﻿//オリジナルは以下にあります
//https://support.embarcadero.com/print/35955
//Delphi XE10 で動作確認を行っています
//2021-05-21 大石剛司
//BXA03027@nifty.ne.jp
//
//使用する場合は、以下を設定します
//Memo1 は、MemoCrLf.pas を使用したコンポーネントを使用します
//改行の表示は Mmeo1 で行い、StringGrid では、改行は表示しません
//セル内に改行がある場合にセル内の複数行を表示する場合は、セルの高さを変更して下さい
//
//type
//  TFormViewGrid = class(TForm)
//
//  private
//    BoolEditModified: Boolean;
//    BoolEditModified: Boolean;
//    IntLastRow: Integer;
//    IntLastCol: Integer;
//
//procedure TFormViewGrid.StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
//var
//  bModified: Boolean;
//begin
//  //編集中
//  if MenuEditMode.Checked = TRUE then
//  begin
//    //前回と異なる
//    if (ACol <> IntLastCol) or (ARow <> IntLastRow)  then
//    begin
//      //前回内容を設定
//      StringGrid1.Cells[IntLastCol, IntLastRow] := Memo1.Text;
//    end;
//    //取得
//    bModified := Memo1.Modified;
//  end
//  else
//  begin
//    bModified := FALSE;
//  end;
//  //再設定する
//  if (BoolEditModified = FALSE) and (bModified = TRUE) then
//  begin
//    BoolEditModified := TRUE;
//  end;
//
//  //新しい内容を設定
//  Memo1.Text := StringGrid1.Cells[ACol, ARow];
//  //リセット
//  BoolMemoShift := FALSE;
//
//  //リセット
//  Memo1.Modified := FALSE;
//
//  //記憶
//  IntLastCol := ACol;
//  IntLastRow := ARow;
//end;
//
//procedure TFormViewGrid.Memo1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
//begin
//  //選択範囲の再描画
//  if Memo1.SelLength > 0 then
//  begin
//    BoolMemoShift := TRUE;
//  end;
//end;
//
//procedure TFormViewGrid.Memo1KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
//begin
//  //選択範囲の再描画
//  if BoolMemoShift = TRUE then
//  begin
//    BoolMemoShift := FALSE;
//    Memo1.Repaint;
//  end;
//end;
//
//procedure TFormViewGrid.Memo1MouseMove(Sender: TObject; Shift: TShiftState; X,Y: Integer);
//begin
//  //選択範囲の再描画
//  if Memo1.SelLength > 0 then
//  begin
//    BoolMemoShift := TRUE;
//  end;
//end;
//
//procedure TFormViewGrid.Memo1MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
//begin
//  //選択範囲の再描画
//  if BoolMemoShift = TRUE then
//  begin
//    BoolMemoShift := FALSE;
//    Memo1.Repaint;
//  end;
//end;
//
//
//----------------------------------------------------------------
//  MlsGrid - MultiLine StringGrid Component
//----------------------------------------------------------------

   unit MlsGrid;

   interface

   uses
     Windows, Messages, SysUtils, Classes, Graphics,
     Controls, Forms, Dialogs, Grids, StdCtrls;
   
   type
     //TMLInplaceEditの宣言 ----------------------
     TMLInplaceEdit = class(TInplaceEdit)
     private
       { Private 宣言 }
     protected
       { Protected 宣言 }
       procedure CreateParams(var Params:TCreateParams); override;
       procedure KeyDown(var Key: Word; Shift: TShiftState); override;
       procedure KeyPress(var Key:Char); override;
     public
       { Public 宣言 }
     published
       { Published 宣言 }
     end;

     //TMLStringGridの宣言------------------------
     TMLStringGrid = class(TStringGrid)
     private
       { Private 宣言 }
       FWordWrap: Boolean;
       FCurrentBackColor: TColor;
       FFixedBackColor: TColor;
       FDarkMode: Boolean;
       procedure SetWordWrap(Value: Boolean);
       procedure SetCurrentBackColor(Value: TColor);
       procedure SetFixedBackColor(Value: TColor);
       procedure SetDarkMode(Value: Boolean);
     protected
       { Protected 宣言 }
       procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
       procedure KeyDown(var Key: Word; Shift:TShiftState); override;
       procedure KeyPress(var Key:Char); override;
       function CreateEditor: TInplaceEdit; override;
     public
       { Public 宣言 }
       //protectedを再宣言
       procedure DeleteRow(ARow: Longint); reintroduce;
       procedure DeleteColumn(ACol: Longint); reintroduce;
       procedure MoveRow(FromIndex, ToIndex: Longint); reintroduce;
       procedure MoveColumn(FromIndex, ToIndex: Longint); reintroduce;
       //protectedを再宣言
     published
       { Published 宣言 }
       property WordWrap : Boolean read FWordWrap
                                   write SetWordWrap
                                   default False;
       //背景色とモード
       property CurrentBackColor : TColor read  FCurrentBackColor
                                          write SetCurrentBackColor
                                          default clWindow;
       property FixedBackColor : TColor   read  FFixedBackColor
                                          write SetFixedBackColor
                                          default clWindow;
       property DarkMode : Boolean        read  FDarkMode
                                          write SetDarkMode
                                          default False;
     end;

   procedure Register;


   implementation


   //TMLInplaceEdit -----------------------------
   procedure TMLInplaceEdit.CreateParams(var Params: TCreateParams);
   begin
     inherited CreateParams(Params);
     // スタイルに複数行エディットを追加
     Params.Style := Params.Style or ES_AUTOHSCROLL or ES_AUTOVSCROLL;
   end;

   procedure TMLInplaceEdit.KeyDown(var Key:Word; Shift: TShiftState);
   begin
     inherited KeyDown(Key, Shift);
   end;

   procedure TMLInplaceEdit.KeyPress(var Key:Char);
   begin
     inherited KeyPress(Key);
   end;


   //TMLStringGrid ------------------------------
   procedure TMLStringGrid.DeleteRow(ARow: Longint);
   begin
     inherited DeleteRow(ARow);
   end;

   procedure TMLStringGrid.DeleteColumn(ACol: Longint);
   begin
     inherited DeleteColumn(ACol);
   end;

   procedure TMLStringGrid.MoveRow(FromIndex, ToIndex: Longint);
   begin
     inherited MoveRow(FromIndex, ToIndex);
   end;

   procedure TMLStringGrid.MoveColumn(FromIndex, ToIndex: Longint);
   begin
     inherited MoveColumn(FromIndex, ToIndex);
   end;

   function TMLStringGrid.CreateEditor: TInplaceEdit;
   begin
     // TInplaceEditのかわりにTMLInplaceEditを生成する
     Result := TInplaceEdit(TMLInplaceEdit.Create(Self));
   end;

   procedure TMLStringGrid.KeyDown(var Key: Word; Shift: TShiftState);
   begin
     inherited KeyDown(Key, Shift);
   end;

   procedure TMLStringGrid.KeyPress(var Key:Char);
   begin
     inherited KeyPress(Key);
   end;

   procedure TMLStringGrid.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);

     //関数内関数の開始
     procedure DrawCellText;
     var
       S: String;
       Options: Integer;
     begin
       S := Cells[ACol, ARow];
       ARect.Left := ARect.Left + 0;
       ARect.Top  := ARect.Top  + 0;
       // メモの表示で描画
       Options := DT_EDITCONTROL;
       if FWordWrap = TRUE then
       begin
         Options := Options or DT_WORDBREAK;
       end;
       if FDarkMode = FALSE then
       begin
         //現在セル
         if (ARow = Row) and (ACol = Col) then
         begin
           Canvas.Brush.Color := FCurrentBackColor;
           Canvas.FillRect(ARect);
         end
         //固定セル
         else if (ARow < FixedRows) or (ACol < FixedCols) then
         begin
           Canvas.Brush.Color := FFixedBackColor;
           Canvas.FillRect(ARect);
         end
         //複数選択は現在セルと同じ描画にする
         else if (ACol >= Selection.Left) and (ACol <= Selection.Right)  and
                 (ARow >= Selection.Top ) and (ARow <= Selection.Bottom) then
         begin
           Canvas.Brush.Color := FCurrentBackColor;
           Canvas.FillRect(ARect);
         end
         //現在セル以外
         else
         begin
           Canvas.Brush.Color := Self.Color;
           Canvas.FillRect(ARect);
         end;
         //表示
         DrawTextEx(Canvas.Handle, PChar(S), Length(S), ARect, Options, nil);
       end
       else
       begin
         //表示
         DrawTextEx(Canvas.Handle, PChar(S), Length(S), ARect, Options, nil);
       end;
     end;
     //関数内関数の終了

   var
     StoreDef: Boolean;
   begin
     //設定を記憶
     StoreDef := DefaultDrawing;
     //一旦無効にする
     DefaultDrawing := False;
     //DefaultDrawingが有効なら改行表示を行う
     if StoreDef = TRUE then
     begin
       DrawCellText;
     end
     //DefaultDrawingが無効ならonDrawSellを行う
     else
     begin
       inherited DrawCell(ACol, ARow, ARect, AState);
     end;
     //設定を戻す
     DefaultDrawing := StoreDef;
   end;

   procedure TMLStringGrid.SetWordWrap(Value: Boolean);
   begin
     if Value <> FWordWrap then
     begin
       FWordWrap := Value;
       Invalidate;
     end;
   end;

   procedure TMLStringGrid.SetCurrentBackColor(Value: Tcolor);
   begin
     if Value <> FCurrentBackColor then
     begin
       FCurrentBackColor := Value;
       Invalidate;
     end;
   end;

   procedure TMLStringGrid.SetFixedBackColor(Value: Tcolor);
   begin
     if Value <> FFixedBackColor then
     begin
       FFixedBackColor := Value;
       Invalidate;
     end;
   end;

   procedure TMLStringGrid.SetDarkMode(Value: Boolean);
   begin
     if Value <> FDarkMode then
     begin
       FDarkMode := Value;
       Invalidate;
     end;
   end;

   //---------------------------------------------
   procedure Register;
   begin
     RegisterComponents('Samples', [TMLStringGrid]);
   end;

   end.