﻿unit Remove;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls, Vcl.ComCtrls,
  IniFiles, MyUtility;

const
  FORM_WIDTH_SET  = 640;
  FORM_HEIGHT_SET = 430;
  FORM_MAX        = 'MAX';
  FORM_MIN        = 'MIN';
  FORM_NORMAL     = 'NORMAL';
  LIST_COLUM_WIDH_0    =  30; //
  LIST_COLUM_WIDH_1    =  60; //root
  LIST_COLUM_WIDH_2    = 105; //volume
  LIST_COLUM_WIDH_3    =  85; //type
  LIST_COLUM_WIDH_4    = 115; //serial
  LIST_COLUM_WIDH_5    =  55; //file
  LIST_COLUM_WIDH_6    = 220; //disk
  LIST_COLUM_WIDH_7    = 220; //free
  LIST_COLUM_WIDH_8    = 220; //use

type
  TRadioGroup = class(Vcl.ExtCtrls.TRadioGroup)
  published
    property OnKeyDown;
  end;

type
  TFormRemove = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    ButtonOK: TButton;
    ButtonCancel: TButton;
    Panel3: TPanel;
    Label1: TLabel;
    ButtonUpdate: TButton;
    Timer1: TTimer;
    RadioGroup9: TRadioGroup;
    ListView1: TListView;

    procedure CreateParams(var Params: TCreateParams); override;
    procedure WmGetMinMaxInfo(var msg : TWMGetMinMaxInfo); message WM_GETMINMAXINFO;

    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);

    procedure ButtonOKClick(Sender: TObject);
    procedure ButtonCancelClick(Sender: TObject);
    procedure ButtonUpdateClick(Sender: TObject);

    procedure ListView1DblClick(Sender: TObject);
    procedure ListView1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);

    procedure SetListViewNewItem(iViewIndex: Integer; sItem: String);
    procedure Timer1Timer(Sender: TObject);

    procedure RadioGroupKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure RadioGroup9Click(Sender: TObject);

  private
    { Private 宣言 }
    FormPosLeft: Integer;
    FormPosTop: Integer;
    FormPosWidth: Integer;
    FormPosHeight: Integer;
    FormPosRight: Integer;
    FormPosButtom: Integer;
    MonitorTop: Integer;
    MonitorLeft: Integer;
    MonitorRight: Integer;
    MonitorButtom: Integer;
    MonitorWidth: Integer;
    MonitorHeight: Integer;
    StrFormState: String;

    BoolJapanese: Boolean;
    StrLanguageFileName: String;

  public
    { Public 宣言 }

  end;

var
  FormRemove: TFormRemove;

implementation

{$R *.dfm}

uses
  Main;

function OpenVolume(ADrive: char): THandle;
var
  RootName: string;
  VolumeName: string;
  AccessFlags: DWORD;
begin
  //ルート文字
  // ADrive + ':\' kills the syntax highlighting
  RootName := ADrive + ':\';

  //タイプを取得
  case GetDriveType(PChar(RootName)) of

    DRIVE_REMOVABLE:
      AccessFlags := GENERIC_READ or GENERIC_WRITE;

    DRIVE_FIXED:
      AccessFlags := GENERIC_READ or GENERIC_WRITE;

    DRIVE_CDROM:
      AccessFlags := GENERIC_READ;

  else
    Result := INVALID_HANDLE_VALUE;
    exit;
  end;

  //設定
  VolumeName := Format('\\.\%s:', [ADrive]);
  //作成
  Result := CreateFile(PChar(VolumeName), AccessFlags, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
  //結果
  if Result = INVALID_HANDLE_VALUE then
  begin
    RaiseLastOSError;
  end;
end;

function LockVolume(AVolumeHandle: THandle): boolean;

const
  LOCK_TIMEOUT = 5 * 1000; // 5 Seconds
  LOCK_RETRIES = 5;
  LOCK_SLEEP = LOCK_TIMEOUT div LOCK_RETRIES;

// #define FSCTL_LOCK_VOLUME CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 6, METHOD_BUFFERED, FILE_ANY_ACCESS)
  _FSCTL_LOCK_VOLUME = (9 shl 16) or (0 shl 14) or (6 shl 2) or 0;

var
  Retries: integer;
  BytesReturned: Cardinal;
begin
  for Retries := 1 to LOCK_RETRIES do
  begin
    //取得
    Result := DeviceIoControl(AVolumeHandle, FSCTL_LOCK_VOLUME, nil, 0, nil, 0, BytesReturned, nil);
    //結果
    if Result then
    begin
      break;
    end;
    //スリープ
    Sleep(LOCK_SLEEP);
  end;
end;

function DismountVolume(AVolumeHandle: THandle): boolean;

const
// #define FSCTL_DISMOUNT_VOLUME CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 8, METHOD_BUFFERED, FILE_ANY_ACCESS)
  _FSCTL_DISMOUNT_VOLUME = (9 shl 16) or (0 shl 14) or (8 shl 2) or 0;

var
  BytesReturned: Cardinal;
begin
  //取得
  Result := DeviceIoControl(AVolumeHandle, FSCTL_DISMOUNT_VOLUME, nil, 0, nil, 0, BytesReturned, nil);
  //結果
  if not Result then
  begin
    RaiseLastOSError;
  end;
end;

function PreventRemovalOfVolume(AVolumeHandle: THandle; APreventRemoval: boolean): boolean;

const
// #define IOCTL_STORAGE_MEDIA_REMOVAL CTL_CODE(IOCTL_STORAGE_BASE, 0x0201, METHOD_BUFFERED, FILE_READ_ACCESS)
  _IOCTL_STORAGE_MEDIA_REMOVAL = ($2d shl 16) or (1 shl 14) or ($201 shl 2) or 0;

type
  TPreventMediaRemoval = record
    PreventMediaRemoval: BOOL;
  end;

var
  BytesReturned: Cardinal;
  PMRBuffer: TPreventMediaRemoval;
begin
  //設定
  PMRBuffer.PreventMediaRemoval := APreventRemoval;
  //取得
  Result := DeviceIoControl(AVolumeHandle, IOCTL_STORAGE_MEDIA_REMOVAL,
    @PMRBuffer, SizeOf(TPreventMediaRemoval), nil, 0, BytesReturned, nil);
  //結果
  if not Result then
  begin
    RaiseLastOSError;
  end;
end;

function AutoEjectVolume(AVolumeHandle: THandle): boolean;

const
// #define IOCTL_STORAGE_EJECT_MEDIA CTL_CODE(IOCTL_STORAGE_BASE, 0x0202, METHOD_BUFFERED, FILE_READ_ACCESS)
//  IOCTL_STORAGE_EJECT_MEDIA = ($2d shl 16) or (1 shl 14) or ($202 shl 2) or 0;
  _IOCTL_STORAGE_EJECT_MEDIA: Cardinal      = $002d4808;
  _IOCTL_STORAGE_EJECTION_CONTROL: Cardinal = $002d0940;

var
  BytesReturned: Cardinal;
begin
  //取得
  Result := DeviceIoControl(AVolumeHandle, IOCTL_STORAGE_EJECT_MEDIA, nil, 0, nil, 0, BytesReturned, nil);
  //結果
  if not Result then
  begin
    RaiseLastOSError;
  end;
end;

function EjectVolume(ADrive: char): boolean;
var
  VolumeHandle: THandle;
begin
  //結果
  Result := FALSE;

  //開く
  VolumeHandle := OpenVolume(ADrive);
  //失敗
  if VolumeHandle = INVALID_HANDLE_VALUE then
  begin
    exit;
  end;

  //実行
  try
    //ロック
    if LockVolume(VolumeHandle) then
    begin
      //取り外し
      if DismountVolume(VolumeHandle) then
      begin
        // Set prevent removal to false and eject the volume
        if PreventRemovalOfVolume(VolumeHandle, FALSE) then
        begin
          Result := AutoEjectVolume(VolumeHandle);
        end;
      end;
    end
    //失敗
    else
    begin
      //設定
      if FormRemove.BoolJapanese = TRUE then
      begin
        FormMain.StrMsg := 'Information : ' + ADrive + ' ドライブは使用中です';
      end
      else
      begin
        FormMain.StrMsg := 'Information : ' + ADrive + ' Drive is in using.' + #13#10 +
                           '[OK]';
      end;
      ShowMessage(FormMain.StrMsg);
      Result := FALSE;
    end;
  finally
    //閉じる
    CloseHandle(VolumeHandle);
  end;
end;




//クリックした時にフォームをアクティブにするために必要
procedure TFormRemove.CreateParams(var Params: TCreateParams);
begin
  inherited;
  Params.WndParent := Application.Handle;
end;

procedure TFormRemove.WmGetMinMaxInfo(var msg : TWMGetMinMaxInfo);
begin
  //最小サイズ制限
  with msg.MinMaxInfo^.ptMinTrackSize do
  begin
    x := FORM_WIDTH_SET;
    y := FORM_HEIGHT_SET;
  end;
  //最大サイズ制限
  with msg.MinMaxInfo^.ptMaxTrackSize do
  begin
    //
  end;
  //最大化時サイズ制限
  with msg.MinMaxInfo^.ptMaxSize do
  begin
    //
  end;
end;


procedure TFormRemove.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  //IMEをオフ
  SetImeMode(Self.Handle, imClose);
end;

procedure TFormRemove.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
;
end;

procedure TFormRemove.FormCreate(Sender: TObject);
var
  IniFile: TIniFile;
begin
  //イベント登録
  RadioGroup9.OnKeyDown := RadioGroupKeyDown;

  //MessageDialog の音を消す対策
  UseLatestCommonDialogs := FALSE;

  //確保
  IniFile := TIniFile.Create(ChangeFileExt(Application.ExeName, '.ini'));
  //読込
  FormPosLeft   := IniFile.ReadInteger('FormRemove', 'Left',   0);
  FormPosTop    := IniFile.ReadInteger('FormRemove', 'Top',    0);
  FormPosWidth  := IniFile.ReadInteger('FormRemove', 'Width',  FORM_WIDTH_SET);
  FormPosHeight := IniFile.ReadInteger('FormRemove', 'Height', FORM_HEIGHT_SET);
  StrFormState  := IniFile.ReadString ('FormRemove', 'State',  FORM_NORMAL);
  FormPosRight  := FormPosLeft + FormPosWidth;
  FormPosButtom := FormPosTop  + FormPosHeight;
  //破棄
  IniFile.Free;

  //初期化
  Timer1.Enabled := FALSE;

  //初期化
  ListView1.Columns.Add.Caption := '　';
  ListView1.Columns.Add.Caption := 'Root';
  ListView1.Columns.Add.Caption := 'Volume';
  ListView1.Columns.Add.Caption := 'Type';
  ListView1.Columns.Add.Caption := 'Serial';
  ListView1.Columns.Add.Caption := 'File';
  ListView1.Columns.Add.Caption := 'Disk';
  ListView1.Columns.Add.Caption := 'Free';
  ListView1.Columns.Add.Caption := 'Use';
  ListView1.ViewStyle := vsReport;
  ListView1.RowSelect := TRUE;
  ListView1.MultiSelect := FALSE;
  ListView1.Visible := TRUE;
  ListView1.ReadOnly := TRUE;
  ListView1.HideSelection := FALSE;     //フォーカスが移動しても残す

  //確保
  IniFile := TIniFile.Create(ChangeFileExt(Application.ExeName, '.ini'));
  //設定
  ListView1.Columns[ 0].Width := IniFile.ReadInteger('ListViewWidth4',  '0', LIST_COLUM_WIDH_0);
  ListView1.Columns[ 1].Width := IniFile.ReadInteger('ListViewWidth4',  '1', LIST_COLUM_WIDH_1);
  ListView1.Columns[ 2].Width := IniFile.ReadInteger('ListViewWidth4',  '2', LIST_COLUM_WIDH_2);
  ListView1.Columns[ 3].Width := IniFile.ReadInteger('ListViewWidth4',  '3', LIST_COLUM_WIDH_3);
  ListView1.Columns[ 4].Width := IniFile.ReadInteger('ListViewWidth4',  '4', LIST_COLUM_WIDH_4);
  ListView1.Columns[ 5].Width := IniFile.ReadInteger('ListViewWidth4',  '5', LIST_COLUM_WIDH_5);
  ListView1.Columns[ 6].Width := IniFile.ReadInteger('ListViewWidth4',  '6', LIST_COLUM_WIDH_6);
  ListView1.Columns[ 7].Width := IniFile.ReadInteger('ListViewWidth4',  '7', LIST_COLUM_WIDH_7);
  ListView1.Columns[ 8].Width := IniFile.ReadInteger('ListViewWidth4',  '8', LIST_COLUM_WIDH_8);
  //破棄
  IniFile.Free;
end;

procedure TFormRemove.FormDestroy(Sender: TObject);
var
  IniFile: TIniFile;
begin
  //確保
  IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
  //画面状態
  if WindowState = wsMaximized then
  begin
    IniFile.WriteString('FormRemove', 'State', FORM_MAX);
  end
  else if WindowState = wsMinimized then
  begin
    IniFile.WriteString('FormRemove', 'State', FORM_MIN);
  end
  else
  begin
    IniFile.WriteString ('FormRemove', 'State',  FORM_NORMAL);
    IniFile.WriteInteger('FormRemove', 'Left',   FormRemove.Left);
    IniFile.WriteInteger('FormRemove', 'Top',    FormRemove.Top);
    IniFile.WriteInteger('FormRemove', 'Width',  FormRemove.Width);
    IniFile.WriteInteger('FormRemove', 'Height', FormRemove.Height);
  end;
  //破棄
  IniFile.Free;

  //確保
  IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
  //設定
  IniFile.WriteInteger('ListViewWidth4',  '0', ListView1.Columns[ 0].Width);
  IniFile.WriteInteger('ListViewWidth4',  '1', ListView1.Columns[ 1].Width);
  IniFile.WriteInteger('ListViewWidth4',  '2', ListView1.Columns[ 2].Width);
  IniFile.WriteInteger('ListViewWidth4',  '3', ListView1.Columns[ 3].Width);
  IniFile.WriteInteger('ListViewWidth4',  '4', ListView1.Columns[ 4].Width);
  IniFile.WriteInteger('ListViewWidth4',  '5', ListView1.Columns[ 5].Width);
  IniFile.WriteInteger('ListViewWidth4',  '6', ListView1.Columns[ 6].Width);
  IniFile.WriteInteger('ListViewWidth4',  '7', ListView1.Columns[ 7].Width);
  IniFile.WriteInteger('ListViewWidth4',  '8', ListView1.Columns[ 8].Width);
  //保存
  IniFile.WriteInteger('RemovableHardWare',   'Index', RadioGroup9.ItemIndex);
  //破棄
  IniFile.Free;
end;

procedure TFormRemove.FormResize(Sender: TObject);
begin
;
end;

procedure TFormRemove.FormShow(Sender: TObject);
var
  IniFile: TIniFile;
begin
  //Formが自動生成でないので、CreateでなくShowで行う
  //FormKeyDownのイベントを設定する
  FormRemove.KeyPreview := TRUE;
  //GetSystemMetricst関数を使用して，仮想画面全体のサイズを取得
  MonitorTop    := GetSystemMetrics(SM_YVIRTUALSCREEN);
  MonitorLeft   := GetSystemMetrics(SM_XVIRTUALSCREEN);
  MonitorWidth  := GetSystemMetrics(SM_CXVIRTUALSCREEN);;
  MonitorHeight := GetSystemMetrics(SM_CYVIRTUALSCREEN);;
  MonitorRight  := MonitorLeft + MonitorWidth;
  MonitorButtom := MonitorTop  + MonitorHeight;

  //表示できれば 前回の状態で表示
  if ( (MonitorTop  <= FormPosTop  + FORM_TOP_SHIFT ) and (FormPosTop    <= MonitorButtom) ) and
     ( (MonitorTop  <= FormPosButtom                ) and (FormPosButtom <= MonitorButtom) ) and
     ( (MonitorLeft <= FormPosLeft + FORM_LEFT_SHIFT) and (FormPosLeft   <= MonitorRight ) ) and
     ( (MonitorLeft <= FormPosRight                 ) and (FormPosRight  <= MonitorRight ) ) then
  begin
    FormRemove.SetBounds(FormPosLeft, FormPosTop, FormPosWidth, FormPosHeight);
  end
  //画面内に表示できなければ デフォルトで表示
  else
  begin
    FormRemove.SetBounds(FormPosLeft, FormPosTop, FormPosWidth, FormPosHeight);
  end;

  //最大
  if StrFormState = FORM_MAX then
  begin
    FormRemove.WindowState := wsMaximized;
  end
  //最小
  else if StrFormState = FORM_MIN then
  begin
    FormRemove.WindowState := wsMinimized;
  end
  //標準
  else
  begin
    //
  end;

  //IMEをオフ
  SetImeMode(FormRemove.Handle, imClose);

  //設定
  BoolJapanese := FormMain.BoolJapanese;
  //設定
  StrLanguageFileName := ExtractFilePath(ParamStr(0)) + FOLDER_LANGUAGE  + '\' +
                         ChangeFileExt(ExtractFileName(ParamStr(0)), '') + '\' + 'Remove.txt';
  if BoolJapanese = TRUE then
  begin
    //日本語
    FormMain.SetJapanese(FormRemove, StrLanguageFileName, Sender);
  end
  else
  begin
    //英語
    FormMain.SetEnglish(FormRemove, StrLanguageFileName, Sender);
  end;

  //確保
  IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
  //読込
  RadioGroup9.ItemIndex := IniFile.ReadInteger('RemovableHardWare',   'Index', 0);
  //破棄
  IniFile.Free;

  //ドライブ取得
  ButtonUpdateClick(Sender);

end;

procedure TFormRemove.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  IniFile: TIniFile;
begin
  //ESC
  if      (Shift = []) and (key = VK_ESCAPE) then
  begin
    ButtonCancelClick(Sender);
  end
  //日本語
  else if ((Shift = [ssShift, ssCtrl]) and (key = VK_J)) and (FormMain.BoolDeveloper = TRUE) then
  begin
    //設定
    BoolJapanese := TRUE;
    //日本語
    FormMain.SetJapanese(FormRemove, StrLanguageFileName, Sender);
    //確保
    IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
    //読込
    RadioGroup9.ItemIndex := IniFile.ReadInteger('RemovableHardWare',   'Index', 0);
    //破棄
    IniFile.Free;
  end
  //英語
  else if ((Shift = [ssShift, ssCtrl]) and (key = VK_E)) and (FormMain.BoolDeveloper = TRUE) then
  begin
    //設定
    BoolJapanese := FALSE;
    //英語
    FormMain.SetEnglish(FormRemove, StrLanguageFileName, Sender);
    //確保
    IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
    //読込
    RadioGroup9.ItemIndex := IniFile.ReadInteger('RemovableHardWare',   'Index', 0);
    //破棄
    IniFile.Free;
  end
  //
  else
  begin
    //
  end;
end;

procedure TFormRemove.RadioGroupKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  //RETURN
  if (Shift = []) and (key = VK_RETURN) then
  begin
    key := 0;
    ButtonOK.SetFocus;
    ButtonOKClick(Sender);
  end
  else
  begin
    //
  end;
end;

procedure TFormRemove.RadioGroup9Click(Sender: TObject);
begin
  //ドライブ取得
  ButtonUpdateClick(Sender);
end;

procedure TFormRemove.ListView1DblClick(Sender: TObject);
begin
  ButtonOK.SetFocus;
  ButtonOKClick(Sender);
end;

procedure TFormRemove.ListView1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  //ENTER
  if      (Shift = []) and (key = VK_RETURN) then
  begin
    key := 0;
    ButtonOK.SetFocus;
    ButtonOKClick(Sender);
  end
  else
  begin
    //
  end;
end;

procedure TFormRemove.ButtonOKClick(Sender: TObject);
var
  sStr1: String;
  sStr2: String;
  sStr11: String;
  sStr21: String;
  sStr22: String;
  sStr23: String;
  bResult: Boolean;
begin
  if Timer1.Enabled = TRUE then
  begin
    Exit;
  end;

  if ListView1.Items.Count > 0 then
  begin
    //ドライブ文字
    sStr1 := Copy(ListView1.Items[ListView1.ItemIndex].SubItems[0], 1, 1);
    //ルート
    sStr2 := Copy(FormMain.StrCurrent, 1, 1);
    if UpperCase(sStr1) = UpperCase(sStr2) then
    begin
      //設定
      if FormRemove.BoolJapanese = TRUE then
      begin
        FormMain.StrMsg := 'Information : カレントのドライブと取り出しのドライブが同一の場合は操作できません';
      end
      else
      begin
        FormMain.StrMsg := 'Information : Drive can not remove, because of drive is same as current drive.' + #13#10 +
                           '[OK]';
      end;
      ShowMessage(FormMain.StrMsg);
      Exit;
    end;


    //USBremove
    if      RadioGroup9.ItemIndex = 0 then
    begin
      //設定
      sStr11 := '';
      if FormMain.StrAnnsFMPath = '' then
      begin
        sStr11 := ExtractFilePath(ParamStr(0)) + USB_REMOVE_EXE;
      end
      else
      begin
        sStr11 := FormMain.StrAnnsFMPath + USB_REMOVE_EXE;
        if FileExists(sStr11) = FALSE then
        begin
          sStr11 := ExtractFilePath(ParamStr(0)) + USB_REMOVE_EXE;
        end;
      end;

      //ファイルがある
      if FileExists(sStr11) = TRUE then
      begin
        //実行ファイル名
        sStr21 := sStr11;
        //R は、成功して 取り外した時のメッセージを出しません
        sStr22 := sStr1+ ': /R';
        //ファイル名
        sStr23 := '';
        //ExecuteProgram(sExeName: String; sOption: String; sFileName: String): Boolean;
        FormMain.ExecuteProgram(sStr21, sStr22, sStr23);

        //再取得
        Timer1.Enabled := TRUE;
        //設定
        if FormRemove.BoolJapanese = TRUE then
        begin
          FormMain.StrMsg := 'Information : ドライブ ' + sStr1 + ' の取外を操作しました' + #13#10 +
                             '結果はリスト一覧で確認して下さい';
        end
        else
        begin
          FormMain.StrMsg := 'Information : Drive ' + sStr1 + ' had removed.' + #13#10 +
                             'Confirm result at drive list.' + #13#10 +
                             '[OK]';
        end;
        ShowMessage(FormMain.StrMsg);
      end
      else
      begin
        //設定
        if FormRemove.BoolJapanese = TRUE then
        begin
          FormMain.StrMsg := 'Information : ファイルがありません' + #13#10 +
                             USB_REMOVE_EXE;
        end
        else
        begin
          FormMain.StrMsg := 'Information : File is not exist.' + #13#10 +
                             USB_REMOVE_EXE + #13#10 +
                             '[OK]';
        end;
        ShowMessage(FormMain.StrMsg);
      end;
    end
    //EjctClse
    else if RadioGroup9.ItemIndex = 1 then
    begin
      //設定
      sStr11 := '';
      if FormMain.StrAnnsFMPath = '' then
      begin
        sStr11 := ExtractFilePath(ParamStr(0)) + EJECT_CLOSE_EXE;
      end
      else
      begin
        sStr11 := FormMain.StrAnnsFMPath + EJECT_CLOSE_EXE;
        if FileExists(sStr11) = FALSE then
        begin
          sStr11 := ExtractFilePath(ParamStr(0)) + EJECT_CLOSE_EXE;
        end;
      end;

      //ファイルがある
      if FileExists(sStr11) = TRUE then
      begin
        //実行ファイル名
        sStr21 := sStr11;
        //オプション
        //E は、Eject トレイを開きます
        sStr22 := sStr1+ ':\' + ' ' + '/E';
        //ファイル名
        sStr23 := '';
        //ExecuteProgram(sExeName: String; sOption: String; sFileName: String): Boolean;
        FormMain.ExecuteProgram(sStr21, sStr22, sStr23);

        //再取得
        Timer1.Enabled := TRUE;
        //設定
        if FormRemove.BoolJapanese = TRUE then
        begin
          FormMain.StrMsg := 'Information : ドライブ ' + sStr1 + ' のメディアの排出を操作しました' + #13#10 +
                             '結果はリスト一覧で確認して下さい';
        end
        else
        begin
          FormMain.StrMsg := 'Information : Drive ' + sStr1 + ' media had removed.' + #13#10 +
                             'Confirm result at drive list.' + #13#10 +
                             '[OK]';
        end;
        ShowMessage(FormMain.StrMsg);
      end
      else
      begin
        //設定
        if FormRemove.BoolJapanese = TRUE then
        begin
          FormMain.StrMsg := 'Information : ファイルがありません' + #13#10 +
                             EJECT_CLOSE_EXE;
        end
        else
        begin
          FormMain.StrMsg := 'Information : File is not exist.' + #13#10 +
                             EJECT_CLOSE_EXE + #13#10 +
                             '[OK]';
        end;
        ShowMessage(FormMain.StrMsg);
      end;
    end
    //API
    else
    begin
        //取り出し
        bResult := FALSE;
        try
          bResult := EjectVolume(char(sStr1[1]));
        except
          //設定
          if FormRemove.BoolJapanese = TRUE then
          begin
            FormMain.StrMsg := 'Information : ドライブ ' + sStr1 + ' のメディアの排出が出来ませんでした';
          end
          else
          begin
            FormMain.StrMsg := 'Information : Drive ' + sStr1 + ' media could not removed.' + #13#10 +
                               '[OK]';
          end;
          ShowMessage(FormMain.StrMsg);
        end;
        if bResult = TRUE then
        begin
          //再取得
          Timer1.Enabled := TRUE;
          //設定
          if FormRemove.BoolJapanese = TRUE then
          begin
            FormMain.StrMsg := 'Information : ドライブ ' + sStr1 + ' のメディアを排出しました';
          end
          else
          begin
            FormMain.StrMsg := 'Information : Drive ' + sStr1 + ' media had removed.' + #13#10 +
                               '[OK]';
          end;
          ShowMessage(FormMain.StrMsg);
        end;
    end;
  end;

end;

procedure TFormRemove.ButtonCancelClick(Sender: TObject);
begin
  //終了
  ModalResult := mrCancel;
end;

procedure TFormRemove.SetListViewNewItem(iViewIndex: Integer; sItem: String);
var
  nItem: TListItem;
  CsvList2: TStringList;
begin
  //確保
  CsvList2 := TStringList.Create;
  //展開
  StringToCsvList(sItem, CsvList2);

  //1行を設定
  nItem := ListView1.Items.Add;
  nItem.Caption := '';
  //
  ListView1.Items.Item[iViewIndex].SubItems.add(CsvList2[0]);
  ListView1.Items.Item[iViewIndex].SubItems.add(CsvList2[1]);
  ListView1.Items.Item[iViewIndex].SubItems.add(CsvList2[2]);
  ListView1.Items.Item[iViewIndex].SubItems.add(CsvList2[3]);
  ListView1.Items.Item[iViewIndex].SubItems.add(CsvList2[4]);
  ListView1.Items.Item[iViewIndex].SubItems.add(CsvList2[5]);
  ListView1.Items.Item[iViewIndex].SubItems.add(CsvList2[6]);
  ListView1.Items.Item[iViewIndex].SubItems.add(CsvList2[7]);

  //破棄
  CsvList2.Free;
end;

procedure TFormRemove.Timer1Timer(Sender: TObject);
begin
  Timer1.Enabled := FALSE;
  ButtonUpdateClick(Sender);
end;

procedure TFormRemove.ButtonUpdateClick(Sender: TObject);
  //ドライブタイプ取得
  function GetType(RootPathName: string): string;
  var
    iRet1: Integer;
  begin
    iRet1 := GetDriveType(PWideChar(RootPathName));
    case iRet1 of
      DRIVE_NO_ROOT_DIR: Result := 'No Root';
      DRIVE_REMOVABLE:   Result := 'Removable';
      DRIVE_FIXED:       Result := 'HDD';
      DRIVE_REMOTE:      Result := 'Network';
      DRIVE_CDROM:       Result := 'CD-ROM';
      DRIVE_RAMDISK:     Result := 'RamDisk';
      else               Result := 'Unknown';
    end;
  end;
var
  StrList1: TStringList;
  CsvList1: TStringList;
  sStr1: String;
  iRow1: Integer;

  iIndex1: Integer;
  iIndex2: Integer;
  iLogDrives: Cardinal;
  sDriveName : String;
  cVol: array[0..255] of Char;
  cFileSys: array[0..255] of Char;
  iSerialNumber: Cardinal;
  iMaxN: Cardinal;
  iFSF: Cardinal;
  iSectPerClstr: Cardinal;
  iBytePerSect: Cardinal;
  iFreeClstr: Cardinal;
  iClstr: Cardinal;
  iDiskSpace: Int64;
  iFreeSpace: Int64;
  iUsedSpace: Int64;
begin
  //確保
  //確保
  StrList1 := TStringList.Create;
  CsvList1 := TStringList.Create;

  //現在使用可能なディスクドライブのビットマスクを取得
  iLogDrives := GetLogicalDrives;

  //ドライブの最大数は25(A-Z)
  for iIndex1 := 0 to 25 do
  begin
    //ビット操作
    if (iLogDrives and (1 shl iIndex1)) <> 0 then
    begin
      //設定
      sDriveName := Char(Ord('A') + iIndex1) + ':\';
      sStr1 := '';

      //ドライブの情報取得
      //USB のマルチカードで、メディアが入っていない場合に FALSE が返る
      if GetVolumeInformation(PWideChar(sDriveName), cVol, 256, @iSerialNumber, iMaxN, iFSF, cFileSys, 256) = TRUE then
      begin
        //ディスクの使用状況取得
        GetDiskFreeSpace(PWideChar(sDriveName), iSectPerClstr, iBytePerSect, iFreeClstr, iClstr);
        iDiskSpace := Int64(iClstr) * Int64(iBytePerSect) * Int64(iSectPerClstr);
        iFreeSpace := Int64(iFreeClstr) * Int64(iBytePerSect) * Int64(iSectPerClstr);
        iUsedSpace := iDiskSpace - iFreeSpace;

        //USBremove
        if      RadioGroup9.ItemIndex = 0 then
        begin
            //情報を作成
            sStr1 := sDriveName                        + #09 +
                     StrPas(cVol)                      + #09 +
                     GetType(sDriveName)               + #09 +
                     '0x' + IntToHex(iSerialNumber, 8) + #09 +
                     StrPas(cFileSys)                  + #09 +
                     ConvertInt64ToStr20(iDiskSpace)   + #09 +
                     ConvertInt64ToStr20(iFreeSpace)   + #09 +
                     ConvertInt64ToStr20(iUsedSpace);
        end
        //EjctClse
        else if RadioGroup9.ItemIndex = 1 then
        begin
          //CD-ROM
          if (UpperCase(GetType(sDriveName)) = UpperCase('CD-ROM')) then
          begin
            //情報を作成
            sStr1 := sDriveName                        + #09 +
                     StrPas(cVol)                      + #09 +
                     GetType(sDriveName)               + #09 +
                     '0x' + IntToHex(iSerialNumber, 8) + #09 +
                     StrPas(cFileSys)                  + #09 +
                     ConvertInt64ToStr20(iDiskSpace)   + #09 +
                     ConvertInt64ToStr20(iFreeSpace)   + #09 +
                     ConvertInt64ToStr20(iUsedSpace);
          end
          else
          begin
            //
          end;
        end
        //API
        else
        begin
          //フロッピーは除外
          if (UpperCase(sDriveName) = UpperCase('A:\')) or (UpperCase(sDriveName) = UpperCase('B:\')) then
          begin
            //
          end
          else
          begin
            //情報を作成
            sStr1 := sDriveName                        + #09 +
                     StrPas(cVol)                      + #09 +
                     GetType(sDriveName)               + #09 +
                     '0x' + IntToHex(iSerialNumber, 8) + #09 +
                     StrPas(cFileSys)                  + #09 +
                     ConvertInt64ToStr20(iDiskSpace)   + #09 +
                     ConvertInt64ToStr20(iFreeSpace)   + #09 +
                     ConvertInt64ToStr20(iUsedSpace);
          end;
        end;
      end
      else
      begin
        //USBremove
        if      RadioGroup9.ItemIndex = 0 then
        begin
            //情報を作成
            sStr1 := sDriveName + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     '';
        end
        //EjctClse
        else if RadioGroup9.ItemIndex = 1 then
        begin
          //CD-ROM
          if (UpperCase(GetType(sDriveName)) = UpperCase('CD-ROM')) then
          begin
            //情報を作成
            sStr1 := sDriveName + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     '';
          end
          else
          begin
            //
          end;
        end
        //API
        else
        begin
          //フロッピーは除外
          if (UpperCase(sDriveName) = UpperCase('A:\')) or (UpperCase(sDriveName) = UpperCase('B:\')) then
          begin
            //
          end
          else
          begin
            //情報を作成
            sStr1 := sDriveName + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     ''         + #09 +
                     '';
          end;
        end;

      end;

      //
      if sStr1 <> '' then
      begin
        StrList1.Add(sStr1);
      end;
    end;
  end;

  //設定
  ListView1.Items.BeginUpdate;
  ListView1.Items.Clear;
  iIndex2 := 0;
  for iRow1 := 0 to StrList1.Count-1 do
  begin
    //ここで各項目を設定してインデックスを加算
    CsvList1.Clear;
    StringToCsvlist(StrList1[iRow1], CsvList1);
    //ドライブを判定
    if (UpperCase(CsvList1.Strings[2]) = UpperCase('Removable')) or (UpperCase(CsvList1.Strings[2]) = UpperCase('CD-ROM')) then
    begin
      SetListViewNewItem(iIndex2, StrList1[iRow1]);
      iIndex2 := iIndex2+1;
    end
  end;
  if ListView1.Items.Count > 0 then
  begin
    ListView1.ItemIndex := 0;
  end;
  ListView1.Items.EndUpdate;

  //破棄
  StrList1.Free;
  CsvList1.Free;
end;


end.
