unit ViewGrid;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.Grids, MlsGrid, Vcl.ExtCtrls, Vcl.Menus,
  IniFiles, MyUtility, Vcl.StdCtrls, Vcl.ComCtrls, System.UITypes, Masks,
  ClipBrd, MemoCrLf;

const
  FORM_WIDTH_SET         = 740;
  FORM_HEIGHT_SET        = 480;
  TITLE_VIEW             = ' - Obh\';
  TITLE_VIEW_ENG         = ' - ViewGrid';
  TEXT_HEIGHT            = 22;
  TEXT_SIZE              = 14;
  GRID_COL_WIDTH         = 64;
  GRID_LINE_WIDTH        = 2;
  CSVLIST_ITEM_MAX       = 1024;

  CHAR_UNKNOW_CHECK_SIZE_MAX = 4095;

  CHAR_AUTO               = 'o';
  CHAR_SJIS               = 'Shift_JIS';
  CHAR_JIS                = 'JIS';
  CHAR_EUCJP              = 'EUC-JP';
  CHAR_UTF7               = 'UTF7';
  CHAR_UTF8_BOM_ON        = 'UTF8  [BOML]';
  CHAR_UTF8_BOM_OFF       = 'UTF8  [BOM]';
  CHAR_UTF16_BOM_ON_BE    = 'UTF16 [BOML BigEndian]';
  CHAR_UTF16_BOM_OFF_BE   = 'UTF16 [BOM BigEndian]';
  CHAR_UTF16_BOM_ON_LE    = 'UTF16 [BOML LittleEndian]';
  CHAR_UTF16_BOM_OFF_LE   = 'UTF16 [BOM LittleEndian]';
  CHAR_UNKNOWN            = 's';
  CHAR_BINARY             = 'oCit@C';

  CHAR_SJIS_ENG               = 'Shift_JIS';
  CHAR_JIS_ENG                = 'JIS';
  CHAR_EUCJP_ENG              = 'EUC-JP';
  CHAR_UTF7_ENG               = 'UTF7';
  CHAR_UTF8_BOM_ON_ENG        = 'UTF8  [WithBOM]';
  CHAR_UTF8_BOM_OFF_ENG       = 'UTF8  [NonBOM]';
  CHAR_UTF16_BOM_ON_BE_ENG    = 'UTF16 [WithBOM BigEndian]';
  CHAR_UTF16_BOM_OFF_BE_ENG   = 'UTF16 [NonBOM BigEndian]';
  CHAR_UTF16_BOM_ON_LE_ENG    = 'UTF16 [WithBOM LittleEndian]';
  CHAR_UTF16_BOM_OFF_LE_ENG   = 'UTF16 [NonBOM LittleEndian]';
  CHAR_UNKNOWN_ENG            = 'UnKnown';
  CHAR_BINARY_ENG             = 'BinaryFile';

  ENCODE_AUTO             = 0;
  ENCODE_SJIS             = 1;
  ENCODE_JIS              = 2;
  ENCODE_EUCJP            = 3;
  ENCODE_UTF7             = 4;
  ENCODE_UTF8_BOM_ON      = 5;
  ENCODE_UTF8_BOM_OFF     = 6;
  ENCODE_UTF16_BOM_ON_BE  = 7;
  ENCODE_UTF16_BOM_OFF_BE = 8;
  ENCODE_UTF16_BOM_ON_LE  = 9;
  ENCODE_UTF16_BOM_OFF_LE = 10;
  ENCODE_UNKNOWN          = 20;

  RET_CRLF              = 'Cr+Lf(Windows) ';
  RET_CR                = 'Cr(OS9)';
  RET_LF                = 'Lf(OSX,Android)';
  RET_NONE              = '';
  RET_UNKNOWN           = 's';

  RET_CRLF_ENG          = 'Cr+Lf(Windows) ';
  RET_CR_ENG            = 'Cr(OS9)';
  RET_LF_ENG            = 'Lf(OSX,Android)';
  RET_NONE_ENG          = 'None';
  RET_UNKNOWN_ENG       = 'UnKnown';

  RETURN_CRLF       = 0;
  RETURN_CR         = 1;
  RETURN_LF         = 2;
  RETURN_NONE       = 3;

  EDITOR_READ_ONLY      = '';
  EDITOR_READ_ONLY_ENG  = 'ReadOnly';
  FILE_READ_ONLY        = 't@Cǎp';
  FILE_READ_ONLY_ENG    = 'ReadOnlyFile';
  EDIT_ENABLE_JPN   = 'ҏW';
  EDIT_DESABLE_JPN  = '';
  EDIT_ENABLE_ENG   = 'Editting';
  EDIT_DESABLE_ENG  = '';

  SEPARATER_TAB                  = 'TAB';
  SEPARATER_DOUBLE_CAUTATION     = '","';
  SEPARATER_COMMA                = ',';
  SEPARATER_COMMA_DOUBLE         = ', or ","';
  SEPARATER_UNKNOWN              = 'UnKnown';
  SEPARATER_FIXED                = '';
  SEPARATER_TAB_ENG              = 'TAB';
  SEPARATER_DOUBLE_CAUTATION_ENG = '","';
  SEPARATER_COMMA_ENG            = ',';
  SEPARATER_COMMA_DOUBLE_ENG     = ', or ","';
  SEPARATER_UNKNOWN_ENG          = 'UnKnown';
  SEPARATER_FIXED_ENG            = '';
  SEPARATER_CHAR_FIXED           = '';

  // ShiftJIS ̃t@Cɂ
  UNKOWN_DST_LIST    = 'BonsFMmini-UnKnown.txt';

type
  SJISString   = type AnsiString(932);   // SJIS
  EUCString    = type AnsiString(20932); // EUC
  JISString    = type AnsiString(50220); // JIS ISO-2022JP
  UTF7String   = type AnsiString(65000); // UTF7
  UTF8String   = type AnsiString(65001); // UTF8

type
  TFormViewGrid = class(TForm)
    MainMenu: TMainMenu;
    MenuFile: TMenuItem;
    MenuEscape: TMenuItem;
    MenuQuit: TMenuItem;
    MenuView: TMenuItem;
    MenuLfToTab: TMenuItem;
    MenuLoadFile: TMenuItem;
    MenuEditMode: TMenuItem;
    MenuRatio1: TMenuItem;
    MenuRatio2: TMenuItem;
    MenuRatio3: TMenuItem;
    MenuRatio4: TMenuItem;
    MenuRatio5: TMenuItem;
    MenuRatio6: TMenuItem;
    MenuRatio7: TMenuItem;
    MenuRatio8: TMenuItem;
    MenuRatio9: TMenuItem;
    MenuRatio10: TMenuItem;
    MenuReadSJIS: TMenuItem;
    MenuReadJIS: TMenuItem;
    MenuReadEUCJP: TMenuItem;
    MenuReadUTF7: TMenuItem;
    MenuReadUTF8Bom: TMenuItem;
    MenuReadUTF8: TMenuItem;
    MenuReadUTF16BomBe: TMenuItem;
    MenuReadUTF16Be: TMenuItem;
    MenuReadUTF16BomLe: TMenuItem;
    MenuReadUTF16Le: TMenuItem;
    MenuReadAuto: TMenuItem;
    MenuColumnWidthAdjust: TMenuItem;
    StatusBar1: TStatusBar;
    StatusBar2: TStatusBar;
    Timer4: TTimer;
    Timer5: TTimer;
    MenuDeleteRow: TMenuItem;
    MenuEdit: TMenuItem;
    MenuFileOverWrite: TMenuItem;
    N4: TMenuItem;
    N6: TMenuItem;
    N7: TMenuItem;
    MenuFileSaveAs: TMenuItem;
    SaveDialog1: TSaveDialog;
    N3: TMenuItem;
    MenuRowFixed: TMenuItem;
    MenuColFixed: TMenuItem;
    N8: TMenuItem;
    MenuFontColor: TMenuItem;
    MenuBackColor: TMenuItem;
    FontDialog1: TFontDialog;
    ColorDialog1: TColorDialog;
    MenuFixedBackColor: TMenuItem;
    ColorDialog2: TColorDialog;
    MenuFontColoInit: TMenuItem;
    MenuRowFixedClear: TMenuItem;
    MenuColFixedClear: TMenuItem;
    MenuCurrentBackColor: TMenuItem;
    ColorDialog3: TColorDialog;
    N9: TMenuItem;
    N10: TMenuItem;
    MenuMemoHeight1: TMenuItem;
    MenuMemoHeight2: TMenuItem;
    MenuMemoHeight3: TMenuItem;
    MenuMemoHeight4: TMenuItem;
    MenuMemoHeight5: TMenuItem;
    MenuMemoHeight6: TMenuItem;
    MenuMemoHeight7: TMenuItem;
    MenuMemoHeight8: TMenuItem;
    MenuMemoHeight9: TMenuItem;
    MenuMemoHeight10: TMenuItem;
    MenuStatusColRow: TMenuItem;
    MenuStatusRowCol: TMenuItem;
    MenuSetting: TMenuItem;
    N11: TMenuItem;
    Panel1: TPanel;
    Label1: TLabel;
    Edit1: TEdit;
    ButtonGridNext: TButton;
    ButtonGridPrev: TButton;
    MenuSearch: TMenuItem;
    MenuSearchWordNext: TMenuItem;
    MenuSearchWordPrev: TMenuItem;
    Panel2: TPanel;
    ButtonMemoPrev: TButton;
    ButtonMemoNext: TButton;
    Label2: TLabel;
    PopupMenu1: TPopupMenu;
    PopUpCopy1: TMenuItem;
    PopUpSelectAll1: TMenuItem;
    PopUpPaste1: TMenuItem;
    PopUpCut1: TMenuItem;
    MenuCopyRow: TMenuItem;
    MenuInsertRow: TMenuItem;
    MenuPasteRow: TMenuItem;
    MenuAddRow: TMenuItem;
    N12: TMenuItem;
    MenuDeleteCol: TMenuItem;
    MenuInsertCol: TMenuItem;
    MenuCopyCol: TMenuItem;
    MenuPasteCol: TMenuItem;
    MenuAddCol: TMenuItem;
    MenuCopyToClip: TMenuItem;
    MenuSaveTab: TMenuItem;
    MenuSaveDoubleCautation: TMenuItem;
    MenuSaveComma: TMenuItem;
    N13: TMenuItem;
    MenuSaveSJIS: TMenuItem;
    MenuSaveJIS: TMenuItem;
    MenuSaveEUCJP: TMenuItem;
    MenuSaveUTF7: TMenuItem;
    MenuSaveUTF8Bom: TMenuItem;
    MenuSaveUTF8: TMenuItem;
    MenuSaveUTF16BomBe: TMenuItem;
    MenuSaveUTF16Be: TMenuItem;
    MenuSaveUTF16BomLe: TMenuItem;
    MenuSaveUTF16Le: TMenuItem;
    MenuCsvAuto: TMenuItem;
    MenuCsvEx1: TMenuItem;
    MenuSameAsAnnsView: TMenuItem;
    StatusBar3: TStatusBar;
    StringGrid1: TMLStringGrid;
    Panel3: TPanel;
    N1: TMenuItem;
    MenuRead: TMenuItem;
    MenuSave: TMenuItem;
    N2: TMenuItem;
    N16: TMenuItem;
    MenuClipTab: TMenuItem;
    MenuClipDoubleCautation: TMenuItem;
    MenuClipComma: TMenuItem;
    Memo1: TMemoCrLf;
    ColorDialog4: TColorDialog;
    MenuCrLfColor: TMenuItem;
    N5: TMenuItem;
    N14: TMenuItem;
    N15: TMenuItem;
    MenuSaveFixedLength: TMenuItem;
    MenuCsvEx2: TMenuItem;
    MenuSaveCsvEx2: TMenuItem;
    MenuClipCsvEx2: TMenuItem;

    procedure CreateParams(var Params: TCreateParams); override;
    procedure WmGetMinMaxInfo(var msg : TWMGetMinMaxInfo); message WM_GETMINMAXINFO;

    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormResize(Sender: TObject);
    procedure FormShow(Sender: TObject);

    procedure MenuRatio1Click(Sender: TObject);
    procedure MenuRatio2Click(Sender: TObject);
    procedure MenuRatio3Click(Sender: TObject);
    procedure MenuRatio4Click(Sender: TObject);
    procedure MenuRatio5Click(Sender: TObject);
    procedure MenuRatio6Click(Sender: TObject);
    procedure MenuRatio7Click(Sender: TObject);
    procedure MenuRatio8Click(Sender: TObject);
    procedure MenuRatio9Click(Sender: TObject);
    procedure MenuRatio10Click(Sender: TObject);
    procedure MenuColumnWidthAdjustClick(Sender: TObject);
    procedure MenuReadAutoClick(Sender: TObject);
    procedure MenuReadSJISClick(Sender: TObject);
    procedure MenuReadJISClick(Sender: TObject);
    procedure MenuReadEUCJPClick(Sender: TObject);
    procedure MenuReadUTF7Click(Sender: TObject);
    procedure MenuReadUTF8BomClick(Sender: TObject);
    procedure MenuReadUTF8Click(Sender: TObject);
    procedure MenuReadUTF16BomBeClick(Sender: TObject);
    procedure MenuReadUTF16BeClick(Sender: TObject);
    procedure MenuReadUTF16BomLeClick(Sender: TObject);
    procedure MenuReadUTF16LeClick(Sender: TObject);
    procedure MenuLfToTabClick(Sender: TObject);
    procedure MenuEditModeClick(Sender: TObject);
    procedure MenuLoadFileClick(Sender: TObject);
    procedure MenuDeleteRowClick(Sender: TObject);
    procedure MenuFileOverWriteClick(Sender: TObject);
    procedure MenuEscapeClick(Sender: TObject);
    procedure MenuQuitClick(Sender: TObject);
    procedure MenuFileSaveAsClick(Sender: TObject);
    procedure MenuRowFixedClick(Sender: TObject);
    procedure MenuColFixedClick(Sender: TObject);
    procedure MenuFontColorClick(Sender: TObject);
    procedure MenuBackColorClick(Sender: TObject);
    procedure MenuFixedBackColorClick(Sender: TObject);
    procedure MenuFontColoInitClick(Sender: TObject);
    procedure MenuRowFixedClearClick(Sender: TObject);
    procedure MenuColFixedClearClick(Sender: TObject);
    procedure MenuMemoHeight1Click(Sender: TObject);
    procedure MenuMemoHeight2Click(Sender: TObject);
    procedure MenuMemoHeight3Click(Sender: TObject);
    procedure MenuMemoHeight4Click(Sender: TObject);
    procedure MenuMemoHeight5Click(Sender: TObject);
    procedure MenuMemoHeight6Click(Sender: TObject);
    procedure MenuMemoHeight7Click(Sender: TObject);
    procedure MenuMemoHeight8Click(Sender: TObject);
    procedure MenuMemoHeight9Click(Sender: TObject);
    procedure MenuMemoHeight10Click(Sender: TObject);
    procedure MenuStatusRowColClick(Sender: TObject);
    procedure MenuStatusColRowClick(Sender: TObject);
    procedure MenuCurrentBackColorClick(Sender: TObject);
    procedure MenuSearchWordNextClick(Sender: TObject);
    procedure MenuSearchWordPrevClick(Sender: TObject);
    procedure MenuCopyRowClick(Sender: TObject);
    procedure MenuInsertRowClick(Sender: TObject);
    procedure MenuPasteRowClick(Sender: TObject);

    procedure Timer4Timer(Sender: TObject);
    procedure Timer5Timer(Sender: TObject);

    procedure GetCharCodeRetCode(FileName:String; var CharCode: Integer; var RetCode: Integer; Sender: TObject);

    procedure LoadCsvFile(Sender: TObject);
    procedure SetCsvFile(Sender: TObject);
    procedure SaveCsvFile(FileName: String; Sender: TObject);
    procedure FixedColumnWidthAdjust(Sender: TObject);
    function  GetSeparater(SrcString: String; Sender: TObject): String;
    function  ConvertCellToCsvEx2(SrcString: String; Sender: TObject): String;
    procedure ConvertLfToTab(Sender: TObject);
    procedure StrListToStringGrid(StrList: TStringList; Sender: TObject);
    procedure StringToCsvListEx1(SrcString: String; CsvList: TstringList);
    procedure StringToCsvListEx2(SrcString: String; CsvList: TstringList);

    procedure SetShortCutKey(Sender: TObject);
    procedure SetShortCutNew(Number: String; MenuItemName: TMenuItem; Sender: TObject);
    procedure StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);

    procedure ButtonGridNextClick(Sender: TObject);
    procedure ButtonMemoNextClick(Sender: TObject);
    procedure ButtonMemoPrevClick(Sender: TObject);
    procedure ButtonGridPrevClick(Sender: TObject);

    procedure PopUpCopy1Click(Sender: TObject);
    procedure PopUpCut1Click(Sender: TObject);
    procedure PopUpPaste1Click(Sender: TObject);
    procedure PopUpSelectAll1Click(Sender: TObject);

    procedure MenuAddRowClick(Sender: TObject);
    procedure MenuDeleteColClick(Sender: TObject);
    procedure MenuCopyColClick(Sender: TObject);
    procedure MenuPasteColClick(Sender: TObject);
    procedure MenuInsertColClick(Sender: TObject);
    procedure MenuAddColClick(Sender: TObject);
    procedure MenuCopyToClipClick(Sender: TObject);
    procedure MenuSaveSJISClick(Sender: TObject);
    procedure MenuSaveJISClick(Sender: TObject);
    procedure MenuSaveEUCJPClick(Sender: TObject);
    procedure MenuSaveUTF7Click(Sender: TObject);
    procedure MenuSaveUTF8BomClick(Sender: TObject);
    procedure MenuSaveUTF8Click(Sender: TObject);
    procedure MenuSaveUTF16BomBeClick(Sender: TObject);
    procedure MenuSaveUTF16BeClick(Sender: TObject);
    procedure MenuSaveUTF16BomLeClick(Sender: TObject);
    procedure MenuSaveUTF16LeClick(Sender: TObject);
    procedure MenuSaveTabClick(Sender: TObject);
    procedure MenuSaveDoubleCautationClick(Sender: TObject);
    procedure MenuSaveCommaClick(Sender: TObject);
    procedure MenuCsvAutoClick(Sender: TObject);
    procedure MenuCsvEx1Click(Sender: TObject);
    procedure MenuSameAsAnnsViewClick(Sender: TObject);
    procedure MenuClipTabClick(Sender: TObject);
    procedure MenuClipDoubleCautationClick(Sender: TObject);
    procedure MenuClipCommaClick(Sender: TObject);
    procedure MenuCrLfColorClick(Sender: TObject);

    procedure SetSameAsAnnsView(Sender: TObject);

    procedure MemoStatus(Sender: TObject);
    procedure Memo1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Memo1KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Memo1Click(Sender: TObject);
    procedure Memo1MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure MenuSaveFixedLengthClick(Sender: TObject);
    procedure MenuCsvEx2Click(Sender: TObject);
    procedure MenuSaveCsvEx2Click(Sender: TObject);
    procedure MenuClipCsvEx2Click(Sender: TObject);

  private
    { Private 錾 }
    FormPosLeft: Integer;
    FormPosTop: Integer;
    FormPosWidth: Integer;
    FormPosHeight: Integer;
    FormPosRight: Integer;
    FormPosButtom: Integer;
    MonitorTop: Integer;
    MonitorLeft: Integer;
    MonitorRight: Integer;
    MonitorButtom: Integer;
    MonitorWidth: Integer;
    MonitorHeight: Integer;
    StrFormState: String;

    StrLanguageFileName: String;
    StrLoadFileName: String;
    StrLoadChar: String;
    StrLoadReturn: String;
    StrSeparater: String;
    StrSaveChar: String;
    BoolJapanese: Boolean;
    BoolLfInLine: Boolean;
    BoolUnKnownCharCode: Boolean;
    BoolFileReadOnly: Boolean;
    BoolEditModified: Boolean;
    BoolKeyRun: Boolean;
    IntLoadFileSize: Integer;
    IntReadEncode1: Integer;
    IntReadReturn1: Integer;
    IntTextHeight: Integer;
    IntLastRow: Integer;
    IntLastCol: Integer;
    IntMemoHeightIndex: Integer;
    IntGridRatioIndex: Integer;
    IntStrToCsvIndex: Integer;

    StrListLoadFile: TStringList;
    StrListLoadCsvFile: TStringList;
    StrListSaveFile: TStringList;
    StrListKeyMap: TStringList;
    StrListCopyRow: TStringList;
    StrListCopyCol: TStringList;
    CsvList11: TStringList;

  public
    { Public 錾 }

  end;

var
  FormViewGrid: TFormViewGrid;

implementation

uses
  Main;

{$R *.dfm}

//NbNɃtH[ANeBuɂ邽߂ɕKv
procedure TFormViewGrid.CreateParams(var Params: TCreateParams);
begin
  inherited;
  Params.WndParent := Application.Handle;
end;

procedure TFormViewGrid.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  //IMEIt
  SetImeMode(FormViewGrid.Handle, imClose);
end;

procedure TFormViewGrid.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  wRet8: Word;
  sStr8: String;
begin
  //ҏWĂ
  if (BoolEditModified = TRUE) or (Memo1.Modified = TRUE) then
  begin
    //mF
    if BoolJapanese = TRUE then
    begin
      sStr8 := 'Information : ҏWĂ܂' + #13#10 +
               'I܂H';
    end
    else
    begin
      sStr8 := 'Information : Item is modified.' + #13#10 +
               'Exit ?' + #13#10 +
               '[YES] / [NO]';
    end;
    wRet8 := MessageDlg(sStr8, mtInformation, [mbYes, mbNo ], 0);
    if wRet8 = mrYes then
    begin
      //I
    end
    else
    begin
      //~
      CanClose := FALSE;
    end;
  end;

end;

procedure TFormViewGrid.FormCreate(Sender: TObject);
var
  IniFile: TIniFile;
  sStr3: String;
  iWidth1: Integer;
begin
  //MessageDialog ̉΍
  UseLatestCommonDialogs := FALSE;

  //m
  IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
  //tH[ʒu
  FormPosLeft   := IniFile.ReadInteger('FormViewGrid', 'Left',   0);
  FormPosTop    := IniFile.ReadInteger('FormViewGrid', 'Top',    0);
  FormPosWidth  := IniFile.ReadInteger('FormViewGrid', 'Width',  FORM_WIDTH_SET);
  FormPosHeight := IniFile.ReadInteger('FormViewGrid', 'Height', FORM_HEIGHT_SET);
  StrFormState  := IniFile.ReadString ('FormViewGrid', 'State',  FORM_NORMAL);
  FormPosRight  := FormPosLeft + FormPosWidth;
  FormPosButtom := FormPosTop  + FormPosHeight;
  //j
  IniFile.Free;

  //Zbg
  Timer4.Enabled := FALSE;
  Timer5.Enabled := FALSE;

  //m
  IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
  //
  Memo1.Height        := IniFile.ReadInteger('ViewGrid', 'MemoHeight', 150);
  Panel3.Height       := IniFile.ReadInteger('ViewGrid', 'MemoHeight', 150);
  //u
  MenuLfToTab.Checked := IniFile.ReadBool('ViewGrid', 'LfToTab', FALSE);
  //s
  MenuStatusRowCol.Checked := IniFile.ReadBool('ViewGrid', 'RowCol', FALSE);
  MenuStatusColRow.Checked := IniFile.ReadBool('ViewGrid', 'ColRow', TRUE);
  //R[h
  MenuReadAuto.Checked          :=          IniFile.ReadBool  ('ViewGrid', 'ReadAuto',       TRUE);
  MenuReadSJIS.Checked          :=          IniFile.ReadBool  ('ViewGrid', 'ReadSJIS',       FALSE);
  MenuReadJIS.Checked           :=          IniFile.ReadBool  ('ViewGrid', 'ReadJIS',        FALSE);
  MenuReadEUCJP.Checked         :=          IniFile.ReadBool  ('ViewGrid', 'ReadEUCJP',      FALSE);
  MenuReadUTF7.Checked          :=          IniFile.ReadBool  ('ViewGrid', 'ReadUTF7',       FALSE);
  MenuReadUTF8Bom.Checked       :=          IniFile.ReadBool  ('ViewGrid', 'ReadUTF8Bom',    FALSE);
  MenuReadUTF8.Checked          :=          IniFile.ReadBool  ('ViewGrid', 'ReadUTF8',       FALSE);
  MenuReadUTF16BomBe.Checked    :=          IniFile.ReadBool  ('ViewGrid', 'ReadUTF16BomBe', FALSE);
  MenuReadUTF16Be.Checked       :=          IniFile.ReadBool  ('ViewGrid', 'ReadUTF16Be',    FALSE);
  MenuReadUTF16BomLe.Checked    :=          IniFile.ReadBool  ('ViewGrid', 'ReadUTF16BomLe', FALSE);
  MenuReadUTF16Le.Checked       :=          IniFile.ReadBool  ('ViewGrid', 'ReadUTF16Le',    FALSE);
  MenuSaveSJIS.Checked          :=          IniFile.ReadBool  ('ViewGrid', 'SaveSJIS',       TRUE);
  MenuSaveJIS.Checked           :=          IniFile.ReadBool  ('ViewGrid', 'SaveJIS',        FALSE);
  MenuSaveEUCJP.Checked         :=          IniFile.ReadBool  ('ViewGrid', 'SaveEUCJP',      FALSE);
  MenuSaveUTF7.Checked          :=          IniFile.ReadBool  ('ViewGrid', 'SaveUTF7',       FALSE);
  MenuSaveUTF8Bom.Checked       :=          IniFile.ReadBool  ('ViewGrid', 'SaveUTF8Bom',    FALSE);
  MenuSaveUTF8.Checked          :=          IniFile.ReadBool  ('ViewGrid', 'SaveUTF8',       FALSE);
  MenuSaveUTF16BomBe.Checked    :=          IniFile.ReadBool  ('ViewGrid', 'SaveUTF16BomBe', FALSE);
  MenuSaveUTF16Be.Checked       :=          IniFile.ReadBool  ('ViewGrid', 'SaveUTF16Be',    FALSE);
  MenuSaveUTF16BomLe.Checked    :=          IniFile.ReadBool  ('ViewGrid', 'SaveUTF16BomLe', FALSE);
  MenuSaveUTF16Le.Checked       :=          IniFile.ReadBool  ('ViewGrid', 'SaveUTF16Le',    FALSE);
  MenuSaveTab.Checked              :=       IniFile.ReadBool  ('ViewGrid', 'SaveTAB',              TRUE);
  MenuSaveDoubleCautation.Checked  :=       IniFile.ReadBool  ('ViewGrid', 'SaveDoubleCautation',  FALSE);
  MenuSaveComma.Checked            :=       IniFile.ReadBool  ('ViewGrid', 'SaveComma',            FALSE);
  MenuSaveFixedLength.Checked      :=       IniFile.ReadBool  ('ViewGrid', 'SaveFixedLength',      FALSE);
  MenuSaveCsvEx2.Checked           :=       IniFile.ReadBool  ('ViewGrid', 'SaveCsvEx2',           FALSE);
  MenuClipTab.Checked              :=       IniFile.ReadBool  ('ViewGrid', 'ClipTab',              TRUE);
  MenuClipDoubleCautation.Checked  :=       IniFile.ReadBool  ('ViewGrid', 'ClipDoubleCautation',  FALSE);
  MenuClipComma.Checked            :=       IniFile.ReadBool  ('ViewGrid', 'ClipComma',            FALSE);
  MenuClipCsvEx2.Checked           :=       IniFile.ReadBool  ('ViewGrid', 'ClipCsvEx2',           FALSE);
  //F
  FontDialog1.Font.Color        := StringToColor(IniFile.ReadString ('ViewGlid', 'FontColor', ColorToString(clBlack)));
  FontDialog1.Font.Name         :=               IniFile.ReadString ('ViewGlid', 'FontName',   'lr SVbN');
  FontDialog1.Font.Size         :=      StrToInt(IniFile.ReadString ('ViewGlid', 'FontSize',   '14'));
  ColorDialog1.Color            := StringToColor(IniFile.ReadString ('ViewGlid', 'Color',        ColorToString(clWhite)));
  ColorDialog2.Color            := StringToColor(IniFile.ReadString ('ViewGlid', 'FixedColor',   ColorToString(clSkyBlue)));
  ColorDialog3.Color            := StringToColor(IniFile.ReadString ('ViewGlid', 'CurrentColor', ColorToString($00FF8000)));
  ColorDialog4.Color            := StringToColor(IniFile.ReadString ('ViewGlid', 'CrLfColor',    ColorToString(clBlue)));
  IntMemoHeightIndex            :=               IniFile.ReadInteger('ViewGlid', 'MemoHeightIndex', 1);
  MenuSameAsAnnsView.Checked    :=               IniFile.ReadBool   ('ViewGlid', 'SameAsAnnsView',  FALSE);
  //
  Edit1.Text  := IniFile.ReadString('ViewGlid', 'SearchWord', '');
  //j
  IniFile.Free;

  //
  StringGrid1.DrawingStyle := gdsClassic;
//  StringGrid1.DoubleBuffered := TRUE;
  StringGrid1.Options := StringGrid1.Options + [goColSizing] + [goThumbTracking];
  //DrawingStyle := gdsClassic Kv
  StringGrid1.Canvas.Font.Name  := FontDialog1.Font.Name;
  StringGrid1.Canvas.Font.Color := FontDialog1.Font.Color;
  StringGrid1.Canvas.Font.Size  := FontDialog1.Font.Size;
  StringGrid1.Font.Name  := FontDialog1.Font.Name;
  StringGrid1.Font.Color := FontDialog1.Font.Color;
  StringGrid1.Font.Size  := FontDialog1.Font.Size;
  StringGrid1.Color      := ColorDialog1.Color;
  StringGrid1.FixedColor := ColorDialog2.Color;
  StringGrid1.CurrentBackColor := ColorDialog3.Color;
  StringGrid1.FixedBackColor := ColorDialog2.Color;
  //ݒ
  IntTextHeight := TEXT_HEIGHT * StringGrid1.Canvas.Font.Size div TEXT_SIZE;
  //
  StringGrid1.DefaultRowHeight := IntTextHeight;
  StringGrid1.DefaultColWidth  := GRID_COL_WIDTH;
  StringGrid1.GridLineWidth    := GRID_LINE_WIDTH;

  //
  if FormMain.IntVclStyle = 0 then
  begin
    StringGrid1.DarkMode := FALSE;
    Memo1.DarkMode := FALSE;
  end
  else
  begin
    StringGrid1.DarkMode := TRUE;
    Memo1.DarkMode := TRUE;
  end;
    //Fݒ
    Memo1.DarkBackColor := FormMain.Color;


  //
  Memo1.Font.Name  := FontDialog1.Font.Name;
  Memo1.Font.Color := FontDialog1.Font.Color;
  Memo1.Font.Size  := FontDialog1.Font.Size;
  Memo1.Color      := ColorDialog1.Color;
  Memo1.CrLfColor  := ColorDialog4.Color;
  Memo1.ReadOnly := TRUE;

  //mF
  if FormMain.StrAnnsFMPath = '' then
  begin
    sStr3 := ExtractFilePath(ParamStr(0)) + ANNSFM_INI;
  end
  else
  begin
    sStr3 := FormMain.StrAnnsFMPath + ANNSFM_INI;
    if FileExists(sStr3) = FALSE then
    begin
      sStr3 := ExtractFilePath(ParamStr(0)) + ANNSFM_INI;
    end;
  end;
  if FileExists(sStr3) = TRUE then
  begin
    //
  end
  else
  begin
    MenuSameAsAnnsView.Checked := FALSE;
  end;
  //ݒ
  if MenuSameAsAnnsView.Checked = TRUE then
  begin
    //ݒ
    MenuFontColor.Enabled := FALSE;
    MenuBackColor.Enabled := FALSE;
    MenuFontColoInit.Enabled := FALSE;
  end
  else
  begin
    //ݒ
    MenuFontColor.Enabled := TRUE;
    MenuBackColor.Enabled := TRUE;
    MenuFontColoInit.Enabled := TRUE;
  end;
  //ݒ
  SetSameAsAnnsView(Sender);

  //
  StatusBar1.Panels.Items[0].Width :=  80;
  StatusBar1.Panels.Items[1].Width :=  90;
  StatusBar1.Panels.Items[2].Width := 120;
  StatusBar1.Panels.Items[3].Width :=  90;
  StatusBar1.Panels.Items[4].Width := 210;
  StatusBar1.Panels.Items[5].Width := 120;
  StatusBar1.Panels.Items[6].Width :=  20;
  StatusBar2.Panels.Items[0].Width :=  80;
  StatusBar2.Panels.Items[1].Width :=  90;
  StatusBar2.Panels.Items[2].Width := 120;
  StatusBar2.Panels.Items[3].Width :=  90;
  StatusBar2.Panels.Items[4].Width := 210;
  StatusBar2.Panels.Items[5].Width := 120;
  StatusBar2.Panels.Items[6].Width :=  20;
  iWidth1 := StatusBar3.ClientWidth div 6;
  StatusBar3.Panels.Items[0].Width := iWidth1;
  StatusBar3.Panels.Items[1].Width := iWidth1;
  StatusBar3.Panels.Items[2].Width := iWidth1;
  StatusBar3.Panels.Items[3].Width := iWidth1;
  StatusBar3.Panels.Items[4].Width := iWidth1;
  StatusBar3.Panels.Items[5].Width := iWidth1;

  //
  IntLastCol := 0;
  IntLastRow := 0;

  //ݒ
  MenuCsvAuto.Checked := TRUE;
  MenuCsvEx1.Checked  := FALSE;
  MenuCsvEx2.Checked  := FALSE;
  IntStrToCsvIndex := 0;

  //
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;
  if IntMemoHeightIndex = 1 then
  begin
    MenuMemoHeight1.Checked  := TRUE;
    MenuMemoHeight1Click(Sender);
  end;
  if IntMemoHeightIndex = 2 then
  begin
    MenuMemoHeight2.Checked  := TRUE;
    MenuMemoHeight2Click(Sender);
  end;
  if IntMemoHeightIndex = 3 then
  begin
    MenuMemoHeight3.Checked  := TRUE;
    MenuMemoHeight3Click(Sender);
  end;
  if IntMemoHeightIndex = 4 then
  begin
    MenuMemoHeight4.Checked  := TRUE;
    MenuMemoHeight4Click(Sender);
  end;
  if IntMemoHeightIndex = 5 then
  begin
    MenuMemoHeight5.Checked  := TRUE;
    MenuMemoHeight5Click(Sender);
  end;
  if IntMemoHeightIndex = 6 then
  begin
    MenuMemoHeight6.Checked  := TRUE;
    MenuMemoHeight6Click(Sender);
  end;
  if IntMemoHeightIndex = 7 then
  begin
    MenuMemoHeight7.Checked  := TRUE;
    MenuMemoHeight7Click(Sender);
  end;
  if IntMemoHeightIndex = 8 then
  begin
    MenuMemoHeight8.Checked  := TRUE;
    MenuMemoHeight8Click(Sender);
  end;
  if IntMemoHeightIndex = 9 then
  begin
    MenuMemoHeight9.Checked  := TRUE;
    MenuMemoHeight9Click(Sender);
  end;
  if IntMemoHeightIndex = 10 then
  begin
    MenuMemoHeight10.Checked  := TRUE;
    MenuMemoHeight10Click(Sender);
  end;

  //ݒ
  MenuRatio1Click(Sender);

  //ݒ
  BoolJapanese := FormMain.BoolJapanese;
  BoolEditModified := FALSE;
  BoolKeyRun := FALSE;

  //m
  StrListLoadFile := TStringList.Create;
  StrListLoadCsvFile := TStringList.Create;
  StrListSaveFile := TStringList.Create;
  StrListKeyMap := TStringList.Create;
  StrListCopyRow := TStringList.Create;
  StrListCopyCol := TStringList.Create;
  CsvList11 := TStringList.Create;

  //ݒ
  SetShortCutKey(Sender);
end;

procedure TFormViewGrid.FormDestroy(Sender: TObject);
var
  IniFile: TIniFile;
begin
  //m
  IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
  //ʏ
  if WindowState = wsMaximized then
  begin
    IniFile.WriteString('FormViewGrid', 'State', FORM_MAX);
  end
  else if WindowState = wsMinimized then
  begin
    IniFile.WriteString('FormViewGrid', 'State', FORM_MIN);
  end
  else
  begin
    IniFile.WriteString ('FormViewGrid', 'State',  FORM_NORMAL);
    IniFile.WriteInteger('FormViewGrid', 'Left',   FormViewGrid.Left);
    IniFile.WriteInteger('FormViewGrid', 'Top',    FormViewGrid.Top);
    IniFile.WriteInteger('FormViewGrid', 'Width',  FormViewGrid.Width);
    IniFile.WriteInteger('FormViewGrid', 'Height', FormViewGrid.Height);
  end;
  //j
  IniFile.Free;

  //m
  IniFile := TIniFile.Create (ChangeFileExt(ParamStr(0), '.ini'));
  //
  IniFile.WriteInteger('ViewGrid', 'MemoHeight', Memo1.Height);
  //u
  IniFile.WriteBool   ('ViewGrid', 'LfToTab', MenuLfToTab.Checked);
  //s
  IniFile.WriteBool('ViewGrid', 'RowCol', MenuStatusRowCol.Checked);
  IniFile.WriteBool('ViewGrid', 'ColRow', MenuStatusColRow.Checked);
  //R[h
  IniFile.WriteBool   ('ViewGrid', 'ReadAuto',             MenuReadAuto.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadSJIS',             MenuReadSJIS.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadJIS',              MenuReadJIS.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadEUCJP',            MenuReadEUCJP.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadUTF7',             MenuReadUTF7.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadUTF8Bom',          MenuReadUTF8Bom.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadUTF8',             MenuReadUTF8.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadUTF16BomBe',       MenuReadUTF16BomBe.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadUTF16Be',          MenuReadUTF16Be.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadUTF16BomLe',       MenuReadUTF16BomLe.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ReadUTF16Le',          MenuReadUTF16Le.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveSJIS',             MenuSaveSJIS.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveJIS',              MenuSaveJIS.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveEUCJP',            MenuSaveEUCJP.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveUTF7',             MenuSaveUTF7.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveUTF8Bom',          MenuSaveUTF8Bom.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveUTF8',             MenuSaveUTF8.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveUTF16BomBe',       MenuSaveUTF16BomBe.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveUTF16Be',          MenuSaveUTF16Be.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveUTF16BomLe',       MenuSaveUTF16BomLe.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveUTF16Le',          MenuSaveUTF16Le.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveTAB',              MenuSaveTab.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveDoubleCautation',  MenuSaveDoubleCautation.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveComma',            MenuSaveComma.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveFixedLength',      MenuSaveFixedLength.Checked);
  IniFile.WriteBool   ('ViewGrid', 'SaveCsvEx2',           MenuSaveCsvEx2.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ClipTAB',              MenuClipTab.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ClipDoubleCautation',  MenuClipDoubleCautation.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ClipComma',            MenuClipComma.Checked);
  IniFile.WriteBool   ('ViewGrid', 'ClipCsvEx2',           MenuClipCsvEx2.Checked);
  IniFile.WriteString ('ViewGlid', 'FontColor',    ColorToString(FontDialog1.Font.Color));
  IniFile.WriteString ('ViewGlid', 'FontName',     FontDialog1.Font.Name);
  IniFile.WriteString ('ViewGlid', 'FontSize',     IntToStr(FontDialog1.Font.Size));
  IniFile.WriteString ('ViewGlid', 'Color',        ColorToString(ColorDialog1.Color));
  IniFile.WriteString ('ViewGlid', 'FixedColor',   ColorToString(ColorDialog2.Color));
  IniFile.WriteString ('ViewGlid', 'CurrentColor', ColorToString(ColorDialog3.Color));
  IniFile.WriteString ('ViewGlid', 'CrLfColor',    ColorToString(ColorDialog4.Color));
  IniFile.WriteInteger('ViewGlid', 'MemoHeightIndex', IntMemoHeightIndex);
  IniFile.WriteBool   ('ViewGlid', 'SameAsAnnsView',  MenuSameAsAnnsView.Checked);
  //
  IniFile.WriteString('ViewGlid', 'SearchWord', Edit1.Text);
  //j
  IniFile.Free;

  //j
  StrListLoadFile.Free;
  StrListLoadCsvFile.Free;
  StrListSaveFile.Free;
  StrListKeyMap.Free;
  StrListCopyRow.Free;
  StrListCopyCol.Free;
  CsvList11.Free;
end;

procedure TFormViewGrid.FormResize(Sender: TObject);
var
  iWidth1: Integer;
begin
  //ݒ
  iWidth1 := StatusBar3.ClientWidth div 6;
  StatusBar3.Panels.Items[0].Width := iWidth1;
  StatusBar3.Panels.Items[1].Width := iWidth1;
  StatusBar3.Panels.Items[2].Width := iWidth1;
  StatusBar3.Panels.Items[3].Width := iWidth1;
  StatusBar3.Panels.Items[4].Width := iWidth1;
  StatusBar3.Panels.Items[5].Width := iWidth1;
end;

procedure TFormViewGrid.FormShow(Sender: TObject);
begin
  //FormłȂ̂ŁACreatełȂShowōs
  //FormKeyDowñCxgݒ肷
  FormViewGrid.KeyPreview := TRUE;

  //GetSystemMetricst֐gpāCzʑŜ̃TCY擾
  MonitorTop    := GetSystemMetrics(SM_YVIRTUALSCREEN);
  MonitorLeft   := GetSystemMetrics(SM_XVIRTUALSCREEN);
  MonitorWidth  := GetSystemMetrics(SM_CXVIRTUALSCREEN);
  MonitorHeight := GetSystemMetrics(SM_CYVIRTUALSCREEN);
  MonitorRight  := MonitorLeft + MonitorWidth;
  MonitorButtom := MonitorTop  + MonitorHeight;

  //\ł ȌԂŕ\
  if ( (MonitorTop  <= FormPosTop  + FORM_TOP_SHIFT ) and (FormPosTop    <= MonitorButtom) ) and
     ( (MonitorTop  <= FormPosButtom                ) and (FormPosButtom <= MonitorButtom) ) and
     ( (MonitorLeft <= FormPosLeft + FORM_LEFT_SHIFT) and (FormPosLeft   <= MonitorRight ) ) and
     ( (MonitorLeft <= FormPosRight                 ) and (FormPosRight  <= MonitorRight ) ) then
  begin
    FormViewGrid.SetBounds(FormPosLeft, FormPosTop, FormPosWidth, FormPosHeight);
  end
  //ʓɕ\łȂ ftHgŕ\
  else
  begin
    FormViewGrid.SetBounds(FormPosLeft, FormPosTop, FormPosWidth, FormPosHeight);
  end;

  //ő
  if StrFormState = FORM_MAX then
  begin
    FormViewGrid.WindowState := wsMaximized;
  end
  //ŏ
  else if StrFormState = FORM_MIN then
  begin
    FormViewGrid.WindowState := wsMinimized;
  end
  //W
  else
  begin
    //
  end;

  //IMEIt
  SetImeMode(FormViewGrid.Handle, imClose);

  //ݒ
  StrLoadFileName := FormMain.StrSetFormViewGrid;

  //ݒ
  BoolJapanese := FormMain.BoolJapanese;
  //ݒ
  StrLanguageFileName := ExtractFilePath(ParamStr(0)) + FOLDER_LANGUAGE  + '\' +
                         ChangeFileExt(ExtractFileName(ParamStr(0)), '') + '\' + 'ViewGrid.txt';
  if BoolJapanese = TRUE then
  begin
    //{
    FormMain.SetJapanese(FormViewGrid, StrLanguageFileName, Sender);
  end
  else
  begin
    //p
    FormMain.SetEnglish(FormViewGrid, StrLanguageFileName, Sender);
  end;

  //FormCreate ł͍sƃG[ɂȂ
  if BoolJapanese = TRUE then
  begin
    FormViewGrid.Caption := StrLoadFileName + TITLE_VIEW;
  end
  else
  begin
    FormViewGrid.Caption := StrLoadFileName + TITLE_VIEW_ENG;
  end;

  //t@CǍ
  LoadCsvFile(Sender);
  SetCsvFile(Sender);
end;

procedure TFormViewGrid.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  //ESC
  if      ((Shift = []) and (key = VK_ESCAPE)) and (MenuEditMode.Checked = FALSE) then
  begin
    //s
    MenuEscapeClick(Sender);
  end
  //̃Z
  else if ((Shift = [ssCtrl]) and (key = VK_UP)) and (Memo1.Focused = TRUE) then
  begin
    if StringGrid1.Row > 2 then
    begin
      StringGrid1.Row := StringGrid1.Row-1;
    end;
  end
  else if ((Shift = [ssCtrl]) and (key = VK_DOWN)) and (Memo1.Focused = TRUE) then
  begin
    if StringGrid1.Row < (StrInggrid1.RowCount-1) then
    begin
      StringGrid1.Row := StringGrid1.Row+1;
    end;
  end
  else if ((Shift = [ssCtrl]) and (key = VK_LEFT)) and (Memo1.Focused = TRUE) then
  begin
    if StringGrid1.Col > 2 then
    begin
      StringGrid1.Col := StringGrid1.Col-1;
    end;
  end
  else if ((Shift = [ssCtrl]) and (key = VK_RIGHT)) and (Memo1.Focused = TRUE) then
  begin
    if StringGrid1.Col < (StrInggrid1.ColCount-1) then
    begin
      StringGrid1.Col := StringGrid1.Col+1;
    end;
  end
  //{
  else if ((Shift = [ssShift, ssCtrl]) and (key = VK_J)) and (FormMain.BoolDeveloper = TRUE) then
  begin
    //ݒ
    BoolJapanese := TRUE;
    //{
    FormMain.SetJapanese(FormViewGrid, StrLanguageFileName, Sender);
  end
  //p
  else if ((Shift = [ssShift, ssCtrl]) and (key = VK_E)) and (FormMain.BoolDeveloper = TRUE) then
  begin
    //ݒ
    BoolJapanese := FALSE;
    //p
    FormMain.SetEnglish(FormViewGrid, StrLanguageFileName, Sender);
  end
  //
  else
  begin
    //
  end;
end;

procedure TFormViewGrid.WmGetMinMaxInfo(var msg : TWMGetMinMaxInfo);
begin
  //ŏTCY
  with msg.MinMaxInfo^.ptMinTrackSize do
  begin
    x := FORM_WIDTH_SET;
    y := FORM_HEIGHT_SET;
  end;
  //őTCY
  with msg.MinMaxInfo^.ptMaxTrackSize do
  begin
    //
  end;
  //ő剻TCY
  with msg.MinMaxInfo^.ptMaxSize do
  begin
    //
  end;
end;

procedure TFormViewGrid.MenuRatio1Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := TRUE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 1;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio2Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := TRUE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 2;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio3Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := TRUE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 3;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio4Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := TRUE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 4;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio5Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := TRUE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 5;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio6Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := TRUE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 6;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio7Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := TRUE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 7;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio8Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := TRUE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 8;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio9Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := TRUE;
  MenuRatio10.Checked := FALSE;
  IntGridRatioIndex   := 9;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuRatio10Click(Sender: TObject);
var
  iRow1: Integer;
begin
  //ݒ
  MenuRatio1.Checked  := FALSE;
  MenuRatio2.Checked  := FALSE;
  MenuRatio3.Checked  := FALSE;
  MenuRatio4.Checked  := FALSE;
  MenuRatio5.Checked  := FALSE;
  MenuRatio6.Checked  := FALSE;
  MenuRatio7.Checked  := FALSE;
  MenuRatio8.Checked  := FALSE;
  MenuRatio9.Checked  := FALSE;
  MenuRatio10.Checked := TRUE;
  IntGridRatioIndex   := 10;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
  end;
end;

procedure TFormViewGrid.MenuReadAutoClick(Sender: TObject);
begin
  if MenuReadAuto.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := TRUE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadSJISClick(Sender: TObject);
begin
  if MenuReadSJIS.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := TRUE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadJISClick(Sender: TObject);
begin
  if MenuReadJIS.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := TRUE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadEUCJPClick(Sender: TObject);
begin
  if MenuReadEUCJP.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := TRUE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadUTF7Click(Sender: TObject);
begin
  if MenuReadUTF7.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := TRUE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadUTF8BomClick(Sender: TObject);
begin
  if MenuReadUTF8Bom.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := TRUE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadUTF8Click(Sender: TObject);
begin
  if MenuReadUTF8.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := TRUE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadUTF16BomBeClick(Sender: TObject);
begin
  if MenuReadUTF16BomBe.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := TRUE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadUTF16BeClick(Sender: TObject);
begin
  if MenuReadUTF16Be.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := TRUE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadUTF16BomLeClick(Sender: TObject);
begin
  if MenuReadUTF16BomLe.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := TRUE;
    MenuReadUTF16Le.Checked    := FALSE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuReadUTF16LeClick(Sender: TObject);
begin
  if MenuReadUTF16Le.Checked = FALSE then
  begin
    //ݒ
    MenuReadAuto.Checked       := FALSE;
    MenuReadSJIS.Checked       := FALSE;
    MenuReadJIS.Checked        := FALSE;
    MenuReadEUCJP.Checked      := FALSE;
    MenuReadUTF7.Checked       := FALSE;
    MenuReadUTF8Bom.Checked    := FALSE;
    MenuReadUTF8.Checked       := FALSE;
    MenuReadUTF16BomBe.Checked := FALSE;
    MenuReadUTF16Be.Checked    := FALSE;
    MenuReadUTF16BomLe.Checked := FALSE;
    MenuReadUTF16Le.Checked    := TRUE;

    //Ǎ
    if StrLoadFileName <> '' then
    begin
      //Ǎ
      LoadCsvFile(Sender);
      SetCsvFile(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuDeleteRowClick(Sender: TObject);
var
  iRow1: Integer;
  iTopRow1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //
  if MenuEditMode.Checked = FALSE then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : ҏW[hł͂܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Status is not Edit Mode.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //
  if StringGrid1.Row <= 1 then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : wb_[s͍폜ł܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Header row can not delete.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //
  if StringGrid1.RowCount <= 2 then
  begin
    BoolKeyRun := FALSE;
    Exit;
  end;

  //L
  iTopRow1 := StringGrid1.TopRow;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //ŏI
  if StringGrid1.RowCount = 3 then
  begin
    //NA
    StringGrid1.Rows[Stringgrid1.Row].Text := '';
    //ݒ
    StringGrid1.TopRow := iTopRow1;
  end
  //ŏIȊO
  else
  begin
    //NA
    StringGrid1.Rows[Stringgrid1.Row].Text := '';
    //V@\
    StringGrid1.DeleteRow(Stringgrid1.Row);
    //ݒ
    StringGrid1.TopRow := iTopRow1;
  end;

  //Đݒ
  for iRow1 := 0 to StringGrid1.RowCount-2 do
  begin
    //s
    StringGrid1.Cells[0, iRow1+2] := Format('%.05d', [iRow1+1]);
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuInsertRowClick(Sender: TObject);
var
  iRow1: Integer;
  iTopRow1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //
  if MenuEditMode.Checked = FALSE then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : ҏW[hł͂܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Status is not Edit Mode.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  if StringGrid1.Row < 2 then
  begin
    BoolKeyRun := FALSE;
    Exit;
  end;

  //L
  iTopRow1 := StringGrid1.TopRow;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //s𑝂₷
  StringGrid1.RowCount := StringGrid1.RowCount+1;
  //
  if StringGrid1.Row < StringGrid1.RowCount-1 then
  begin
    //Rs[
    for iRow1 := StringGrid1.RowCount-1 downto StringGrid1.Row+1 do
    begin
      StringGrid1.Rows[iRow1].Assign(StringGrid1.Rows[iRow1-1]);
    end;
    //ݍs̓NA
    StringGrid1.Rows[StringGrid1.Row].Clear;
    //ݒ
    StringGrid1.TopRow := iTopRow1;
  end;

  //Đݒ
  for iRow1 := 0 to StringGrid1.RowCount-2 do
  begin
    //s
    StringGrid1.Cells[0, iRow1+2] := Format('%.05d', [iRow1+1]);
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuAddRowClick(Sender: TObject);
var
  iRow1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //
  if MenuEditMode.Checked = FALSE then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : ҏW[hł͂܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Status is not Edit Mode.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //s𑝂₷
  StringGrid1.RowCount := StringGrid1.RowCount+1;
  //ݍs̓NA
  StringGrid1.Rows[StringGrid1.RowCount-1].Clear;
  //ݒ
  StringGrid1.TopRow := StringGrid1.RowCount-1;
  StringGrid1.Row    := StringGrid1.RowCount-1;

  //Đݒ
  for iRow1 := 0 to StringGrid1.RowCount-2 do
  begin
    //s
    StringGrid1.Cells[0, iRow1+2] := Format('%.05d', [iRow1+1]);
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuCopyRowClick(Sender: TObject);
var
  iCol1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //NA
  StrListCopyRow.Clear;
  //ǉ
  for iCol1 := 2 to StringGrid1.ColCount-1 do
  begin
    StrListCopyRow.Add(StringGrid1.Cells[iCol1, StringGrid1.Row]);
  end;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuPasteRowClick(Sender: TObject);
var
  iCol1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //
  if MenuEditMode.Checked = FALSE then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : ҏW[hł͂܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Status is not Edit Mode.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //
  if StrListCopyRow.Count = 0 then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : Rs[f[^܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Copy data is empty.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //s
  for iCol1 := 2 to StringGrid1.ColCount-1 do
  begin
    StringGrid1.Cells[iCol1, StringGrid1.Row] := StrListCopyRow.Strings[iCol1-2];
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuDeleteColClick(Sender: TObject);
var
  iCol1: Integer;
  iLeftCol1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //
  if MenuEditMode.Checked = FALSE then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : ҏW[hł͂܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Status is not Edit Mode.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //
  if StringGrid1.Col <= 1 then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : wb_[͍폜ł܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Header column can not delete.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //
  if StringGrid1.ColCount <= 2 then
  begin
    BoolKeyRun := FALSE;
    Exit;
  end;


  //L
  iLeftCol1 := StringGrid1.LeftCol;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //ŏI
  if StringGrid1.ColCount = 3 then
  begin
    //NA
    StringGrid1.Cols[Stringgrid1.Col].Text := '';
    //ݒ
    StringGrid1.LeftCol := iLeftCol1;
  end
  //ŏIȊO
  else
  begin
    //NA
    StringGrid1.Cols[Stringgrid1.Col].Text := '';
    //V@\
    StringGrid1.DeleteColumn(Stringgrid1.Col);
    //ݒ
    StringGrid1.LeftCol := iLeftCol1;
  end;

  //Đݒ
  for iCol1 := 0 to StringGrid1.ColCount-2 do
  begin
    //
    StringGrid1.Cells[iCol1+2, 0] := InttoStr(iCol1+1);
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuCopyColClick(Sender: TObject);
var
  iRow1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //NA
  StrListCopyCol.Clear;
  //ǉ
  for iRow1 := 2 to StringGrid1.RowCount-1 do
  begin
    StrListCopyCol.Add(StringGrid1.Cells[StringGrid1.Col, iRow1]);
  end;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuPasteColClick(Sender: TObject);
var
  iRow1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //
  if MenuEditMode.Checked = FALSE then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : ҏW[hł͂܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Status is not Edit Mode.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //
  if StrListCopyCol.Count = 0 then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : Rs[f[^܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Copy data is empty.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //s
  for iRow1 := 2 to StringGrid1.ColCount-1 do
  begin
    StringGrid1.Cells[StringGrid1.Col, iRow1] := StrListCopyCol.Strings[iRow1-2];
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuInsertColClick(Sender: TObject);
var
  iCol1: Integer;
  iLeftCol1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //
  if MenuEditMode.Checked = FALSE then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : ҏW[hł͂܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Status is not Edit Mode.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  if StringGrid1.Col < 2 then
  begin
    BoolKeyRun := FALSE;
    Exit;
  end;

  //L
  iLeftCol1 := StringGrid1.LeftCol;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //𑝂₷
  StringGrid1.ColCount := StringGrid1.ColCount+1;
  //
  if StringGrid1.Col < StringGrid1.ColCount-1 then
  begin
    //Rs[
    for iCol1 := StringGrid1.ColCount-1 downto StringGrid1.Col+1 do
    begin
      StringGrid1.Cols[iCol1].Assign(StringGrid1.Cols[iCol1-1]);
    end;
    //ݍs̓NA
    StringGrid1.Cols[StringGrid1.Col].Clear;
    //ݒ
    StringGrid1.LeftCol := iLeftCol1;
  end;

  //Đݒ
  for iCol1 := 0 to StringGrid1.ColCount-2 do
  begin
    //
    StringGrid1.Cells[iCol1+2, 0] := InttoStr(iCol1+1);
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuAddColClick(Sender: TObject);
var
  iCol1: Integer;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //
  if MenuEditMode.Checked = FALSE then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : ҏW[hł͂܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : Status is not Edit Mode.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    BoolKeyRun := FALSE;
    Exit;
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //s𑝂₷
  StringGrid1.ColCount := StringGrid1.ColCount+1;
  //ݍs̓NA
  StringGrid1.Cols[StringGrid1.ColCount-1].Clear;
  //ݒ
  StringGrid1.LeftCol := StringGrid1.ColCount-1;
  StringGrid1.Col     := StringGrid1.ColCount-1;

  //Đݒ
  for iCol1 := 0 to StringGrid1.ColCount-2 do
  begin
    //
    StringGrid1.Cells[iCol1+2, 0] := InttoStr(iCol1+1);
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuColFixedClick(Sender: TObject);
var
  iRow1: Integer;
begin
  for iRow1 := 2 to StringGrid1.RowCount do
  begin
    StringGrid1.Cells[1, iRow1] := StringGrid1.Cells[StringGrid1.col, iRow1];
  end;
end;

procedure TFormViewGrid.MenuColFixedClearClick(Sender: TObject);
var
  iRow1: Integer;
begin
  for iRow1 := 2 to StringGrid1.RowCount do
  begin
    StringGrid1.Cells[1, iRow1] := '';
  end;
end;

procedure TFormViewGrid.MenuRowFixedClick(Sender: TObject);
var
  iCol1: Integer;
begin
  for iCol1 := 2 to StringGrid1.ColCount do
  begin
    StringGrid1.Cells[iCol1, 1] := StringGrid1.Cells[iCol1, StringGrid1.Row];
  end;
end;

procedure TFormViewGrid.MenuClipTabClick(Sender: TObject);
begin
  //ݒ
  MenuClipTab.Checked             := TRUE;
  MenuClipDoubleCautation.Checked := FALSE;
  MenuClipComma.Checked           := FALSE;
  MenuClipCsvEx2.Checked          := FALSE;
end;

procedure TFormViewGrid.MenuClipDoubleCautationClick(Sender: TObject);
begin
  //ݒ
  MenuClipTab.Checked             := FALSE;
  MenuClipDoubleCautation.Checked := TRUE;
  MenuClipComma.Checked           := FALSE;
  MenuClipCsvEx2.Checked          := FALSE;
end;

procedure TFormViewGrid.MenuClipCommaClick(Sender: TObject);
begin
  //ݒ
  MenuClipTab.Checked             := FALSE;
  MenuClipDoubleCautation.Checked := FALSE;
  MenuClipComma.Checked           := TRUE;
  MenuClipCsvEx2.Checked          := FALSE;
end;

procedure TFormViewGrid.MenuClipCsvEx2Click(Sender: TObject);
begin
  //ݒ
  MenuClipTab.Checked             := FALSE;
  MenuClipDoubleCautation.Checked := FALSE;
  MenuClipComma.Checked           := FALSE;
  MenuClipCsvEx2.Checked          := TRUE;
end;

procedure TFormViewGrid.MenuRowFixedClearClick(Sender: TObject);
var
  iCol1: Integer;
begin
  for iCol1 := 2 to StringGrid1.ColCount do
  begin
    StringGrid1.Cells[iCol1, 1] := '';
  end;
end;

procedure TFormViewGrid.MenuColumnWidthAdjustClick(Sender: TObject);
var
  iRow1: Integer;
  iCol1: Integer;
  iMax1: Integer;
  iWidth1: Integer;
begin
  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //s
  for iCol1 := 0 to StringGrid1.ColCount-1 do
  begin
    iMax1 := 0;
    for iRow1 := 0 to StringGrid1.RowCount-1 do
    begin
      //擾
      iWidth1 := StringGrid1.Canvas.TextWidth(StringGrid1.Cells[iCol1, iRow1]) + 5;
      if iWidth1 > iMax1 then
      begin
        iMax1 := iWidth1;
      end;
    end;
    //ݒ
    StringGrid1.ColWidths[iCol1]:= iMax1;
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;
end;

procedure TFormViewGrid.MenuEditModeClick(Sender: TObject);
begin
  //ݒ
  if MenuEditMode.Checked = FALSE then
  begin
    Memo1.Modified := FALSE;
  end;

  //̐ݒ̂
  MenuEditMode.Checked := TRUE;
  MenuEditMode.Enabled := FALSE;

  //XV
  MemoStatus(Sender);

  //ݒ
  Memo1.ReadOnly := FALSE;
  if BoolJapanese = TRUE then
  begin
    StatusBar1.Panels.Items[1].Text := EDIT_ENABLE_JPN;
    StatusBar2.Panels.Items[1].Text := EDIT_ENABLE_JPN;
  end
  else
  begin
    StatusBar1.Panels.Items[1].Text := EDIT_ENABLE_ENG;
    StatusBar2.Panels.Items[1].Text := EDIT_ENABLE_ENG;
  end;
end;

procedure TFormViewGrid.MenuCopyToClipClick(Sender: TObject);
var
  iLeft1: Integer;
  iRight1: Integer;
  iTop1: Integer;
  iBottom1: Integer;
  StrList1: TStringList;
  iRow1: Integer;
  iCol1: Integer;
  sStr1: String;
  sStr2: String;
begin
  //JԂ𐧌
  if BoolKeyRun = TRUE then
  begin
    Exit;
  end;
  BoolKeyRun := TRUE;

  //擾
  iLeft1   := StringGrid1.Selection.Left;
  iRight1  := StringGrid1.Selection.Right;
  iTop1    := StringGrid1.Selection.Top;
  iBottom1 := StringGrid1.Selection.Bottom;

  //m
  StrList1 := TStringList.Create;

  //s
  for iRow1 := iTop1 to iBottom1 do
  begin
    //NA
    sStr1 := '';
    //
    if      (iRight1 - iLeft1) = 0 then
    begin
          if MenuClipTab.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iLeft1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr2;
          end;
          if MenuClipDoubleCautation.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iLeft1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr2;
          end;
          if MenuClipComma.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iLeft1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr2;
          end;
          if MenuClipCsvEx2.Checked = TRUE then
          begin
            sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iLeft1, iRow1], Sender);
            sStr1 := sStr2;
          end;
          StrList1.Add(sStr1);
    end
    else if (iRight1 - iLeft1) = 1 then
    begin
          if MenuClipTab.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iLeft1,  iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2 + #09;
            sStr2 := StringReplace(StringGrid1.Cells[iRight1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2;
          end;
          if MenuClipDoubleCautation.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iLeft1,  iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + '"' + sStr2 + '","';
            sStr2 := StringReplace(StringGrid1.Cells[iRight1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2 + '"';
          end;
          if MenuClipComma.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iLeft1,  iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2 + ',';
            sStr2 := StringReplace(StringGrid1.Cells[iRight1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2;
          end;
          if MenuClipCsvEx2.Checked = TRUE then
          begin
            sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iLeft1,  iRow1], Sender);
            sStr1 := sStr1 + sStr2 + ',';
            sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iRight1, iRow1], Sender);
            sStr1 := sStr1 + sStr2;
          end;
          StrList1.Add(sStr1);
    end
    else
    begin
      for iCol1 := iLeft1 to iRight1 do
      begin
        //擪
        if iCol1 = iLeft1 then
        begin
          if MenuClipTab.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr2 + #09;
          end;
          if MenuClipDoubleCautation.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := '"' + sStr2 + '","';
          end;
          if MenuClipComma.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr2 + ',';
          end;
          if MenuClipCsvEx2.Checked = TRUE then
          begin
            sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iCol1,  iRow1], Sender);
            sStr1 := sStr2 + ',';
          end;
        end
        //Ō
        else if iCol1 = iRight1 then
        begin
          if MenuClipTab.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2;
          end;
          if MenuClipDoubleCautation.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2 + '"';
          end;
          if MenuClipComma.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2;
          end;
          if MenuClipCsvEx2.Checked = TRUE then
          begin
            sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iCol1,  iRow1], Sender);
            sStr1 := sStr1 + sStr2;
          end;
        end
        //ŌȊO
        else
        begin
          if MenuClipTab.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2 + #09;
          end;
          if MenuClipDoubleCautation.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2 + '","';
          end;
          if MenuClipComma.Checked = TRUE then
          begin
            sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
            sStr1 := sStr1 + sStr2 + ',';
          end;
          if MenuClipCsvEx2.Checked = TRUE then
          begin
            sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iCol1,  iRow1], Sender);
            sStr1 := sStr1 + sStr2 + ',';
          end;
        end;
      end;
      StrList1.Add(sStr1);
    end;
  end;

  //Rs[
  ClipBoard.AsText := StrList1.Text;

  //j
  StrList1.Free;

  //XV
  BoolKeyRun := FALSE;
end;

procedure TFormViewGrid.MenuCsvAutoClick(Sender: TObject);
begin
  //ݒ
  MenuCsvAuto.Checked := TRUE;
  MenuCsvEx1.Checked  := FALSE;
  MenuCsvEx2.Checked  := FALSE;
  IntStrToCsvIndex := 0;
  MenuLoadFileClick(Sender);
end;

procedure TFormViewGrid.MenuCsvEx1Click(Sender: TObject);
begin
  //ݒ
  MenuCsvAuto.Checked := FALSE;
  MenuCsvEx1.Checked  := TRUE;
  MenuCsvEx2.Checked  := FALSE;
  IntStrToCsvIndex := 1;
  MenuLoadFileClick(Sender);
end;

procedure TFormViewGrid.MenuCsvEx2Click(Sender: TObject);
begin
  //ݒ
  MenuCsvAuto.Checked := FALSE;
  MenuCsvEx1.Checked  := FALSE;
  MenuCsvEx2.Checked  := TRUE;
  IntStrToCsvIndex := 2;
  MenuLoadFileClick(Sender);
end;

procedure TFormViewGrid.MenuEscapeClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

procedure TFormViewGrid.MenuFileOverWriteClick(Sender: TObject);
begin
  //ۑ
  SaveCsvFile(StrLoadFileName, Sender);
  //Zbg
  BoolEditModified  := FALSE;
end;

procedure TFormViewGrid.MenuFileSaveAsClick(Sender: TObject);
var
  wRet8: Word;
  sStr8: String;
begin
  //ݒ
  SaveDialog1.InitialDir := ExtractFilePath(StrLoadFileName);
  SaveDialog1.FileName   := ExtractFileName(StrLoadFileName);
  if SaveDialog1.Execute = TRUE then
  begin
    //
    if FileExists(SaveDialog1.FileName) = TRUE then
    begin
      //mF
      if BoolJapanese = TRUE then
      begin
        sStr8 := 'Information : ꖼ̃t@C܂' + #13#10 +
                 '㏑ۑ܂H';
      end
      else
      begin
        sStr8 := 'Information : Same file is already exists.' + #13#10 +
                 'Over write  file ?' + #13#10 +
                 '[YES] / [NO]';
      end;
      wRet8 := MessageDlg(sStr8, mtInformation, [mbYes, mbNo ], 0);
      if wRet8 = mrYes then
      begin
        //ۑ
        SaveCsvFile(SaveDialog1.FileName, Sender);
        //Zbg
        BoolEditModified  := FALSE;
      end
      else
      begin
        //Ȃ
        Exit;
      end;
    end
    else
    begin
        //ۑ
        SaveCsvFile(SaveDialog1.FileName, Sender);
        //Zbg
        BoolEditModified  := FALSE;
    end;
  end;
end;

procedure TFormViewGrid.MenuLfToTabClick(Sender: TObject);
begin
  //ݒ
  if MenuLfToTab.Checked = FALSE then
  begin
    MenuLfToTab.Checked := TRUE;
  end
  else
  begin
    MenuLfToTab.Checked := FALSE;
  end;
end;

procedure TFormViewGrid.MenuLoadFileClick(Sender: TObject);
var
  iRow1: Integer;
  iCol1: Integer;
begin
  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //NA
  for iRow1 := 0 to StringGrid1.RowCount-1 do
  begin
    for iCol1 := 0 to StringGrid1.ColCount-1 do
    begin
      StringGrid1.Cells[iCol1, iRow1] := '';
    end;
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;

  //t@CǍ
  LoadCsvFile(Sender);
  SetCsvFile(Sender);
  //ҏWs
  MenuEditMode.Checked := FALSE;
  MenuEditMode.Enabled := TRUE;
  //Đݒ
  FormResize(Sender);

  //Đݒ
  if BoolJapanese = TRUE then
  begin
    //\
    StatusBar3.Panels.Items[0].Text := '';
    StatusBar3.Panels.Items[1].Text := '';
  end
  else
  begin
    //\
    StatusBar3.Panels.Items[0].Text := '';
    StatusBar3.Panels.Items[1].Text := '';
  end;
end;

procedure TFormViewGrid.MenuMemoHeight1Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := TRUE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 1) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 1) + (IntTextHeight * 2);
  IntMemoHeightIndex := 1;
end;

procedure TFormViewGrid.MenuMemoHeight2Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := TRUE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 2) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 2) + (IntTextHeight * 2);
  IntMemoHeightIndex := 2;
end;

procedure TFormViewGrid.MenuMemoHeight3Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := TRUE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 3) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 3) + (IntTextHeight * 2);
  IntMemoHeightIndex := 3;
end;

procedure TFormViewGrid.MenuMemoHeight4Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := TRUE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 4) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 4) + (IntTextHeight * 2);
  IntMemoHeightIndex := 4;
end;

procedure TFormViewGrid.MenuMemoHeight5Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := TRUE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 5) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 5) + (IntTextHeight * 2);
  IntMemoHeightIndex := 5;
end;

procedure TFormViewGrid.MenuMemoHeight6Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := TRUE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 6) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 6) + (IntTextHeight * 2);
  IntMemoHeightIndex := 6;
end;

procedure TFormViewGrid.MenuMemoHeight7Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := TRUE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 7) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 7) + (IntTextHeight * 2);
  IntMemoHeightIndex := 7;
end;

procedure TFormViewGrid.MenuMemoHeight8Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := TRUE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 8) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 8) + (IntTextHeight * 2);
  IntMemoHeightIndex := 8;
end;

procedure TFormViewGrid.MenuMemoHeight9Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := TRUE;
  MenuMemoHeight10.Checked := FALSE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 9) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 9) + (IntTextHeight * 2);
  IntMemoHeightIndex := 9;
end;

procedure TFormViewGrid.MenuMemoHeight10Click(Sender: TObject);
begin
  //ݒ
  MenuMemoHeight1.Checked  := FALSE;
  MenuMemoHeight2.Checked  := FALSE;
  MenuMemoHeight3.Checked  := FALSE;
  MenuMemoHeight4.Checked  := FALSE;
  MenuMemoHeight5.Checked  := FALSE;
  MenuMemoHeight6.Checked  := FALSE;
  MenuMemoHeight7.Checked  := FALSE;
  MenuMemoHeight8.Checked  := FALSE;
  MenuMemoHeight9.Checked  := FALSE;
  MenuMemoHeight10.Checked := TRUE;

  //ݒ
  Memo1.Height  := (IntTextHeight * 10) + (IntTextHeight * 2);
  Panel3.Height := (IntTextHeight * 10) + (IntTextHeight * 2);
  IntMemoHeightIndex := 10;
end;

procedure TFormViewGrid.MenuQuitClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

procedure TFormViewGrid.Timer4Timer(Sender: TObject);
begin
  //Zbg
  Timer4.Enabled := FALSE;

  //ݒ
  if BoolJapanese = TRUE then
  begin
    FormMain.StrMsg := 'Information : s(CrLf)܂łɁAɉs(Lf)܂';
  end
  else
  begin
    FormMain.StrMsg := 'Information : LineFeed [Lf] is exist, before CarrigeRetun and LineFeed [CrLf].' + #13#10 +
                       '[OK]';
  end;
  ShowMessage(FormMain.StrMsg);
end;

procedure TFormViewGrid.Timer5Timer(Sender: TObject);
begin
  //Zbg
  Timer5.Enabled := FALSE;

  //ݒ
  if BoolJapanese = TRUE then
  begin
    FormMain.StrMsg := 'Information : sȕ܂ł܂̂ŏ֎~t@CƂĈ܂';
  end
  else
  begin
    FormMain.StrMsg := 'Information : Some letters can not acccept, so file type lokks like Read Only File.' + #13#10 +
                       '[OK]';
  end;
  ShowMessage(FormMain.StrMsg);
end;

procedure TFormViewGrid.GetCharCodeRetCode(FileName:String; var CharCode: Integer; var RetCode: Integer; Sender: TObject);
var
  i: Integer;
  iCr1: Integer;
  iLf1: Integer;
  iMax1: Integer;
  hFindData1: THandle;
  pBuff1: PAnsiChar;
  File1: TSearchRec;
  FileStream1: TFileStream;
  bFind1: Boolean;
  bFind2: Boolean;
  MemoryStream1 :TMemoryStream;
  bSJIS_E3: Boolean;
begin
    //t@CTCY擾
    hFindData1 := FindFirst(Filename, faAnyFile, File1);
    try
      ;
    finally
      FindClose(File1);
    end;
    //s
    if hFindData1 <> 0 then
    begin
      //ݒ
      if BoolJapanese = TRUE then
      begin
        FormMain.StrMsg := 'Information : t@C܂';
      end
      else
      begin
        FormMain.StrMsg := 'Information : File is not exist.' + #13#10 +
                           '[OK]';
      end;
      ShowMessage(FormMain.StrMsg);
      Exit;
    end;

    //ă`FbNTCY̏w
    iMax1 := File1.Size;
    if iMax1 > CHAR_UNKNOW_CHECK_SIZE_MAX then
    begin
      iMax1 := CHAR_UNKNOW_CHECK_SIZE_MAX;
    end;

    //
    CharCode := ENCODE_SJIS;
    RetCode := RETURN_CRLF;
    BoolUnKnownCharCode := FALSE;
    //TCY
    if File1.Size = 0 then
    begin
      CharCode := ENCODE_SJIS;
      RetCode := RETURN_NONE;
    end
    //eL
    else
    begin
      //TCYw肵ăm
      pBuff1 := AllocMem(File1.Size);
      try
        //ǂݍ
        try
          //̓t@Cݒ肵ēǂݍ
          FileStream1 := TFileStream.Create(FileName, fmOpenRead);
          try
            FileStream1.Read(pBuff1^, File1.Size);
          finally
            FileStream1.Free;
          end;
        except
          //ReadOnly J 103 ̃G[ɂȂ΍
        end;

        //

        //Size = 1
        if      File1.Size = 1 then
        begin
          CharCode := ENCODE_SJIS;
        end
        //Size == 2
        else if File1.Size = 2 then
        begin
          //ݒ
          i := 0;
          //UTF16 Little BOM
          if      (pBuff1[i+0] = #$FF) and (pBuff1[i+1] = #$FE) then
          begin
            CharCode := ENCODE_UTF16_BOM_ON_LE;
          end
          //UTF16 Big BOM
          else if (pBuff1[i+0] = #$FE) and (pBuff1[i+1] = #$FF) then
          begin
            CharCode := ENCODE_UTF16_BOM_ON_BE;
          end

          // EUC Ђ炪
          else if ( ((pBuff1[i+0] = #$A4) and (pBuff1[i+1] >= #$A1)) and
                    ((pBuff1[i+0] = #$A4) and (pBuff1[i+1] <= #$F3)) )
          then
          begin
            CharCode := ENCODE_EUCJP;
          end
          // EUC J^Ji
          else if ( ((pBuff1[i+0] = #$A5) and (pBuff1[i+1] >= #$A1)) and
                    ((pBuff1[i+0] = #$A5) and (pBuff1[i+1] <= #$F6)) )
          then
          begin
            CharCode := ENCODE_EUCJP;
          end
          // EUC 
          else if ( ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] >= #$B0)) and
                    ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] <= #$B9)) )
          then
          begin
            CharCode := ENCODE_EUCJP;
          end
          // EUC p
          else if ( ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] >= #$C1)) and
                    ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] <= #$Da)) )
          then
          begin
            CharCode := ENCODE_EUCJP;
          end
          // EUC p
          else if ( ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] >= #$E1)) and
                    ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] <= #$FA)) )
          then
          begin
            CharCode := ENCODE_EUCJP;
          end

          // SJIS Ђ炪
          else if ( ((pBuff1[i+0] = #$82) and (pBuff1[i+1] >= #$9F)) and
                    ((pBuff1[i+0] = #$82) and (pBuff1[i+1] <= #$F1)) )
          then
          begin
            CharCode := ENCODE_SJIS;
          end
          // SJIS J^Ji
          else if ( ((pBuff1[i+0] = #$83) and (pBuff1[i+1] >= #$40)) and
                    ((pBuff1[i+0] = #$83) and (pBuff1[i+1] <= #$96)) )
          then
          begin
             CharCode := ENCODE_SJIS;
          end
          // SJIS 
          else if ( ((pBuff1[i+0] = #$82) and (pBuff1[i+1] >= #$4F)) and
                    ((pBuff1[i+0] = #$82) and (pBuff1[i+1] <= #$58)) )
          then
          begin
             CharCode := ENCODE_SJIS;
          end
          // SJIS p
          else if ( ((pBuff1[i+0] = #$82) and (pBuff1[i+1] >= #$60)) and
                    ((pBuff1[i+0] = #$82) and (pBuff1[i+1] <= #$79)) )
          then
          begin
             CharCode := ENCODE_SJIS;
          end
          // SJIS p
          else if ( ((pBuff1[i+0] = #$82) and (pBuff1[i+1] >= #$81)) and
                    ((pBuff1[i+0] = #$82) and (pBuff1[i+1] <= #$9A)) )
          then
          begin
            CharCode := ENCODE_SJIS;
          end

          // JIS Ђ炪
          else if ( ((pBuff1[i+0] = #$24) and (pBuff1[i+1] >= #$21)) and
                    ((pBuff1[i+0] = #$24) and (pBuff1[i+1] <= #$73)) )
          then
          begin
            CharCode := ENCODE_JIS;
          end
          // JIS J^Ji
          else if ( ((pBuff1[i+0] = #$25) and (pBuff1[i+1] >= #$21)) and
                    ((pBuff1[i+0] = #$25) and (pBuff1[i+1] <= #$76)) )
          then
          begin
            CharCode := ENCODE_JIS;
          end
          // JIS 
          else if ( ((pBuff1[i+0] = #$23) and (pBuff1[i+1] >= #$30)) and
                    ((pBuff1[i+0] = #$23) and (pBuff1[i+1] <= #$39)) )
          then
          begin
            CharCode := ENCODE_JIS;
          end
          // JIS p
          else if ( ((pBuff1[i+0] = #$23) and (pBuff1[i+1] >= #$41)) and
                    ((pBuff1[i+0] = #$23) and (pBuff1[i+1] <= #$5A)) )
          then
          begin
            CharCode := ENCODE_JIS;
          end
          // JIS p
          else if ( ((pBuff1[i+0] = #$23) and (pBuff1[i+1] >= #$61)) and
                    ((pBuff1[i+0] = #$23) and (pBuff1[i+1] <= #$7A)) )
          then
          begin
            CharCode := ENCODE_JIS;
          end

          //
          else
          begin
            CharCode := ENCODE_SJIS;
          end;
        end
        //Size >= 3
        else
        begin
          //UTF8 BOM
          if (pBuff1[0] = #$EF) and (pBuff1[1] = #$BB) and (pBuff1[2] = #$BF) then
          begin
            CharCode := ENCODE_UTF8_BOM_ON;
          end
          //UTF16 Little BOM
          else if(pBuff1[0] = #$FF) and (pBuff1[1] = #$FE) then
          begin
            CharCode := ENCODE_UTF16_BOM_ON_LE;
          end
          //UTF16 Big BOM
          else if (pBuff1[0] = #$FE) and (pBuff1[1] = #$FF) then
          begin
            CharCode := ENCODE_UTF16_BOM_ON_BE;
          end
          //ȊO
          else
          begin
            //
            bFind1   := FALSE;
            bSJIS_E3 := FALSE;

            //
            if bFind1 = FALSE then
            begin
              // UTF8  SJIS ͓͈͂̂ōŏ UTF8 𔻒肷
              bSJIS_E3 := FALSE;
              for i := 0 to (File1.Size-4) do
              begin
                //͈͂
                if i >= 1 then
                begin
                  //UTF8  #$E3  SJIS ̊Y͈͂ɂꍇ͏
                  if( ((pBuff1[i-1] = #$81) and (pBuff1[i-0] >= #$E3)) or
                      ((pBuff1[i-1] = #$82) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$83) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$84) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$85) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$86) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$87) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$88) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$89) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$8A) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$8B) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$8C) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$8D) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$8E) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$8F) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$90) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$91) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$92) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$93) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$94) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$95) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$96) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$97) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$98) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$99) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$9A) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$9B) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$9C) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$9D) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$9E) and (pBuff1[i-0] <= #$E3)) or
                      ((pBuff1[i-1] = #$9F) and (pBuff1[i-0] <= #$E3))
                      )
                  then
                  begin
                    bSJIS_E3 := TRUE;
                  end
                  else
                  begin
                    bSJIS_E3 := FALSE;
                  end;
                end;

                // UTF8 Ђ炪
                if( ((pBuff1[i+0] = #$E3) and (pBuff1[i+1] = #$81) and (pBuff1[i+2] >= #$81)) and
                    ((pBuff1[i+0] = #$E3) and (pBuff1[i+1] = #$81) and (pBuff1[i+2] <= #$BF)) )
                then
                begin
                  if bSJIS_E3 = FALSE then
                  begin
                    CharCode := ENCODE_UTF8_BOM_OFF;
                    bFind1 := TRUE;
                    Break;
                  end;
                end;
                // UTF8 Ђ炪
                if( ((pBuff1[i+0] = #$E3) and (pBuff1[i+1] = #$82) and (pBuff1[i+2] >= #$80)) and
                    ((pBuff1[i+0] = #$E3) and (pBuff1[i+1] = #$82) and (pBuff1[i+2] <= #$93)) )
                then
                begin
                  if bSJIS_E3 = FALSE then
                  begin
                    CharCode := ENCODE_UTF8_BOM_OFF;
                    bFind1 := TRUE;
                    Break;
                  end;
                end;
                // UTF8 J^Ji
                if ( ((pBuff1[i+0] = #$E3) and (pBuff1[i+1] = #$82) and (pBuff1[i+2] >= #$A1)) and
                     ((pBuff1[i+0] = #$E3) and (pBuff1[i+1] = #$82) and (pBuff1[i+2] <= #$BF)) )
                then
                begin
                  if bSJIS_E3 = FALSE then
                  begin
                    CharCode := ENCODE_UTF8_BOM_OFF;
                    bFind1 := TRUE;
                    Break;
                  end;
                end;
                // UTF8 J^Ji
                if ( ((pBuff1[i+0] = #$E3) and (pBuff1[i+1] = #$83) and (pBuff1[i+2] >= #$80)) and
                     ((pBuff1[i+0] = #$E3) and (pBuff1[i+1] = #$83) and (pBuff1[i+2] <= #$B6)) )
                then
                begin
                  if bSJIS_E3 = FALSE then
                  begin
                    CharCode := ENCODE_UTF8_BOM_OFF;
                    bFind1 := TRUE;
                    Break;
                  end;
                end;
                // UTF8 
                if ( ((pBuff1[i+0] = #$EF) and (pBuff1[i+1] = #$BC) and (pBuff1[i+2] >= #$90)) and
                     ((pBuff1[i+0] = #$EF) and (pBuff1[i+1] = #$BC) and (pBuff1[i+2] <= #$99)) )
                then
                begin
                  if bSJIS_E3 = FALSE then
                  begin
                    CharCode := ENCODE_UTF8_BOM_OFF;
                    bFind1 := TRUE;
                    Break;
                  end;
                end;
                // UTF8 p
                if ( ((pBuff1[i+0] = #$EF) and (pBuff1[i+1] = #$BC) and (pBuff1[i+2] >= #$A1)) and
                     ((pBuff1[i+0] = #$EF) and (pBuff1[i+1] = #$BC) and (pBuff1[i+2] <= #$BA)) )
                then
                begin
                  if bSJIS_E3 = FALSE then
                  begin
                    CharCode := ENCODE_UTF8_BOM_OFF;
                    bFind1 := TRUE;
                    Break;
                  end;
                end;
                // UTF8 p
                if ( ((pBuff1[i+0] = #$EF) and (pBuff1[i+1] = #$BD) and (pBuff1[i+2] >= #$81)) and
                     ((pBuff1[i+0] = #$EF) and (pBuff1[i+1] = #$BD) and (pBuff1[i+2] <= #$9A)) )
                then
                begin
                  if bSJIS_E3 = FALSE then
                  begin
                    CharCode := ENCODE_UTF8_BOM_OFF;
                    bFind1 := TRUE;
                    Break;
                  end;
                end;
              end;
            end;

            //
            if bFind1 = FALSE then
            begin
              for i := 0 to (File1.Size-3) do
              begin
                // SJIS Ђ炪
                if ( ((pBuff1[i+0] = #$82) and (pBuff1[i+1] >= #$9F)) and
                     ((pBuff1[i+0] = #$82) and (pBuff1[i+1] <= #$F1)) )
                then
                begin
                  CharCode := ENCODE_SJIS;
                  bFind1 := TRUE;
                  Break;
                end;
                // SJIS J^Ji
                if ( ((pBuff1[i+0] = #$83) and (pBuff1[i+1] >= #$40)) and
                     ((pBuff1[i+0] = #$83) and (pBuff1[i+1] <= #$96)) )
                then
                begin
                  CharCode := ENCODE_SJIS;
                  bFind1 := TRUE;
                  Break;
                end;
                // SJIS 
                if ( ((pBuff1[i+0] = #$82) and (pBuff1[i+1] >= #$4F)) and
                     ((pBuff1[i+0] = #$82) and (pBuff1[i+1] <= #$58)) )
                then
                begin
                  CharCode := ENCODE_SJIS;
                  bFind1 := TRUE;
                  Break;
                end;
                // SJIS p
                if ( ((pBuff1[i+0] = #$82) and (pBuff1[i+1] >= #$60)) and
                     ((pBuff1[i+0] = #$82) and (pBuff1[i+1] <= #$79)) )
                then
                begin
                  CharCode := ENCODE_SJIS;
                  bFind1 := TRUE;
                  Break;
                end;
                // SJIS p
                if ( ((pBuff1[i+0] = #$82) and (pBuff1[i+1] >= #$81)) and
                     ((pBuff1[i+0] = #$82) and (pBuff1[i+1] <= #$9A)) )
                then
                begin
                  CharCode := ENCODE_SJIS;
                  bFind1 := TRUE;
                  Break;
                end;
              end;
            end;


            //
            if bFind1 = FALSE then
            begin
              for i := 0 to (File1.Size-3) do
              begin
                // EUC Ђ炪
                if ( ((pBuff1[i+0] = #$A4) and (pBuff1[i+1] >= #$A1)) and
                     ((pBuff1[i+0] = #$A4) and (pBuff1[i+1] <= #$F3)) )
                then
                begin
                  CharCode := ENCODE_EUCJP;
                  bFind1 := TRUE;
                  Break;
                end;
                // EUC J^Ji
                if ( ((pBuff1[i+0] = #$A5) and (pBuff1[i+1] >= #$A1)) and
                     ((pBuff1[i+0] = #$A5) and (pBuff1[i+1] <= #$F6)) )
                then
                begin
                  CharCode := ENCODE_EUCJP;
                  bFind1 := TRUE;
                  Break;
                end;
                // EUC 
                if ( ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] >= #$B0)) and
                     ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] <= #$B9)) )
                then
                begin
                  CharCode := ENCODE_EUCJP;
                  bFind1 := TRUE;
                  Break;
                end;
                // EUC p
                if ( ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] >= #$C1)) and
                     ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] <= #$Da)) )
                then
                begin
                  CharCode := ENCODE_EUCJP;
                  bFind1 := TRUE;
                  Break;
                end;
                // EUC p
                if ( ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] >= #$E1)) and
                     ((pBuff1[i+0] = #$A3) and (pBuff1[i+1] <= #$FA)) )
                then
                begin
                  CharCode := ENCODE_EUCJP;
                  bFind1 := TRUE;
                  Break;
                end;
              end;
            end;


            //
            if bFind1 = FALSE then
            begin
              for i := 0 to (File1.Size-3) do
              begin
                // JIS Ђ炪
                if ( ((pBuff1[i+0] = #$24) and (pBuff1[i+1] >= #$21)) and
                     ((pBuff1[i+0] = #$24) and (pBuff1[i+1] <= #$73)) )
                then
                begin
                  CharCode := ENCODE_JIS;
                  bFind1 := TRUE;
                  Break;
                end;
                // JIS J^Ji
                if ( ((pBuff1[i+0] = #$25) and (pBuff1[i+1] >= #$21)) and
                     ((pBuff1[i+0] = #$25) and (pBuff1[i+1] <= #$76)) )
                then
                begin
                  CharCode := ENCODE_JIS;
                  bFind1 := TRUE;
                  Break;
                end;
                // JIS 
                if ( ((pBuff1[i+0] = #$23) and (pBuff1[i+1] >= #$30)) and
                     ((pBuff1[i+0] = #$23) and (pBuff1[i+1] <= #$39)) )
                then
                begin
                  CharCode := ENCODE_JIS;
                  bFind1 := TRUE;
                  Break;
                end;
                // JIS p
                if ( ((pBuff1[i+0] = #$23) and (pBuff1[i+1] >= #$41)) and
                     ((pBuff1[i+0] = #$23) and (pBuff1[i+1] <= #$5A)) )
                then
                begin
                  CharCode := ENCODE_JIS;
                  bFind1 := TRUE;
                  Break;
                end;
                // JIS p
                if ( ((pBuff1[i+0] = #$23) and (pBuff1[i+1] >= #$61)) and
                     ((pBuff1[i+0] = #$23) and (pBuff1[i+1] <= #$7A)) )
                then
                begin
                  CharCode := ENCODE_JIS;
                  bFind1 := TRUE;
                  Break;
                end;
              end;
            end;


            //xoȂ
            if bFind1 = FALSE then
            begin
              //
            end;
            //xoȂ
            if bSJIS_E3 = FALSE then
            begin
              //
            end;
          end;
        end;

        //s
        bFind2 := FALSE;
        //UTF16  BOM ɂJnʒuύX
        if (CharCode = ENCODE_UTF16_BOM_ON_BE) or (CharCode = ENCODE_UTF16_BOM_ON_LE)  then
        begin
          i := 2;
        end
        else
        begin
          i := 0;
        end;
        //2ɏ̂ for ͎gȂ
        while i <= (File1.Size-4) do
        begin
          // UTF16LE CRLF
          if      ( (pBuff1[i+0] = #13) and (pBuff1[i+1] = #00) and (pBuff1[i+2] = #10) and (pBuff1[i+3] = #00) )
          then
          begin
            RetCode := RETURN_CRLF;
            bFind2 := TRUE;
            //R[h̔ł͖Ȃ̂ŉs̕тŔ
            if (CharCode <> ENCODE_UTF16_BOM_ON_BE) and (CharCode <> ENCODE_UTF16_BOM_ON_LE)then
            begin
              CharCode := ENCODE_UTF16_BOM_OFF_LE;
            end;
            Break;
          end
          // UTF16BE CRLF
          else if ( (pBuff1[i+0] = #00) and (pBuff1[i+1] = #13) and (pBuff1[i+2] = #00) and (pBuff1[i+3] = #10) )
          then
          begin
            RetCode := RETURN_CRLF;
            bFind2 := TRUE;
            //R[h̔ł͖Ȃ̂ŉs̕тŔ
            if (CharCode <> ENCODE_UTF16_BOM_ON_BE) and (CharCode <> ENCODE_UTF16_BOM_ON_LE)then
            begin
              CharCode := ENCODE_UTF16_BOM_OFF_BE;
            end;
            Break;
          end
          //
          else
          begin
          end;
          //2̏
          i := i+2;
        end;

        //
        if bFind2 = FALSE then
        begin
          //2ɏ̂ for ͎gȂ
          i := 0;
          while i <= (File1.Size-2) do
          begin
            // UTF16BE CR
            if      ( (pBuff1[i+0] = #00) and (pBuff1[i+1] = #13) )
            then
            begin
              RetCode := RETURN_CR;
              bFind2 := TRUE;
              //R[h̔ł͖Ȃ̂ŉs̕тŔ
              if (CharCode <> ENCODE_UTF16_BOM_ON_BE) and (CharCode <> ENCODE_UTF16_BOM_ON_LE)then
              begin
                CharCode := ENCODE_UTF16_BOM_OFF_BE;
              end;
              Break;
            end
            // UTF16LE CR
            else if ( (pBuff1[i+0] = #13) and (pBuff1[i+1] = #00) )
            then
            begin
              RetCode := RETURN_CR;
              bFind2 := TRUE;
              //R[h̔ł͖Ȃ̂ŉs̕тŔ
              if (CharCode <> ENCODE_UTF16_BOM_ON_BE) and (CharCode <> ENCODE_UTF16_BOM_ON_LE)then
              begin
                CharCode := ENCODE_UTF16_BOM_OFF_LE;
              end;
              Break;
            end
            // UTF16BE LF
            else if ( (pBuff1[i+0] = #00) and (pBuff1[i+1] = #10) )
            then
            begin
              RetCode := RETURN_LF;
              bFind2 := TRUE;
              //R[h̔ł͖Ȃ̂ŉs̕тŔ
              if (CharCode <> ENCODE_UTF16_BOM_ON_BE) and (CharCode <> ENCODE_UTF16_BOM_ON_LE)then
              begin
                CharCode := ENCODE_UTF16_BOM_OFF_BE;
              end;
              Break;
            end
            // UTF16LE LF
            else if ( (pBuff1[i+0] = #10) and (pBuff1[i+1] = #00) )
            then
            begin
              RetCode := RETURN_LF;
              bFind2 := TRUE;
              //R[h̔ł͖Ȃ̂ŉs̕тŔ
              if (CharCode <> ENCODE_UTF16_BOM_ON_BE) and (CharCode <> ENCODE_UTF16_BOM_ON_LE)then
              begin
                CharCode := ENCODE_UTF16_BOM_OFF_LE;
              end;
              Break;
            end
            //
            else
            begin
            end;
            //2̏
            i := i+2;
          end;
        end;

        //
        if bFind2 = FALSE then
        begin
          iCr1 := 0;
          iLf1 := 0;
          BoolLfInLine := FALSE;
          for i := 0 to (File1.Size-1) do
          begin
            // CR
            if      ( (pBuff1[i+0] = #13) )
            then
            begin
              iCr1 := iCr1+1;
            end
            // LF
            else if ( (pBuff1[i+0] = #10) )
            then
            begin
              iLf1 := iLf1+1;
            end
            //
            else
            begin
            end;
          end;
          //ݒ
          if (iCr1 > 0) and (iLf1 > 0) then
          begin
            RetCode := RETURN_CRLF;
            //s Lf 
            if iCr1 <> iLf1 then
            begin
              BoolLfInLine := TRUE;
            end
            else
            begin
              BoolLfInLine := FALSE;
            end;
          end
          else if (iCr1 > 0) then
          begin
              RetCode := RETURN_CR;
          end
          else if (iLf1 > 0) then
          begin
              RetCode := RETURN_LF;
          end
          else
          begin
              RetCode := RETURN_NONE;
          end;
        end;
        //x\Ȃ
        if bFind2 = TRUE then
        begin
        end;

        //ēx
         if (CharCode = ENCODE_UTF16_BOM_ON_BE) or (CharCode = ENCODE_UTF16_BOM_OFF_BE) or
            (CharCode = ENCODE_UTF16_BOM_ON_LE) or (CharCode = ENCODE_UTF16_BOM_OFF_LE) then
         begin
           //
         end
         //ȊO
         else
         begin
           //m
           MemoryStream1 := TMemoryStream.Create;
           try
              //
              for i := 0 to (iMax1-1) do
              begin
                //s
                //#$1B  JIS Ŏgp
                if (pBuff1[i] in [#$00..#$08, #$0B..#$0C, #$0E..#$1A, #$1C..#$1F]) then
                begin
                  BoolUnKnownCharCode := TRUE;
                  Break;
                end;
              end;
              //s̏ꍇ
              if BoolUnKnownCharCode = TRUE then
              begin
                for i := 0 to (File1.Size-1) do
                begin
                  if (pBuff1[i] in [#$00..#$08, #$0B..#$0C, #$0E..#$1F]) then
                  begin
                    //
                  end
                  else
                  begin
                    MemoryStream1.Write(pBuff1[i], Length(pBuff1[i]));
                  end;
                end;
              end;
              //ۑ
              MemoryStream1.SaveToFile(ExtractFilePath(ParamStr(0)) + UNKOWN_DST_LIST);
           finally
             //j
             MemoryStream1.Free;
           end;
         end;
      finally
        FreeMem(pBuff1);
      end;
    end;


end;

procedure TFormViewGrid.LoadCsvFile(Sender: TObject);
var
  hFindData1: THandle;
  File1: TSearchRec;
  FileStream1 :TFileStream;
  pBuff1: PAnsiChar;
  pAnsiChar1: PAnsiChar;
  iRow1: Integer;
  iRow2: Integer;
  iStart1: Integer;
  iEnd1: Integer;
  iCopyCnt: Integer;
  sjis: SJISString;
  utf8: UTF8String;
begin
  //t@C擾
  hFindData1 := FindFirst(StrLoadFileName, faAnyFile, File1);
  try
    ;
  finally
    FindClose(File1);
  end;
  //s
  if hFindData1 <> 0 then
  begin
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : t@C܂';
    end
    else
    begin
      FormMain.StrMsg := 'Information : File is not exist.' + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    Exit;
  end;
  //t@CTCY
  IntLoadFileSize := File1.Size;

  //\
  StatusBar1.Panels.Items[0].Text := '';
  StatusBar1.Panels.Items[1].Text := '';
  StatusBar1.Panels.Items[2].Text := '';
  StatusBar1.Panels.Items[3].Text := '';
  StatusBar1.Panels.Items[4].Text := '';
  StatusBar1.Panels.Items[5].Text := '';
  StatusBar1.Panels.Items[6].Text := '';
  StatusBar2.Panels.Items[0].Text := '';
  StatusBar2.Panels.Items[1].Text := '';
  StatusBar2.Panels.Items[2].Text := '';
  StatusBar2.Panels.Items[3].Text := '';
  StatusBar2.Panels.Items[4].Text := '';
  StatusBar2.Panels.Items[5].Text := '';
  StatusBar2.Panels.Items[6].Text := '';

  //
  IntReadEncode1 := ENCODE_UNKNOWN;
  IntReadReturn1 := RETURN_NONE;

  //͎o
  if MenuReadAuto.Checked = TRUE then
  begin
    //擾
    GetCharCodeRetCode(StrLoadFileName, IntReadEncode1, IntReadReturn1, Sender);
  end
  //͐ݒɏ]
  else
  begin
    //擾 spɎs
    GetCharCodeRetCode(StrLoadFileName, IntReadEncode1, IntReadReturn1, Sender);
    //ݒ
    if      MenuReadSJIS.Checked = TRUE then
    begin
      IntReadEncode1 := ENCODE_SJIS;
    end
    else if MenuReadUTF8Bom.Checked = TRUE then
    begin
      IntReadEncode1 := ENCODE_UTF8_BOM_ON;
    end
    else if MenuReadUTF8.Checked = TRUE then
    begin
      IntReadEncode1 := ENCODE_UTF8_BOM_OFF;
    end
    else if MenuReadUTF16BomBe.Checked = TRUE then
    begin
      IntReadEncode1 := ENCODE_UTF16_BOM_ON_BE;
    end
    else if MenuReadUTF16Be.Checked = TRUE then
    begin
      IntReadEncode1 := ENCODE_UTF16_BOM_OFF_BE;
    end
    else if MenuReadUTF16BomLe.Checked = TRUE then
    begin
      IntReadEncode1 := ENCODE_UTF16_BOM_ON_LE;
    end
    else if MenuReadUTF16Le.Checked = TRUE then
    begin
      IntReadEncode1 := ENCODE_UTF16_BOM_OFF_LE;
    end
    else
    begin
      IntReadEncode1 := ENCODE_UNKNOWN;
    end;
  end;

  //NA
  StrListLoadFile.Clear;

  //UnKnown
  if BoolUnKnownCharCode = TRUE then
  begin
    Timer5.Enabled := TRUE;
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_BINARY;
    end
    else
    begin
      StrLoadChar := CHAR_BINARY_ENG;
    end;
  end
  //SJIS
  else if IntReadEncode1 = ENCODE_SJIS then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_SJIS;
    end
    else
    begin
      StrLoadChar := CHAR_SJIS_ENG;
    end;
  end
  //UTF8 BOM
  else if IntReadEncode1 = ENCODE_UTF8_BOM_ON then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_UTF8_BOM_ON;
    end
    else
    begin
      StrLoadChar := CHAR_UTF8_BOM_ON_ENG;
    end;
  end
  //UTF8
  else if IntReadEncode1 = ENCODE_UTF8_BOM_OFF then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_UTF8_BOM_OFF;
    end
    else
    begin
      StrLoadChar := CHAR_UTF8_BOM_OFF_ENG;
    end;
  end
  //UTF16BOM BE
  else if IntReadEncode1 = ENCODE_UTF16_BOM_ON_BE then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_UTF16_BOM_ON_BE;
    end
    else
    begin
      StrLoadChar := CHAR_UTF16_BOM_ON_BE_ENG;
    end;
  end
  //UTF16 BE
  else if IntReadEncode1 = ENCODE_UTF16_BOM_OFF_BE then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_UTF16_BOM_OFF_BE;
    end
    else
    begin
      StrLoadChar := CHAR_UTF16_BOM_OFF_BE_ENG;
    end;
  end
  //UTF16BOM LE
  else if IntReadEncode1 = ENCODE_UTF16_BOM_ON_LE then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_UTF16_BOM_ON_LE;
    end
    else
    begin
      StrLoadChar := CHAR_UTF16_BOM_ON_LE_ENG;
    end;
  end
  //UTF16 LE
  else if IntReadEncode1 = ENCODE_UTF16_BOM_OFF_LE then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_UTF16_BOM_OFF_LE;
    end
    else
    begin
      StrLoadChar := CHAR_UTF16_BOM_OFF_LE_ENG;
    end;
  end
  //͂Ȃ
  else
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadChar := CHAR_UNKNOWN;
    end
    else
    begin
      StrLoadChar := CHAR_UNKNOWN_ENG;
    end;
  end;

  //CR+LF
  if IntReadReturn1 = RETURN_CRLF then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadReturn := RET_CRLF;
    end
    else
    begin
      StrLoadReturn := RET_CRLF_ENG;
    end;
  end
  //CR
  else if IntReadReturn1 = RETURN_CR then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadReturn := RET_CR;
    end
    else
    begin
      StrLoadReturn := RET_CR_ENG;
    end;
  end
  //LF
  else if IntReadReturn1 = RETURN_LF then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadReturn := RET_LF;
    end
    else
    begin
      StrLoadReturn := RET_LF_ENG;
    end;
  end
  //NONE
  else if IntReadReturn1 = RETURN_NONE then
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadReturn := RET_NONE;
    end
    else
    begin
      StrLoadReturn := RET_NONE_ENG;
    end;
  end
  //NONE
  else
  begin
    if BoolJapanese = TRUE then
    begin
      StrLoadReturn := RET_UNKNOWN;
    end
    else
    begin
      StrLoadReturn := RET_UNKNOWN_ENG;
    end;
  end;

  //֎~ ܂ s
  if (Bool(File1.Attr and FILE_ATTRIBUTE_READONLY) = TRUE) or (BoolUnKnownCharCode = TRUE) then
  begin
    //ݒ
    BoolFileReadOnly := TRUE;
    //ݒ
    if BoolJapanese = TRUE then
    begin
      //\
      StatusBar1.Panels.Items[2].Text := FILE_READ_ONLY;
      StatusBar1.Panels.Items[4].Text := StrLoadChar;
      StatusBar1.Panels.Items[5].Text := StrLoadReturn;
      StatusBar1.Panels.Items[6].Text := '';
      StatusBar2.Panels.Items[2].Text := FILE_READ_ONLY;
      StatusBar2.Panels.Items[4].Text := StrLoadChar;
      StatusBar2.Panels.Items[5].Text := StrLoadReturn;
      StatusBar2.Panels.Items[6].Text := '';
    end
    else
    begin
      //\
      StatusBar1.Panels.Items[2].Text := FILE_READ_ONLY_ENG;
      StatusBar1.Panels.Items[4].Text := StrLoadChar;
      StatusBar1.Panels.Items[5].Text := StrLoadReturn;
      StatusBar1.Panels.Items[6].Text := '';
      StatusBar2.Panels.Items[2].Text := FILE_READ_ONLY_ENG;
      StatusBar2.Panels.Items[4].Text := StrLoadChar;
      StatusBar2.Panels.Items[5].Text := StrLoadReturn;
      StatusBar2.Panels.Items[6].Text := '';
    end;
  end
  else
  begin
    //ݒ
    BoolFileReadOnly := FALSE;
    //ݒ
    if BoolJapanese = TRUE then
    begin
      //\
      StatusBar1.Panels.Items[2].Text := '';
      StatusBar1.Panels.Items[4].Text := StrLoadChar;
      StatusBar1.Panels.Items[5].Text := StrLoadReturn;
      StatusBar1.Panels.Items[6].Text := '';
      StatusBar2.Panels.Items[2].Text := '';
      StatusBar2.Panels.Items[4].Text := StrLoadChar;
      StatusBar2.Panels.Items[5].Text := StrLoadReturn;
      StatusBar2.Panels.Items[6].Text := '';
    end
    else
    begin
      //\
      StatusBar1.Panels.Items[2].Text := '';
      StatusBar1.Panels.Items[4].Text := StrLoadChar;
      StatusBar1.Panels.Items[5].Text := StrLoadReturn;
      StatusBar1.Panels.Items[6].Text := '';
      StatusBar2.Panels.Items[2].Text := '';
      StatusBar2.Panels.Items[4].Text := StrLoadChar;
      StatusBar2.Panels.Items[5].Text := StrLoadReturn;
      StatusBar2.Panels.Items[6].Text := '';
    end;
  end;


  //TCYw肵ăm
  pBuff1 := AllocMem(File1.Size);
  try
    //ǂݍ
    try
      //̓t@Cݒ肵ēǂݍ
      FileStream1 := TFileStream.Create(StrLoadFileName, fmOpenRead);
      try
        FileStream1.Read(pBuff1^, File1.Size);
      finally
        FileStream1.Free;
      end;

      //͈͂w
      //UTF8 BOM
      if IntReadEncode1 = ENCODE_UTF8_BOM_ON then
      begin
        iStart1 := 3;
        iEnd1   := File1.Size;
      end
      //UTF16BOM BE
      else if IntReadEncode1 = ENCODE_UTF16_BOM_ON_BE then
      begin
        iStart1 := 2;
        iEnd1   := File1.Size;
      end
      //UTF16BOM LE
      else if IntReadEncode1 = ENCODE_UTF16_BOM_ON_LE then
      begin
        iStart1 := 2;
        iEnd1   := File1.Size;
      end
      //ȊO
      else
      begin
        iStart1 := 0;
        iEnd1   := File1.Size;
      end;

      //UTF16 LE
      if      (IntReadEncode1 = ENCODE_UTF16_BOM_ON_LE) or (IntReadEncode1 = ENCODE_UTF16_BOM_OFF_LE) then
      begin
        //2ɏ̂ for ͎gȂ
        iRow1 := iStart1;
        while iRow1 < iEnd1 do
        begin
          //CR+Lf
          if IntReadReturn1 = RETURN_CRLF then
          begin
            if      ( (pBuff1[iRow1+0] = #13) and (pBuff1[iRow1+1] = #00) and (pBuff1[iRow1+2] = #10) and (pBuff1[iRow1+3] = #00) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //4̏
              iRow1 := iRow1 + 4;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 2;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //4̏
              iRow1 := iRow1 + 4;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //CR
          else if IntReadReturn1 = RETURN_CR then
          begin
            if      ( (pBuff1[iRow1+0] = #13) and (pBuff1[iRow1+1] = #00) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 2;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //LF
          else if IntReadReturn1 = RETURN_LF then
          begin
            if      ( (pBuff1[iRow1+0] = #10) and (pBuff1[iRow1+1] = #00) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 2;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //NONE
          else if IntReadReturn1 = RETURN_NONE then
          begin
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end;
          end;
        end;
      end
      //UTF16 BE
      else if (IntReadEncode1 = ENCODE_UTF16_BOM_ON_BE) or (IntReadEncode1 = ENCODE_UTF16_BOM_OFF_BE) then
      begin
        //2ɏ̂ for ͎gȂ
        iRow1 := iStart1;
        while iRow1 < iEnd1 do
        begin
          //CR+Lf
          if IntReadReturn1 = RETURN_CRLF then
          begin
            if      ( (pBuff1[iRow1+0] = #00) and (pBuff1[iRow1+1] = #13) and (pBuff1[iRow1+2] = #00) and (pBuff1[iRow1+3] = #10) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                if (iRow2 mod 2) = 0 then
                begin
                  pAnsiChar1[iRow2+1] := pBuff1[iStart1+iRow2];
                end
                else
                begin
                  pAnsiChar1[iRow2-1] := pBuff1[iStart1+iRow2];
                end;
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //4̏
              iRow1 := iRow1 + 4;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 2;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                if (iRow2 mod 2) = 0 then
                begin
                  pAnsiChar1[iRow2+1] := pBuff1[iStart1+iRow2];
                end
                else
                begin
                  pAnsiChar1[iRow2-1] := pBuff1[iStart1+iRow2];
                end;
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //4̏
              iRow1 := iRow1 + 4;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //CR
          else if IntReadReturn1 = RETURN_CR then
          begin
            if      ( (pBuff1[iRow1+0] = #00) and (pBuff1[iRow1+1] = #13) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                if (iRow2 mod 2) = 0 then
                begin
                  pAnsiChar1[iRow2+1] := pBuff1[iStart1+iRow2];
                end
                else
                begin
                  pAnsiChar1[iRow2-1] := pBuff1[iStart1+iRow2];
                end;
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 2;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                if (iRow2 mod 2) = 0 then
                begin
                  pAnsiChar1[iRow2+1] := pBuff1[iStart1+iRow2];
                end
                else
                begin
                  pAnsiChar1[iRow2-1] := pBuff1[iStart1+iRow2];
                end;
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //LF
          else if IntReadReturn1 = RETURN_LF then
          begin
            if      ( (pBuff1[iRow1+0] = #00) and (pBuff1[iRow1+1] = #10) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                if (iRow2 mod 2) = 0 then
                begin
                  pAnsiChar1[iRow2+1] := pBuff1[iStart1+iRow2];
                end
                else
                begin
                  pAnsiChar1[iRow2-1] := pBuff1[iStart1+iRow2];
                end;
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 2;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                if (iRow2 mod 2) = 0 then
                begin
                  pAnsiChar1[iRow2+1] := pBuff1[iStart1+iRow2];
                end
                else
                begin
                  pAnsiChar1[iRow2-1] := pBuff1[iStart1+iRow2];
                end;
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //NONE
          else if IntReadReturn1 = RETURN_NONE then
          begin
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+2);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                if (iRow2 mod 2) = 0 then
                begin
                  pAnsiChar1[iRow2+1] := pBuff1[iStart1+iRow2];
                end
                else
                begin
                  pAnsiChar1[iRow2-1] := pBuff1[iStart1+iRow2];
                end;
              end;
              pAnsiChar1[iCopyCnt+0] := #00;
              pAnsiChar1[iCopyCnt+1] := #00;
              //ǉ
              StrListLoadFile.Add(WideString(PWideChar(@pAnsiChar1[0])));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+2);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end;
          end;
        end;
      end
      //UTF8
      else if (IntReadEncode1 = ENCODE_UTF8_BOM_ON) or (IntReadEncode1 = ENCODE_UTF8_BOM_OFF) then
      begin
        //2ɏ̂ for ͎gȂ
        iRow1 := iStart1;
        while iRow1 < iEnd1 do
        begin
          //CR+Lf
          if IntReadReturn1 = RETURN_CRLF then
          begin
            if      ( (pBuff1[iRow1+0] = #13) and (pBuff1[iRow1+1] = #10) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              utf8 := pAnsiChar1;
              StrListLoadFile.Add(WideString(utf8));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 1;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              utf8 := pAnsiChar1;
              StrListLoadFile.Add(WideString(utf8));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //CR
          else if IntReadReturn1 = RETURN_CR then
          begin
            if      ( (pBuff1[iRow1+0] = #13) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              utf8 := pAnsiChar1;
              StrListLoadFile.Add(WideString(utf8));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 1;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              utf8 := pAnsiChar1;
              StrListLoadFile.Add(WideString(utf8));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //LF
          else if IntReadReturn1 = RETURN_LF then
          begin
            if      ( (pBuff1[iRow1+0] = #10)) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              utf8 := pAnsiChar1;
              StrListLoadFile.Add(WideString(utf8));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 1;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              utf8 := pAnsiChar1;
              StrListLoadFile.Add(WideString(utf8));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //NONE
          else if IntReadReturn1 = RETURN_NONE then
          begin
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              utf8 := pAnsiChar1;
              StrListLoadFile.Add(WideString(utf8));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end;
          end;
        end;
      end
      //ȊO
      else
      begin
        //2ɏ̂ for ͎gȂ
        iRow1 := iStart1;
        while iRow1 < iEnd1 do
        begin
          //CR+Lf
          if IntReadReturn1 = RETURN_CRLF then
          begin
            if      ( (pBuff1[iRow1+0] = #13) and (pBuff1[iRow1+1] = #10) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              sjis := pAnsiChar1;
              StrListLoadFile.Add(WideString(sjis));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 1;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              sjis := pAnsiChar1;
              StrListLoadFile.Add(WideString(sjis));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //2̏
              iRow1 := iRow1 + 2;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //CR
          else if IntReadReturn1 = RETURN_CR then
          begin
            if      ( (pBuff1[iRow1+0] = #13) ) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              sjis := pAnsiChar1;
              StrListLoadFile.Add(WideString(sjis));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 1;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              sjis := pAnsiChar1;
              StrListLoadFile.Add(WideString(sjis));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //LF
          else if IntReadReturn1 = RETURN_LF then
          begin
            if      ( (pBuff1[iRow1+0] = #10)) then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              sjis := pAnsiChar1;
              StrListLoadFile.Add(WideString(sjis));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end
            else
            begin
              //2̏
              iRow1 := iRow1 + 1;
            end;
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              sjis := pAnsiChar1;
              StrListLoadFile.Add(WideString(sjis));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end;
          end
          //NONE
          else if IntReadReturn1 = RETURN_NONE then
          begin
            //Ō
            if iRow1 = iEnd1-1 then
            begin
              //oCg
              iCopyCnt := iRow1 - iStart1;
              //m
              pAnsiChar1 := AllocMem(iCopyCnt+1);
              //]
              for iRow2 := 0 to iCopyCnt-1 do
              begin
                pAnsiChar1[iRow2] := pBuff1[iStart1+iRow2];
              end;
              pAnsiChar1[iCopyCnt] := #00;
              //ǉ
              sjis := pAnsiChar1;
              StrListLoadFile.Add(WideString(sjis));
              //j
              FreeMem(pAnsiChar1, iCopyCnt+1);
              //1̏
              iRow1 := iRow1 + 1;
              //Jnʒu
              istart1 := iRow1;
            end;
          end;
        end;
      end;
    except
      //ReadOnly J 103 ̃G[ɂȂ΍
    end;
  finally
    FreeMem(pBuff1);
  end;

end;

procedure  TFormViewGrid.ConvertLfToTab(Sender: TObject);
var
  iRow1: Integer;
  sStr1: String;
begin
  //ϊ
  StrListLoadCsvFile.Clear;
  if MenuLfToTab.Checked = TRUE then
  begin
    for iRow1 := 0 to StrListLoadFile.Count-1 do
    begin
      sStr1 := StringReplace(StrListLoadFile.Strings[iRow1], #10, #9, [rfReplaceAll]);
      StrListLoadCsvFile.Add(sStr1);
    end;
  end
  else
  begin
    for iRow1 := 0 to StrListLoadFile.Count-1 do
    begin
      StrListLoadCsvFile.Add(StrListLoadFile.Strings[iRow1]);
    end;
  end;
end;

function  TFormViewGrid.GetSeparater(SrcString: String;Sender: TObject): String;
var
  iPos1: Integer;
  iPos2: Integer;
begin
  //^ũf[^
  iPos1 := Pos(#09, SrcString);
  //dp̃f[^
  iPos2 := Pos('"', SrcString);
  if iPos1 > 0 then
  begin
    Result := SEPARATER_TAB;
  end
  //Đ擪ɂ dp̃f[^
  else if iPos2 = 1 then
  begin
    Result := SEPARATER_DOUBLE_CAUTATION;
  end
  else
  begin
    Result := SEPARATER_COMMA;
  end;
end;

procedure TFormViewGrid.StrListToStringGrid(StrList: TStringList; Sender: TObject);
var
  iRow1: Integer;
  iCol1: Integer;
  iMax1: Integer;
  CsvList1: TStringList;
  sStr2: String;
  CanSelect: Boolean;
begin
  //m
  CsvList1 := TStringList.Create;

  //s
  iMax1 := 0;
  for iRow1 := 0 to StrList.Count-1 do
  begin
    //NA
    CsvList1.Clear;

    //WJ
    if      IntStrToCsvIndex = 0 then
    begin
      StringToCsvList(StrList.Strings[iRow1], CsvList1);
    end
    else if IntStrToCsvIndex = 1 then
    begin
      StringToCsvListEx1(StrList.Strings[iRow1], CsvList1);
    end
    else if IntStrToCsvIndex = 2 then
    begin
      StringToCsvListEx2(StrList.Strings[iRow1], CsvList1);
    end
    else
    begin
      StringToCsvList(StrList.Strings[iRow1], CsvList1);
    end;

    //s
    StringGrid1.Cells[0, iRow1+2] := Format('%.05d', [iRow1+1]);
    StringGrid1.Cells[1, iRow1+2] := Format('%.05d', [iRow1+1]) + ':' + IntToStr(CsvList1.Count);
    //ڐ
    if CsvList1.Count > iMax1 then
    begin
      iMax1 := CsvList1.Count;
    end;

    //e
    for iCol1 := 0 to CsvList1.Count-1 do
    begin
      //#10 ƕҏWJnɉʏŉssȂ
      sStr2 := StringReplace(CsvList1[iCol1], #10, #13#10, [rfReplaceAll]);
      StringGrid1.Cells[iCol1+2, iRow1+2] := sStr2;
    end;
  end;

  //ݒ
  for iRow1 := 2 to  StringGrid1.RowCount-1 do
  begin
    StringGrid1.RowHeights[iRow1] := IntTextHeight*1;
  end;

  //j
  CsvList1.Free;

  //
  for iCol1 := 0 to iMax1 do
  begin
    StringGrid1.Cells[iCol1+2, 0] := IntToStr(iCol1+1);
  end;

  //ݒ
  StringGrid1.RowCount := StrList.Count+2;
  StringGrid1.ColCount := iMax1+2;

  //ݒ
  FixedColumnWidthAdjust(Sender);

  //ݒ
  StringGrid1.Row := 2;
  StringGrid1.Col := 2;
  StringGrid1SelectCell(Sender, StringGrid1.Col, StringGrid1.Row, CanSelect);

end;

procedure TFormViewGrid.SetCsvFile(Sender: TObject);
begin
  //؂蕶I
  if StrListLoadFile.Count > 0 then
  begin
    StrSeparater := GetSeparater(StrListLoadFile.Strings[0], Sender);
  end
  else
  begin
    StrSeparater := SEPARATER_UNKNOWN;
  end;
  //XV
  StatusBar1.Panels.Items[3].Text := StrSeparater;
  StatusBar2.Panels.Items[3].Text := StrSeparater;

  //LfTABɕϊ
  ConvertLfToTab(Sender);

  //\
  StrListToStringGrid(StrListLoadCsvFile, Sender);

  //ݒ
  MenuRatio1Click(Sender);

  //sɉs
  if BoolLfInLine = TRUE then
  begin
    Timer4.Enabled := TRUE;
  end;

end;

procedure TFormViewGrid.MenuFontColorClick(Sender: TObject);
begin
  if FontDialog1.Execute = TRUE then
  begin
    //ݒ
    StringGrid1.Canvas.Font.Name  := FontDialog1.Font.Name;
    StringGrid1.Canvas.Font.Color := FontDialog1.Font.Color;
    StringGrid1.Canvas.Font.Size  := FontDialog1.Font.Size;
    StringGrid1.Font.Name  := FontDialog1.Font.Name;
    StringGrid1.Font.Color := FontDialog1.Font.Color;
    StringGrid1.Font.Size  := FontDialog1.Font.Size;

    Memo1.Font.Name  := FontDialog1.Font.Name;
    Memo1.Font.Color := FontDialog1.Font.Color;
    Memo1.Font.Size  := FontDialog1.Font.Size;

    //ݒ
    IntTextHeight := TEXT_HEIGHT * StringGrid1.Canvas.Font.Size div TEXT_SIZE;
    StringGrid1.DefaultRowHeight := IntTextHeight;
    Memo1.Height  := (IntTextHeight * IntMemoHeightIndex) + (IntTextHeight * 2);
    Panel3.Height := (IntTextHeight * IntMemoHeightIndex) + (IntTextHeight * 2);

    //ݒ
    FixedColumnWidthAdjust(Sender);

    //ݒ
    if MenuRatio1.Checked = TRUE then
    begin
      MenuRatio1Click(Sender);
    end;
    if MenuRatio2.Checked = TRUE then
    begin
      MenuRatio2Click(Sender);
    end;
    if MenuRatio3.Checked = TRUE then
    begin
      MenuRatio3Click(Sender);
    end;
    if MenuRatio4.Checked = TRUE then
    begin
      MenuRatio4Click(Sender);
    end;
    if MenuRatio5.Checked = TRUE then
    begin
      MenuRatio5Click(Sender);
    end;
    if MenuRatio6.Checked = TRUE then
    begin
      MenuRatio6Click(Sender);
    end;
    if MenuRatio7.Checked = TRUE then
    begin
      MenuRatio7Click(Sender);
    end;
    if MenuRatio8.Checked = TRUE then
    begin
      MenuRatio8Click(Sender);
    end;
    if MenuRatio9.Checked = TRUE then
    begin
      MenuRatio9Click(Sender);
    end;
    if MenuRatio10.Checked = TRUE then
    begin
      MenuRatio10Click(Sender);
    end;
  end;
end;

procedure TFormViewGrid.MenuStatusColRowClick(Sender: TObject);
begin
  //ݒ
  MenuStatusRowCol.Checked := FALSE;
  MenuStatusColRow.Checked := TRUE;
  //XV
  StatusBar1.Panels.Items[0].Text := '[' + IntToStr(Stringgrid1.Col-1) + ',' + IntToStr(Stringgrid1.Row-1) + ']';
  StatusBar2.Panels.Items[0].Text := '[' + IntToStr(Stringgrid1.Col-1) + ',' + IntToStr(Stringgrid1.Row-1) + ']';
end;

procedure TFormViewGrid.MenuStatusRowColClick(Sender: TObject);
begin
  //ݒ
  MenuStatusColRow.Checked := FALSE;
  MenuStatusRowCol.Checked := TRUE;
  //XV
  StatusBar1.Panels.Items[0].Text := '[' + IntToStr(Stringgrid1.Row-1) + ',' + IntToStr(Stringgrid1.Col-1) + ']';
  StatusBar2.Panels.Items[0].Text := '[' + IntToStr(Stringgrid1.Row-1) + ',' + IntToStr(Stringgrid1.Col-1) + ']';
end;

procedure TFormViewGrid.MenuSameAsAnnsViewClick(Sender: TObject);
var
  sStr3: String;
begin
  //ݒ
  if MenuSameAsAnnsView.Checked = FALSE then
  begin
    MenuSameAsAnnsView.Checked := TRUE;
  end
  else
  begin
    MenuSameAsAnnsView.Checked := FALSE;
  end;

  //mF
  if FormMain.StrAnnsFMPath = '' then
  begin
    sStr3 := ExtractFilePath(ParamStr(0)) + ANNSFM_INI;
  end
  else
  begin
    sStr3 := FormMain.StrAnnsFMPath + ANNSFM_INI;
    if FileExists(sStr3) = FALSE then
    begin
      sStr3 := ExtractFilePath(ParamStr(0)) + ANNSFM_INI;
    end;
  end;
  if FileExists(sStr3) = TRUE then
  begin
    //
  end
  else
  begin
    MenuSameAsAnnsView.Checked := FALSE;
    //ݒ
    if BoolJapanese = TRUE then
    begin
      FormMain.StrMsg := 'Information : annsFM.ini ̃t@ĈŃ`FbN͖ł' + #13#10 +
                         sStr3;
    end
    else
    begin
      FormMain.StrMsg := 'Information : Checked can not accept, because of annsFM.ini is not exists.' + #13#10 +
                         sStr3 + #13#10 +
                         '[OK]';
    end;
    ShowMessage(FormMain.StrMsg);
    Exit;
  end;

  //ݒ
  if MenuSameAsAnnsView.Checked = TRUE then
  begin
    //ݒ
    MenuFontColor.Enabled := FALSE;
    MenuBackColor.Enabled := FALSE;
    MenuFontColoInit.Enabled := FALSE;
  end
  else
  begin
    //ݒ
    MenuFontColor.Enabled := TRUE;
    MenuBackColor.Enabled := TRUE;
    MenuFontColoInit.Enabled := TRUE;
  end;

  //ݒ
  SetSameAsAnnsView(Sender);

  //ݒ
  Stringgrid1.Repaint;
  FixedColumnWidthAdjust(Sender);
end;

procedure TFormViewGrid.MenuSaveTabClick(Sender: TObject);
begin
  MenuSaveTab.Checked              := TRUE;
  MenuSaveDoubleCautation.Checked  := FALSE;
  MenuSaveComma.Checked            := FALSE;
  MenuSaveFixedLength.Checked      := FALSE;
  MenuSaveCsvEx2.Checked           := FALSE;
end;

procedure TFormViewGrid.MenuSaveDoubleCautationClick(Sender: TObject);
begin
  MenuSaveTab.Checked              := FALSE;
  MenuSaveDoubleCautation.Checked  := TRUE;
  MenuSaveComma.Checked            := FALSE;
  MenuSaveFixedLength.Checked      := FALSE;
  MenuSaveCsvEx2.Checked           := FALSE;
end;

procedure TFormViewGrid.MenuSaveCommaClick(Sender: TObject);
begin
  MenuSaveTab.Checked              := FALSE;
  MenuSaveDoubleCautation.Checked  := FALSE;
  MenuSaveComma.Checked            := TRUE;
  MenuSaveFixedLength.Checked      := FALSE;
  MenuSaveCsvEx2.Checked           := FALSE;
end;

procedure TFormViewGrid.MenuSaveFixedLengthClick(Sender: TObject);
begin
  MenuSaveTab.Checked              := FALSE;
  MenuSaveDoubleCautation.Checked  := FALSE;
  MenuSaveComma.Checked            := FALSE;
  MenuSaveFixedLength.Checked      := TRUE;
  MenuSaveCsvEx2.Checked           := FALSE;
end;

procedure TFormViewGrid.MenuSaveCsvEx2Click(Sender: TObject);
begin
  MenuSaveTab.Checked              := FALSE;
  MenuSaveDoubleCautation.Checked  := FALSE;
  MenuSaveComma.Checked            := FALSE;
  MenuSaveFixedLength.Checked      := FALSE;
  MenuSaveCsvEx2.Checked           := TRUE;
end;

procedure TFormViewGrid.MenuSaveSJISClick(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_SJIS;
  end
  else
  begin
    StrSaveChar := CHAR_SJIS_ENG;
  end;
  MenuSaveSJIS.Checked          := TRUE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveJISClick(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_JIS;
  end
  else
  begin
    StrSaveChar := CHAR_JIS_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := TRUE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveEUCJPClick(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_EUCJP;
  end
  else
  begin
    StrSaveChar := CHAR_EUCJP_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := TRUE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveUTF7Click(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_UTF7;
  end
  else
  begin
    StrSaveChar := CHAR_UTF7_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := TRUE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveUTF8BomClick(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_UTF8_BOM_ON;
  end
  else
  begin
    StrSaveChar := CHAR_UTF8_BOM_ON_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := TRUE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveUTF8Click(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_UTF8_BOM_OFF;
  end
  else
  begin
    StrSaveChar := CHAR_UTF8_BOM_OFF_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := TRUE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveUTF16BomBeClick(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_UTF16_BOM_ON_BE;
  end
  else
  begin
    StrSaveChar := CHAR_UTF16_BOM_ON_BE_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := TRUE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveUTF16BeClick(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_UTF16_BOM_OFF_BE;
  end
  else
  begin
    StrSaveChar := CHAR_UTF16_BOM_OFF_BE_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := TRUE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveUTF16BomLeClick(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_UTF16_BOM_ON_LE;
  end
  else
  begin
    StrSaveChar := CHAR_UTF16_BOM_ON_LE_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := TRUE;
  MenuSaveUTF16Le.Checked       := FALSE;
end;

procedure TFormViewGrid.MenuSaveUTF16LeClick(Sender: TObject);
begin
  if BoolJapanese = TRUE then
  begin
    StrSaveChar := CHAR_UTF16_BOM_OFF_LE;
  end
  else
  begin
    StrSaveChar := CHAR_UTF16_BOM_OFF_LE_ENG;
  end;
  MenuSaveSJIS.Checked          := FALSE;
  MenuSaveJIS.Checked           := FALSE;
  MenuSaveEUCJP.Checked         := FALSE;
  MenuSaveUTF7.Checked          := FALSE;
  MenuSaveUTF8Bom.Checked       := FALSE;
  MenuSaveUTF8.Checked          := FALSE;
  MenuSaveUTF16BomBe.Checked    := FALSE;
  MenuSaveUTF16Be.Checked       := FALSE;
  MenuSaveUTF16BomLe.Checked    := FALSE;
  MenuSaveUTF16Le.Checked       := TRUE;
end;

procedure TFormViewGrid.MenuSearchWordNextClick(Sender: TObject);
begin
  //
  if Memo1.Focused = TRUE then
  begin
    ButtonMemoNextClick(Sender);
  end
  //ȊO
  else
  begin
    ButtonGridNextClick(Sender);
  end;
end;

procedure TFormViewGrid.MenuSearchWordPrevClick(Sender: TObject);
begin
  //
  if Memo1.Focused = TRUE then
  begin
    ButtonMemoPrevClick(Sender);
  end
  //ȊO
  else
  begin
    ButtonGridPrevClick(Sender);
  end;
end;

procedure TFormViewGrid.StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
var
  bModified: Boolean;
begin
  //ҏW
  if MenuEditMode.Checked = TRUE then
  begin
    //OƈقȂ
    if (ACol <> IntLastCol) or (ARow <> IntLastRow)  then
    begin
      //Oeݒ
      StringGrid1.Cells[IntLastCol, IntLastRow] := Memo1.Text;
    end;
    //擾
    bModified := Memo1.Modified;
  end
  else
  begin
    bModified := FALSE;
  end;
  //Đݒ肷
  if (BoolEditModified = FALSE) and (bModified = TRUE) then
  begin
    BoolEditModified := TRUE;
  end;

  //Veݒ
  Memo1.Text := StringGrid1.Cells[ACol, ARow];

  //Zbg
  Memo1.Modified := FALSE;
  //XV
  MemoStatus(Sender);

  //L
  IntLastCol := ACol;
  IntLastRow := ARow;

  //XV
  if MenuStatusColRow.Checked = TRUE then
  begin
    StatusBar1.Panels.Items[0].Text := '[' + IntToStr(ACol-1) + ',' + IntToStr(ARow-1) + ']';
    StatusBar2.Panels.Items[0].Text := '[' + IntToStr(ACol-1) + ',' + IntToStr(ARow-1) + ']';
  end;
  if MenuStatusRowCol.Checked = TRUE then
  begin
    StatusBar1.Panels.Items[0].Text := '[' + IntToStr(ARow-1) + ',' + IntToStr(ACol-1) + ']';
    StatusBar2.Panels.Items[0].Text := '[' + IntToStr(ARow-1) + ',' + IntToStr(ACol-1) + ']';
  end;
end;

procedure TFormViewGrid.MenuBackColorClick(Sender: TObject);
begin
  if ColorDialog1.Execute = TRUE then
  begin
    StringGrid1.Color := ColorDialog1.Color;
    Memo1.Color       := ColorDialog1.Color;
  end;
end;

procedure TFormViewGrid.MenuFixedBackColorClick(Sender: TObject);
begin
  if ColorDialog2.Execute = TRUE then
  begin
    StringGrid1.FixedBackColor := ColorDialog2.Color;
    StringGrid1.Repaint;
  end;
end;

procedure TFormViewGrid.MenuCurrentBackColorClick(Sender: TObject);
begin
  if ColorDialog3.Execute = TRUE then
  begin
    StringGrid1.CurrentBackColor := ColorDialog3.Color;
    StringGrid1.Repaint;
  end;
end;

procedure TFormViewGrid.MenuCrLfColorClick(Sender: TObject);
begin
  if ColorDialog4.Execute = TRUE then
  begin
    Memo1.CrLfColor := ColorDialog4.Color;
  end;
end;

procedure TFormViewGrid.MenuFontColoInitClick(Sender: TObject);
begin
    FontDialog1.Font.Color        := clBlack;
    FontDialog1.Font.Name         := 'lr SVbN';
    FontDialog1.Font.Size         := 14;
    ColorDialog1.Color            := clWhite;
    ColorDialog2.Color            := clSkyBlue;
    ColorDialog3.Color            := $00FF8000;
    ColorDialog4.Color            := clBlue;

    StringGrid1.Canvas.Font.Name  := FontDialog1.Font.Name;
    StringGrid1.Canvas.Font.Color := FontDialog1.Font.Color;
    StringGrid1.Canvas.Font.Size  := FontDialog1.Font.Size;
    StringGrid1.Font.Name  := FontDialog1.Font.Name;
    StringGrid1.Font.Color := FontDialog1.Font.Color;
    StringGrid1.Font.Size  := FontDialog1.Font.Size;
    StringGrid1.Color      := ColorDialog1.Color;
    StringGrid1.CurrentBackColor := ColorDialog3.Color;
    StringGrid1.FixedBackColor   := ColorDialog2.Color;
    StringGrid1.Repaint;

    Memo1.Font.Name  := FontDialog1.Font.Name;
    Memo1.Font.Color := FontDialog1.Font.Color;
    Memo1.Font.Size  := FontDialog1.Font.Size;
    Memo1.Color      := ColorDialog1.Color;
    Memo1.CrLfColor  := ColorDialog4.Color;

  //ݒ
  IntTextHeight := TEXT_HEIGHT * StringGrid1.Canvas.Font.Size div TEXT_SIZE;
  StringGrid1.DefaultRowHeight := IntTextHeight;

  //ݒ
  FixedColumnWidthAdjust(Sender);

  //ݒ
  if MenuRatio1.Checked = TRUE then
  begin
    MenuRatio1Click(Sender);
  end;
  if MenuRatio2.Checked = TRUE then
  begin
    MenuRatio2Click(Sender);
  end;
  if MenuRatio3.Checked = TRUE then
  begin
    MenuRatio3Click(Sender);
  end;
  if MenuRatio4.Checked = TRUE then
  begin
    MenuRatio4Click(Sender);
  end;
  if MenuRatio5.Checked = TRUE then
  begin
    MenuRatio5Click(Sender);
  end;
  if MenuRatio6.Checked = TRUE then
  begin
    MenuRatio6Click(Sender);
  end;
  if MenuRatio7.Checked = TRUE then
  begin
    MenuRatio7Click(Sender);
  end;
  if MenuRatio8.Checked = TRUE then
  begin
    MenuRatio8Click(Sender);
  end;
  if MenuRatio9.Checked = TRUE then
  begin
    MenuRatio9Click(Sender);
  end;
  if MenuRatio10.Checked = TRUE then
  begin
    MenuRatio10Click(Sender);
  end;
end;

procedure TFormViewGrid.FixedColumnWidthAdjust(Sender: TObject);
var
  iRow1: Integer;
  iCol1: Integer;
  iWidth1: Integer;
  iMax1: Integer;
begin
  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 0, 0);

  //s
  for iCol1 := 0 to 1 do
  begin
    iMax1 := 0;
    for iRow1 := 0 to StringGrid1.RowCount do
    begin
      //擾
      iWidth1 := StringGrid1.Canvas.TextWidth(StringGrid1.Cells[iCol1, iRow1]) + 5;
      if iWidth1 > iMax1 then
      begin
        iMax1 := iWidth1;
      end;
    end;
    //ݒ
    StringGrid1.ColWidths[iCol1]:= iMax1;
  end;

  //`搧
  SendMessage(StringGrid1.Handle, WM_SETREDRAW, 1, 0);
  StringGrid1.Repaint;
end;

procedure TFormViewGrid.SetShortCutKey(Sender: TObject);
var
  sStr1: String;
begin
  //annsfm
  if FormMain.IntKeyMenuTypeIndex = KEYBORD_MENU_ANNSFM then
  begin
    //Ǎ
    sStr1 := ExtractFilePath(ParamStr(0)) + KEYMAP_ANNSFM_LIST;
    if FileExists(sStr1) = TRUE then
    begin
      StrListKeyMap.LoadFromFile(sStr1, TEncoding.Unicode);
    end;

    //ݒ
    if StrListKeyMap.Count > 0 then
    begin
      //
    end;
  end;
end;

procedure TFormViewGrid.SetShortCutNew(Number: String; MenuItemName: TMenuItem; Sender: TObject);
var
  sStr3: String;
  sStr5: String;
  sStr6: String;
  sStr7: String;
  sStr8: String;
  sStr9: String;
  sStr11: String;
  sStr12: String;
  sStr13: String;
  sStr14: String;
  sStr15: String;
  sStr16: String;
  sStr17: String;
  iRow1: Integer;
  iFind1: Integer;
  iPos3: Integer;
  iPos4: Integer;
  iPos5: Integer;
  iPos11: Integer;
  iPos12: Integer;
  iPos13: Integer;
  Shift: TShiftState;
  Key: Word;
begin
  //ԍ̃V[gJbgݒ
  iFind1 := -1;
  for iRow1 := 0 to StrListKeyMap.Count-1 do
  begin
    //擾
    sStr11 := StrListKeyMap.Strings[iRow1];
    iPos11 := Pos(#9, sStr11);
    sStr12 := Copy(sStr11, 1,        iPos11-1);      //t@C
    sStr13 := Copy(sStr11, iPos11+1, Length(sStr11));//ȍ~̕
    iPos12 := Pos(#9, sStr13);
    sStr14 := Copy(sStr13, 1,        iPos12-1);      //VK쐬
    sStr15 := Copy(sStr13, iPos12+1, Length(sStr13));//ȍ~̕
    iPos13 := Pos(#9, sStr15);
    sStr16 := Copy(sStr15, 1,        iPos13-1);      //Ctrl+O
    sStr17 := Copy(sStr15, iPos13+1, Length(sStr15));//1001
    //ԍ
    if sStr17 = Number then
    begin
      iFind1 := iRow1;
      Break;
    end;
  end;
  if iFind1 = -1 then
  begin
    //ݒ
    sStr3 := '';
  end
  else
  begin
    //ݒ
    sStr3 := sStr16;
  end;

  //
  iPos3 := Pos('Shift',  sStr3);
  iPos4 := Pos('Ctrl',   sStr3);
  iPos5 := Pos('Alt',    sStr3);
  sStr5 := StringReplace(sStr3, 'Shift', '', [rfReplaceAll]);
  sStr6 := StringReplace(sStr5, 'Ctrl',  '', [rfReplaceAll]);
  sStr7 := StringReplace(sStr6, 'Alt',   '', [rfReplaceAll]);
  sStr8 := StringReplace(sStr7, '+',     '', [rfReplaceAll]);
  sStr9 := Trim         (sStr8);
  //L[
  if      sStr9 = '0' then
  begin
    Key := VK_0;
  end
  else if sStr9 = '1' then
  begin
    Key := VK_1;
  end
  else if sStr9 = '2' then
  begin
    Key := VK_2;
  end
  else if sStr9 = '3' then
  begin
    Key := VK_3;
  end
  else if sStr9 = '4' then
  begin
    Key := VK_4;
  end
  else if sStr9 = '5' then
  begin
    Key := VK_5;
  end
  else if sStr9 = '6' then
  begin
    Key := VK_6;
  end
  else if sStr9 = '7' then
  begin
    Key := VK_7;
  end
  else if sStr9 = '8' then
  begin
    Key := VK_8;
  end
  else if sStr9 = '9' then
  begin
    Key := VK_9;
  end

  else if sStr9 = 'A' then
  begin
    Key := VK_A;
  end
  else if sStr9 = 'B' then
  begin
    Key := VK_B;
  end
  else if sStr9 = 'C' then
  begin
    Key := VK_C;
  end
  else if sStr9 = 'D' then
  begin
    Key := VK_D;
  end
  else if sStr9 = 'E' then
  begin
    Key := VK_E;
  end
  else if sStr9 = 'F' then
  begin
    Key := VK_F;
  end
  else if sStr9 = 'G' then
  begin
    Key := VK_G;
  end
  else if sStr9 = 'H' then
  begin
    Key := VK_H;
  end
  else if sStr9 = 'I' then
  begin
    Key := VK_I;
  end
  else if sStr9 = 'J' then
  begin
    Key := VK_J;
  end
  else if sStr9 = 'K' then
  begin
    Key := VK_K;
  end
  else if sStr9 = 'L' then
  begin
    Key := VK_L;
  end
  else if sStr9 = 'M' then
  begin
    Key := VK_M;
  end
  else if sStr9 = 'N' then
  begin
    Key := VK_N;
  end
  else if sStr9 = 'O' then
  begin
    Key := VK_O;
  end
  else if sStr9 = 'P' then
  begin
    Key := VK_P;
  end
  else if sStr9 = 'Q' then
  begin
    Key := VK_Q;
  end
  else if sStr9 = 'R' then
  begin
    Key := VK_R;
  end
  else if sStr9 = 'S' then
  begin
    Key := VK_S;
  end
  else if sStr9 = 'T' then
  begin
    Key := VK_T;
  end
  else if sStr9 = 'U' then
  begin
    Key := VK_U;
  end
  else if sStr9 = 'V' then
  begin
    Key := VK_V;
  end
  else if sStr9 = 'W' then
  begin
    Key := VK_W;
  end
  else if sStr9 = 'X' then
  begin
    Key := VK_X;
  end
  else if sStr9 = 'Y' then
  begin
    Key := VK_Y;
  end
  else if sStr9 = 'Z' then
  begin
    Key := VK_Z;
  end

  else if sStr9 = ',' then
  begin
    Key := VK_COMMA;
  end
  else if sStr9 = '.' then
  begin
    Key := VK_PIRIOD ;
  end
  else if sStr9 = '/' then
  begin
    Key := VK_SLASH;
  end
  else if sStr9 = ';' then
  begin
    Key := VK_SEMICOLLON;
  end
  else if sStr9 = ':' then
  begin
    Key := VK_COLLON;
  end
  else if sStr9 = ']' then
  begin
    Key := VK_RIGHT_KAGI;
  end
  else if sStr9 = '@' then
  begin
    Key := VK_AT_MARK;
  end
  else if sStr9 = '[' then
  begin
    Key := VK_LEFT_KAGI;
  end
  else if sStr9 = '-' then
  begin
    Key := VK_MINUS;
  end
  else if sStr9 = '^' then
  begin
    Key := VK_HATTO;
  end
  else if sStr9 = '\' then
  begin
    Key := VK_YEN;
  end

  else if sStr9 = 'F1' then
  begin
    Key := VK_F1;
  end
  else if sStr9 = 'F2' then
  begin
    Key := VK_F2;
  end
  else if sStr9 = 'F3' then
  begin
    Key := VK_F3;
  end
  else if sStr9 = 'F4' then
  begin
    Key := VK_F4;
  end
  else if sStr9 = 'F5' then
  begin
    Key := VK_F5;
  end
  else if sStr9 = 'F6' then
  begin
    Key := VK_F6;
  end
  else if sStr9 = 'F7' then
  begin
    Key := VK_F7;
  end
  else if sStr9 = 'F8' then
  begin
    Key := VK_F8;
  end
  else if sStr9 = 'F9' then
  begin
    Key := VK_F9;
  end
  else if sStr9 = 'F10' then
  begin
    Key := VK_F10;
  end
  else if sStr9 = 'F11' then
  begin
    Key := VK_F11;
  end
  else if sStr9 = 'F12' then
  begin
    Key := VK_F12;
  end

  else if sStr9 = 'Space' then
  begin
    Key := VK_SPACE;
  end
  else if sStr9 = 'Esc' then
  begin
    Key := VK_ESCAPE;
  end
  else if sStr9 = 'Enter' then
  begin
    Key := VK_RETURN;
  end
  else if sStr9 = 'Tab' then
  begin
    Key := VK_TAB;
  end
  else if sStr9 = 'Up' then
  begin
    Key := VK_UP;
  end
  else if sStr9 = 'Down' then
  begin
    Key := VK_DOWN;
  end
  else if sStr9 = 'Left' then
  begin
    Key := VK_LEFT;
  end
  else if sStr9 = 'Right' then
  begin
    Key := VK_RIGHT;
  end
  else if sStr9 = 'Home' then
  begin
    Key := VK_HOME;
  end
  else if sStr9 = 'End' then
  begin
    Key := VK_END;
  end
  else if sStr9 = 'PgUp' then
  begin
    Key := VK_PRIOR;
  end
  else if sStr9 = 'PgDn' then
  begin
    Key := VK_NEXT;
  end
  else if sStr9 = 'Ins' then
  begin
    Key := VK_INSERT;
  end
  else if sStr9 = 'Del' then
  begin
    Key := VK_DELETE;
  end
  else if sStr9 = 'BkSp' then
  begin
    Key := VK_BACK;
  end

  else
  begin
    Key := 0;
  end;

  //L[
  if      (iPos3 = 0) and (iPos4 = 0) and (iPos5 = 0) then
  begin
    Shift := [];
  end
  else if (iPos3 > 0) and (iPos4 = 0) and (iPos5 = 0) then
  begin
    Shift := [ssShift];
  end
  else if (iPos3 = 0) and (iPos4 > 0) and (iPos5 = 0) then
  begin
    Shift := [ssCtrl];
  end
  else if (iPos3 > 0) and (iPos4 > 0) and (iPos5 = 0) then
  begin
    Shift := [ssShift, ssCtrl];
  end
  else if (iPos3 = 0) and (iPos4 = 0) and (iPos5 > 0) then
  begin
    Shift := [ssAlt];
  end
  else if (iPos3 > 0) and (iPos4 = 0) and (iPos5 > 0) then
  begin
    Shift := [ssShift, ssAlt];
  end
  else if (iPos3 = 0) and (iPos4 > 0) and (iPos5 > 0) then
  begin
    Shift := [ssCtrl, ssAlt];
  end
  else if (iPos3 > 0) and (iPos4 > 0) and (iPos5 > 0) then
  begin
    Shift := [ssShift, ssCtrl, ssAlt];
  end
  else
  begin
  end;

  //ݒ
  MenuItemName.ShortCut := ShortCut(Key, Shift);

end;

procedure TFormViewGrid.SaveCsvFile(FileName: String; Sender: TObject);
var
  sStr8: String;
  wRet8: Word;
  iRow1: Integer;
  iCol1: Integer;
  iLen1: Integer;
  sStr1: String;
  sStr2: String;
  sStr3: String;
  sStr4: String;
  sStr10: String;
  CsvMax: array [0..CSVLIST_ITEM_MAX] of Integer;
begin

  if MenuLfToTab.Checked = TRUE then
  begin
    //mF
    if BoolJapanese = TRUE then
    begin
      sStr8 := 'Information : Z̉s^uɒuw肳Ă܂' + #13#10 +
               '㏑ۑ܂H';
    end
    else
    begin
      sStr8 := 'Information : Lf is already replaced to TAB.' + #13#10 +
               'Over write  file ?' + #13#10 +
               '[YES] / [NO]';
    end;
    wRet8 := MessageDlg(sStr8, mtInformation, [mbYes, mbNo ], 0);
    if wRet8 = mrYes then
    begin
      //ۑ
    end
    else
    begin
      //Ȃ
      Exit;
    end;
  end;

  //mF
  if BoolJapanese = TRUE then
  begin
    if MenuSaveTab.Checked = TRUE then
    begin
      sStr3 := SEPARATER_TAB;
    end;
    if MenuSaveDoubleCautation.Checked = TRUE then
    begin
      sStr3 := SEPARATER_DOUBLE_CAUTATION;
    end;
    if MenuSaveComma.Checked = TRUE then
    begin
      sStr3 := SEPARATER_COMMA;
    end;
    if MenuSaveFixedLength.Checked = TRUE then
    begin
      sStr3 := SEPARATER_FIXED;
    end;
    if MenuSaveCsvEx2.Checked = TRUE then
    begin
      sStr3 := SEPARATER_COMMA_DOUBLE;
    end;

    if MenuSaveSJIS.Checked = TRUE then
    begin
      sStr4 := CHAR_SJIS;
    end;
    if MenuSaveJIS.Checked = TRUE then
    begin
      sStr4 := CHAR_JIS;
    end;
    if MenuSaveEUCJP.Checked  = TRUE then
    begin
      sStr4 := CHAR_EUCJP;
    end;
    if MenuSaveUTF7.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF7;
    end;
    if MenuSaveUTF8Bom.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF8_BOM_ON;
    end;
    if MenuSaveUTF8.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF8_BOM_OFF;
    end;
    if  MenuSaveUTF16BomBe.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF16_BOM_ON_BE;
    end;
    if MenuSaveUTF16Be.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF16_BOM_OFF_BE;
    end;
    if MenuSaveUTF16BomLe.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF16_BOM_ON_LE;
    end;
    if MenuSaveUTF16Le.Checked  = TRUE then
    begin
      sStr4 := CHAR_UTF16_BOM_OFF_LE;
    end;

    sStr8 := 'Information : ȉ̓eŕۑ܂' + #13#10 +
             '    R[h : ' + sStr4              + #13#10 +
             '    s̉sR[h : CrLf[#13#10]'    + #13#10 +
             '    Z̉sR[h : Lf[#10]'       + #13#10 +
             '    ؂蕶 : ' + '[' + sStr3 + ']'  + #13#10 +
             'ۑ܂H';
  end
  else
  begin
    if MenuSaveTab.Checked = TRUE then
    begin
      sStr3 := SEPARATER_TAB_ENG;
    end;
    if MenuSaveDoubleCautation.Checked = TRUE then
    begin
      sStr3 := SEPARATER_DOUBLE_CAUTATION_ENG;
    end;
    if MenuSaveComma.Checked = TRUE then
    begin
      sStr3 := SEPARATER_COMMA_ENG;
    end;
    if MenuSaveFixedLength.Checked = TRUE then
    begin
      sStr3 := SEPARATER_FIXED_ENG;
    end;
    if MenuSaveCsvEx2.Checked = TRUE then
    begin
      sStr3 := SEPARATER_COMMA_DOUBLE_ENG;
    end;

    if MenuSaveSJIS.Checked = TRUE then
    begin
      sStr4 := CHAR_SJIS_ENG;
    end;
    if MenuSaveJIS.Checked = TRUE then
    begin
      sStr4 := CHAR_JIS_ENG;
    end;
    if MenuSaveEUCJP.Checked  = TRUE then
    begin
      sStr4 := CHAR_EUCJP_ENG;
    end;
    if MenuSaveUTF7.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF7_ENG;
    end;
    if MenuSaveUTF8Bom.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF8_BOM_ON_ENG;
    end;
    if MenuSaveUTF8.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF8_BOM_OFF_ENG;
    end;
    if  MenuSaveUTF16BomBe.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF16_BOM_ON_BE_ENG;
    end;
    if MenuSaveUTF16Be.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF16_BOM_OFF_BE_ENG;
    end;
    if MenuSaveUTF16BomLe.Checked = TRUE then
    begin
      sStr4 := CHAR_UTF16_BOM_ON_LE_ENG;
    end;
    if MenuSaveUTF16Le.Checked  = TRUE then
    begin
      sStr4 := CHAR_UTF16_BOM_OFF_LE_ENG;
    end;

    sStr8 := 'Information : File saved as follows.' + #13#10 +
             '    CharCode : ' + sStr4              + #13#10 +
             '    CrLf at Line End : CrLf[#13#10]'  + #13#10 +
             '    Lf in Cell : Lf[#10]'             + #13#10 +
             '    Separater : ' + '[' + sStr3 + ']' + #13#10 +
             'Save file ?'                          + #13#10 +
             '[YES] / [NO]';
  end;
  wRet8 := MessageDlg(sStr8, mtInformation, [mbYes, mbNo ], 0);
  if wRet8 = mrYes then
  begin
    //ۑ
  end
  else
  begin
    //Ȃ
    Exit;
  end;

  //NA
  StrListSaveFile.Clear;

  //Œ蒷
  if MenuSaveFixedLength.Checked = TRUE then
  begin
    //s
    for iCol1 := 2 to StringGrid1.ColCount-1 do
    begin
      //ݒ
      for iRow1 := 2 to StringGrid1.RowCount-1 do
      begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          iLen1 := Length(AnsiString(sStr2));
          if iLen1 > CsvMax[iCol1] then
          begin
            CsvMax[iCol1] := iLen1;
          end;
      end;
    end;
  end;

  //s
  for iRow1 := 2 to StringGrid1.RowCount-1 do
  begin
    //NA
    sStr1 := '';
    //ݒ
    for iCol1 := 2 to StringGrid1.ColCount-1 do
    begin
      //擪
      if iCol1 = 2 then
      begin
        if MenuSaveTab.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := sStr2 + #09;
        end;
        if MenuSaveDoubleCautation.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := '"' + sStr2 + '","';
        end;
        if MenuSaveComma.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := sStr2 + ',';
        end;
        if MenuSaveFixedLength.Checked = TRUE then
        begin
          sStr2  := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr10 := sStr2 + StringOfChar(' ', CsvMax[iCol1]-Length(AnsiString(sStr2)));
          sStr1  := sStr10 + SEPARATER_CHAR_FIXED;
        end;
        if MenuSaveCsvEx2.Checked = TRUE then
        begin
          sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iCol1, iRow1], Sender);
          sStr1 := sStr2 + ',';
        end;
      end
      //Ō
      else if iCol1 = StringGrid1.ColCount-1 then
      begin
        if MenuSaveTab.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[StringGrid1.ColCount-1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := sStr1 + sStr2;
        end;
        if MenuSaveDoubleCautation.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := sStr1 + sStr2 + '"';
        end;
        if MenuSaveComma.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := sStr1 + sStr2;
        end;
        if MenuSaveFixedLength.Checked = TRUE then
        begin
          sStr2  := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr10 := sStr2 + StringOfChar(' ', CsvMax[iCol1]-Length(AnsiString(sStr2)));
          sStr1  := sStr1 + sStr10;
        end;
        if MenuSaveCsvEx2.Checked = TRUE then
        begin
          sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iCol1, iRow1], Sender);
          sStr1 := sStr1 + sStr2;
        end;
      end
      //ŌȊO
      else
      begin
        if MenuSaveTab.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := sStr1 + sStr2 + #09;
        end;
        if MenuSaveDoubleCautation.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := sStr1 + sStr2 + '","';
        end;
        if MenuSaveComma.Checked = TRUE then
        begin
          sStr2 := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr1 := sStr1 + sStr2 + ',';
        end;
        if MenuSaveFixedLength.Checked = TRUE then
        begin
          sStr2  := StringReplace(StringGrid1.Cells[iCol1, iRow1], #13#10, #10, [rfReplaceAll]);
          sStr10 := sStr2 + StringOfChar(' ', CsvMax[iCol1]-Length(AnsiString(sStr2)));
          sStr1  := sStr1 + sStr10 + SEPARATER_CHAR_FIXED;
        end;
        if MenuSaveCsvEx2.Checked = TRUE then
        begin
          sStr2 := ConvertCellToCsvEx2(StringGrid1.Cells[iCol1, iRow1], Sender);
          sStr1 := sStr1 + sStr2 + ',';
        end;
      end;
    end;
    //ǉ
    StrListSaveFile.Add(sStr1);
  end;

  //ۑ
  if MenuSaveSJIS.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.SaveToFile(FileName, TEncoding.GetEncoding(932));
  end;
  if MenuSaveJIS.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.SaveToFile(FileName, TEncoding.GetEncoding(50220));
  end;
  if MenuSaveEUCJP.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.SaveToFile(FileName, TEncoding.GetEncoding(20932));
  end;
  if MenuSaveUTF7.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.SaveToFile(FileName, TEncoding.UTF7);
  end;
  if MenuSaveUTF8Bom.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.WriteBOM := TRUE;
    StrListSaveFile.SaveToFile(FileName, TEncoding.UTF8);
  end;
  if MenuSaveUTF8.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.WriteBOM := FALSE;
    StrListSaveFile.SaveToFile(FileName, TEncoding.UTF8);
  end;
  if MenuSaveUTF16BomBe.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.WriteBOM := TRUE;
    StrListSaveFile.SaveToFile(FileName, TEncoding.BigEndianUnicode);
  end;
  if MenuSaveUTF16Be.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.WriteBOM := FALSE;
    StrListSaveFile.SaveToFile(FileName, TEncoding.BigEndianUnicode);
  end;
  if MenuSaveUTF16BomLe.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.WriteBOM := TRUE;
    StrListSaveFile.SaveToFile(FileName, TEncoding.Unicode);
  end;
  if MenuSaveUTF16Le.Checked = TRUE then
  begin
    //ۑ
    StrListSaveFile.WriteBOM := FALSE;
    StrListSaveFile.SaveToFile(FileName, TEncoding.Unicode);
  end;
end;

procedure TFormViewGrid.ButtonGridNextClick(Sender: TObject);
var
  iRow1: Integer;
  iCol1: Integer;
  iStartRow1: Integer;
  iStartCol1: Integer;
  iFind1: Integer;
  iFindCol1: Integer;
  iFindRow1: Integer;
begin
    //
    iStartRow1 := StringGrid1.Row;
    iStartCol1 := StringGrid1.Col+1;
    iFind1 := -1;
    iFindCol1 := 2;
    iFindRow1 := 2;
    //
    for iRow1 := iStartRow1 to StringGrid1.RowCount-1 do
    begin
      for iCol1 := iStartCol1 to StringGrid1.ColCount-1 do
      begin
        //
        if Pos(UpperCase(Edit1.Text), UpperCase(StringGrid1.Cells[iCol1, iRow1])) > 0 then
        begin
          iFind1    := iCol1;
          iFindCol1 := iCol1;
          iFindRow1 := iRow1;
        end;
      end;
      if iFind1 = -1 then
      begin
        iStartCol1 := 2;
      end;
      if iFind1 <> -1 then
      begin
        Break;
      end;
    end;
    //
    if iFind1 <> -1 then
    begin
      StringGrid1.Row := iFindRow1;
      StringGrid1.Col := iFindCol1;
    end
    //
    else
    begin
      //ݒ
      if BoolJapanese = TRUE then
      begin
        FormMain.StrMsg := 'Information : ܂';
      end
      else
      begin
        FormMain.StrMsg := 'Information : Search word can not find.' + #13#10 +
                           '[OK]';
      end;
      ShowMessage(FormMain.StrMsg);
    end;
    //tH[JXړ
    StringGrid1.SetFocus;
end;

procedure TFormViewGrid.ButtonGridPrevClick(Sender: TObject);
var
  iRow1: Integer;
  iCol1: Integer;
  iStartRow1: Integer;
  iStartCol1: Integer;
  iFind1: Integer;
  iFindCol1: Integer;
  iFindRow1: Integer;
begin
    //
    iStartRow1 := StringGrid1.Row;
    iStartCol1 := StringGrid1.Col-1;
    iFind1 := -1;
    iFindCol1 := 2;
    iFindRow1 := 2;
    //
    for iRow1 := iStartRow1 downto 2 do
    begin
      for iCol1 := iStartCol1 downto 2 do
      begin
        //
        if Pos(UpperCase(Edit1.Text), UpperCase(StringGrid1.Cells[iCol1, iRow1])) > 0 then
        begin
          iFind1    := iCol1;
          iFindCol1 := iCol1;
          iFindRow1 := iRow1;
        end;
      end;
      if iFind1 = -1 then
      begin
        iStartCol1 := StringGrid1.ColCount-1;
      end;
      if iFind1 <> -1 then
      begin
        Break;
      end;
    end;
    //
    if iFind1 <> -1 then
    begin
      StringGrid1.Row := iFindRow1;
      StringGrid1.Col := iFindCol1;
    end
    //
    else
    begin
      //ݒ
      if BoolJapanese = TRUE then
      begin
        FormMain.StrMsg := 'Information : ܂';
      end
      else
      begin
        FormMain.StrMsg := 'Information : Search word can not find.' + #13#10 +
                           '[OK]';
      end;
      ShowMessage(FormMain.StrMsg);
    end;
    //tH[JXړ
    StringGrid1.SetFocus;
end;

procedure TFormViewGrid.ButtonMemoPrevClick(Sender: TObject);
var
  iFind1: Integer;
  iPos1: Integer;
  iStart1: Integer;
  sStr1: String;
  sStr2: String;
begin
    //Jnʒu
    iStart1 := Memo1.SelStart;
    //JnʒuȑO
    sStr1 := Copy(Memo1.Text, 1, iStart1);
    //
    iFind1 := 0;
    //
    iPos1 := Pos(UpperCase(Edit1.Text), UpperCase(sStr1));
    //
    if iPos1 > 0 then
    begin
      //ݒ
      iFind1 := iFind1 + iPos1 - 1;
      while iPos1 > 0 do
      begin
        //JnʒuȑO
        sStr2 := Copy(sStr1, iPos1+1, Length(sStr1));
        sStr1 := sStr2;
        //
        iPos1 := Pos(UpperCase(Edit1.Text), UpperCase(sStr1));
        //
        if iPos1 > 0 then
        begin
          //ݒ
          iFind1 := iFind1 + iPos1;
        end;
      end;
    end;
    //
    if iFind1 > 0 then
    begin
      //ݒ
      Memo1.SelStart  := iFind1;
      Memo1.SelLength := Length(Edit1.Text);
    end
    //
    else
    begin
      //ݒ
      if BoolJapanese = TRUE then
      begin
        FormMain.StrMsg := 'Information : ܂';
      end
      else
      begin
        FormMain.StrMsg := 'Information : Search word can not find.' + #13#10 +
                           '[OK]';
      end;
      ShowMessage(FormMain.StrMsg);
    end;
    //tH[JXړ
    Memo1.SetFocus;
end;

procedure TFormViewGrid.ButtonMemoNextClick(Sender: TObject);
var
  iPos1: Integer;
  iStart1: Integer;
  sStr1: String;
begin
    //Jnʒu
    iStart1 := Memo1.SelStart + Memo1.SelLength;
    //Jnʒuȍ~
    sStr1 := Copy(Memo1.Text, iStart1+1, Length(Memo1.Text));
    //
    iPos1 := Pos(UpperCase(Edit1.Text), UpperCase(sStr1));
    if iPos1 > 0 then
    begin
      //ݒ
      Memo1.SelStart := iStart1 + iPos1 - 1;
      Memo1.SelLength := Length(Edit1.Text);
    end
    else
    begin
      //ݒ
      if BoolJapanese = TRUE then
      begin
        FormMain.StrMsg := 'Information : ܂';
      end
      else
      begin
        FormMain.StrMsg := 'Information : Search word can not find.' + #13#10 +
                           '[OK]';
      end;
      ShowMessage(FormMain.StrMsg);
    end;
    //tH[JXړ
    Memo1.SetFocus;
end;

procedure TFormViewGrid.PopUpCut1Click(Sender: TObject);
begin
  ClipBoard.AsText := Memo1.SelText;
  Memo1.SelText := '';
  //XV
  MemoStatus(Sender);
end;

procedure TFormViewGrid.PopUpPaste1Click(Sender: TObject);
begin
  Memo1.PasteFromClipboard;
  //XV
  MemoStatus(Sender);
end;

procedure TFormViewGrid.PopUpCopy1Click(Sender: TObject);
begin
  ClipBoard.AsText := Memo1.SelText;
  //XV
  MemoStatus(Sender);
end;

procedure TFormViewGrid.PopUpSelectAll1Click(Sender: TObject);
begin
  Memo1.SelectAll;
  //XV
  MemoStatus(Sender);
end;

procedure TFormViewGrid.StringToCsvListEx1(SrcString: String; CsvList: TstringList);
var
  iPos1: Integer;
  iPos2: Integer;
  sStr1: String;
  sStr2: String;
  sStr3: String;
  sStr4: String;
  sStr5: String;
  iRow1: Integer;
  iRow2: Integer;
  iVal1: Integer;
  iVal2: Integer;
  iVal3: Integer;
  bFind1: Boolean;
  StrList1: TStringList;
  StrList2: TStringList;
  StrList3: TStringList;
  StrList4: TStringList;
begin
  //m
  StrList1 := TStringList.Create;
  StrList2 := TStringList.Create;
  StrList3 := TStringList.Create;
  StrList4 := TStringList.Create;

  //f[^̐ݒ
  //擪ɒǉ
  sStr1 := ',' + SrcString;

  //ŏɒu
  sStr2 := StringReplace(sStr1, '""', #$0C, [rfReplaceAll]);
  //
  bFind1 := FALSE;
  for iRow1 := 0 to Length(sStr2)-1 do
  begin
    if Copy(sStr2, iRow1+1, 1) = '"' then
    begin
      if bFind1 = FALSE then
      begin
        StrList1.Add(IntToStr(iRow1));
        bFind1 := TRUE;
      end
      else
      begin
        StrList2.Add(IntToStr(iRow1));
        bFind1 := FALSE;
      end;
    end
    else if Copy(sStr2, iRow1+1, 1) = ',' then
    begin
        StrList3.Add(IntToStr(iRow1));
    end
    else
    begin
      //
    end;
  end;

  //2dp̊ԂɂJ}͑ΏۊOƂ
  for iRow1 := 0 to StrList3.Count-1 do
  begin
    //ڐ
    if StrList1.Count = StrList2.Count then
    begin
      if StrList1.Count = 0 then
      begin
            StrList4.Add(StrList3.Strings[iRow1]);
      end
      else
      begin
        //ԍ
        iVal1 := StrToInt(StrList3.Strings[iRow1]);
        for iRow2 := 0 to StrList1.Count-1 do
        begin
          //ԍ
          iVal2 := StrToInt(StrList1.Strings[iRow2]);
          //ԍ
          iVal3 := StrToInt(StrList2.Strings[iRow2]);
          //͈͓
          if (iVal1 > iVal2) and (iVal1 < iVal3)  then
          begin
            //
          end
          else
          begin
            StrList4.Add(StrList3.Strings[iRow1]);
          end;
        end;
      end;
    end;
  end;
          StrList4.Add(IntToStr(Length(sStr2)));

//  StrList1.SaveToFile('strlist1.txt');
//  StrList2.SaveToFile('strlist2.txt');
//  StrList3.SaveToFile('strlist3.txt');
//  StrList4.SaveToFile('strlist4.txt');
//  showmessage('');


  //ɖ߂
  sStr3 := StringReplace(sStr2, #$0C, '"', [rfReplaceAll]);

  if StrList4.Count = 1 then
  begin
      //擾
      iPos1 := 0;
      iPos2 := StrToInt(StrList4.Strings[0]);
      sStr4 := Copy(sStr3, iPos1+2, iPos2 - iPos1 - 1);
      if Length(sStr4) > 0 then
      begin
        //ǉ
        CsvList.Add(sStr4);
      end;
  end
  else
  begin
    //ݒ
    for iRow1 := 0 to StrList4.Count-2 do
    begin
      //擾
      iPos1 := StrToInt(StrList4.Strings[iRow1+0]);
      iPos2 := StrToInt(StrList4.Strings[iRow1+1]);
      sStr4 := Copy(sStr3, iPos1+2, iPos2 - iPos1 - 1);
      if Length(sStr4) > 2 then
      begin
        if (Copy(sStr4, 1, 1) = '"') and (Copy(sStr4, Length(sStr4), 1) = '"') then
        begin
          sStr5 := Copy(sStr4, 2, Length(sStr4)-2);
        end
        else
        begin
          sStr5 := sStr4;
        end;
      end
      else
      begin
          sStr5 := sStr4;
      end;
      //ǉ
      CsvList.Add(sStr5);
    end;
  end;

  //j
  StrList1.Free;
  StrList2.Free;
  StrList3.Free;
  StrList4.Free;
end;

procedure TFormViewGrid.StringToCsvListEx2(SrcString: String; CsvList: TstringList);
var
  iPos1: Integer;
  iPos2: Integer;
  sStr1: String;
  sStr2: String;
  sStr3: String;
  sStr4: String;
  sStr5: String;
  iRow1: Integer;
  iRow2: Integer;
  iVal1: Integer;
  iVal2: Integer;
  iVal3: Integer;
  bFind1: Boolean;
  bFind2: Boolean;
  StrList1: TStringList;
  StrList2: TStringList;
  StrList3: TStringList;
  StrList4: TStringList;
begin
  //m
  StrList1 := TStringList.Create;
  StrList2 := TStringList.Create;
  StrList3 := TStringList.Create;
  StrList4 := TStringList.Create;

  //f[^̐ݒ
  //擪ɒǉ
  sStr1 := ',' + SrcString;

  //ŏɒu
  sStr2 := StringReplace(sStr1, '""', #$0C, [rfReplaceAll]);
  //
  bFind1 := FALSE;
  for iRow1 := 0 to Length(sStr2)-1 do
  begin
    if Copy(sStr2, iRow1+1, 1) = '"' then
    begin
      if bFind1 = FALSE then
      begin
        StrList1.Add(IntToStr(iRow1));
        bFind1 := TRUE;
      end
      else
      begin
        StrList2.Add(IntToStr(iRow1));
        bFind1 := FALSE;
      end;
    end
    else if Copy(sStr2, iRow1+1, 1) = ',' then
    begin
        StrList3.Add(IntToStr(iRow1));
    end
    else
    begin
      //
    end;
  end;

  //2dp̊ԂɂJ}͑ΏۊOƂ
  for iRow1 := 0 to StrList3.Count-1 do
  begin
    //ڐ
    if StrList1.Count = StrList2.Count then
    begin
      if StrList1.Count = 0 then
      begin
            StrList4.Add(StrList3.Strings[iRow1]);
      end
      else
      begin
        //ԍ
        iVal1 := StrToInt(StrList3.Strings[iRow1]);
        bFind2 := TRUE;
        for iRow2 := 0 to StrList1.Count-1 do
        begin
          //ԍ
          iVal2 := StrToInt(StrList1.Strings[iRow2]);
          //ԍ
          iVal3 := StrToInt(StrList2.Strings[iRow2]);
          //͈͓
          if (iVal1 > iVal2) and (iVal1 < iVal3)  then
          begin
            bFind2 := FALSE;
          end;
        end;
        if bFind2 = TRUE then
        begin
            StrList4.Add(StrList3.Strings[iRow1]);
        end;
      end;
    end;
  end;
          StrList4.Add(IntToStr(Length(sStr2)));

//  StrList1.SaveToFile('strlist1.txt');
//  StrList2.SaveToFile('strlist2.txt');
//  StrList3.SaveToFile('strlist3.txt');
//  StrList4.SaveToFile('strlist4.txt');
//  showmessage('');


  //ɖ߂
  sStr3 := StringReplace(sStr2, #$0C, '"', [rfReplaceAll]);

  if StrList4.Count = 1 then
  begin
      //擾
      iPos1 := 0;
      iPos2 := StrToInt(StrList4.Strings[0]);
      sStr4 := Copy(sStr3, iPos1+2, iPos2 - iPos1 - 1);
      if Length(sStr4) > 0 then
      begin
        //ǉ
        CsvList.Add(sStr4);
      end;
  end
  else
  begin
    //ݒ
    for iRow1 := 0 to StrList4.Count-2 do
    begin
      //擾
      iPos1 := StrToInt(StrList4.Strings[iRow1+0]);
      iPos2 := StrToInt(StrList4.Strings[iRow1+1]);
      sStr4 := Copy(sStr3, iPos1+2, iPos2 - iPos1 - 1);
      if Length(sStr4) > 2 then
      begin
        if (Copy(sStr4, 1, 1) = '"') and (Copy(sStr4, Length(sStr4), 1) = '"') then
        begin
          sStr5 := Copy(sStr4, 2, Length(sStr4)-2);
        end
        else
        begin
          sStr5 := sStr4;
        end;
      end
      else
      begin
          sStr5 := sStr4;
      end;
      //ǉ
      CsvList.Add(sStr5);
    end;
  end;

  //j
  StrList1.Free;
  StrList2.Free;
  StrList3.Free;
  StrList4.Free;
end;

procedure TFormViewGrid.SetSameAsAnnsView(Sender: TObject);
var
  IniFile: TIniFile;
  sStr3: String;
  iRow1: Integer;
begin
  //f
  if MenuSameAsAnnsView.Checked = TRUE then
  begin
    if FormMain.StrAnnsFMPath = '' then
    begin
      sStr3 := ExtractFilePath(ParamStr(0)) + ANNSFM_INI;
    end
    else
    begin
      sStr3 := FormMain.StrAnnsFMPath + ANNSFM_INI;
      if FileExists(sStr3) = FALSE then
      begin
        sStr3 := ExtractFilePath(ParamStr(0)) + ANNSFM_INI;
      end;
    end;
    if FileExists(sStr3) = TRUE then
    begin
      //m
      IniFile := TIniFile.Create(sStr3);
      //f
      StringGrid1.Font.Name                 :=                     IniFile.ReadString ('annsView', 'annsViewFontName',      'lr SVbN');
      StringGrid1.Font.Size                 :=                     IniFile.ReadInteger('annsView', 'annsViewFontHeight',    14);
      StringGrid1.Font.Color                := StringToColor('$' + IniFile.ReadString ('annsView', 'annsViewColorText',     'FFFFFF'));
      StringGrid1.Color                     := StringToColor('$' + IniFile.ReadString ('annsView', 'annsViewColorBack',     '400000'));
      IntTextHeight := TEXT_HEIGHT * StringGrid1.Font.Size div TEXT_SIZE;
        StringGrid1.RowHeights[0] := IntTextHeight;
        StringGrid1.RowHeights[1] := IntTextHeight;
      for iRow1 := 2 to  StringGrid1.RowCount-1 do
      begin
        StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
      end;
      Memo1.Font.Name                       :=                     IniFile.ReadString ('annsView', 'annsViewFontName',      'lr SVbN');
      Memo1.Font.Size                       :=                     IniFile.ReadInteger('annsView', 'annsViewFontHeight',    14);
      Memo1.Font.Color                      := StringToColor('$' + IniFile.ReadString ('annsView', 'annsViewColorText',     'FFFFFF'));
      Memo1.Color                           := StringToColor('$' + IniFile.ReadString ('annsView', 'annsViewColorBack',     '400000'));
      Memo1.Height                          := (IntTextHeight * IntMemoHeightIndex) + (IntTextHeight * 2);
      Panel3.Height                         := (IntTextHeight * IntMemoHeightIndex) + (IntTextHeight * 2);
      //j
      IniFile.Free;
    end
    else
    begin
      //f
      StringGrid1.Font.Name                 := 'lr SVbN';
      StringGrid1.Font.Size                 := 14;
      StringGrid1.Font.Color                := StringToColor('$' + 'FFFFFF');
      StringGrid1.Color                     := StringToColor('$' + '400000');
      IntTextHeight := TEXT_HEIGHT * StringGrid1.Font.Size div TEXT_SIZE;
        StringGrid1.RowHeights[0] := IntTextHeight;
        StringGrid1.RowHeights[1] := IntTextHeight;
      for iRow1 := 2 to  StringGrid1.RowCount-1 do
      begin
        StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
      end;
      Memo1.Font.Name                       := 'lr SVbN';
      Memo1.Font.Size                       := 14;
      Memo1.Font.Color                      := StringToColor('$' + 'FFFFFF');
      Memo1.Color                           := StringToColor('$' + '400000');
      Memo1.Height                          := (IntTextHeight * IntMemoHeightIndex) + (IntTextHeight * 2);
      Panel3.Height                         := (IntTextHeight * IntMemoHeightIndex) + (IntTextHeight * 2);
    end;
  end
  else
  begin
      //f
      StringGrid1.Font.Name  := FontDialog1.Font.Name;
      StringGrid1.Font.Size  := FontDialog1.Font.Size;
      StringGrid1.Font.Color := FontDialog1.Font.Color;
      StringGrid1.Color      := ColorDialog1.Color;
      IntTextHeight := TEXT_HEIGHT * StringGrid1.Font.Size div TEXT_SIZE;
        StringGrid1.RowHeights[0] := IntTextHeight;
        StringGrid1.RowHeights[1] := IntTextHeight;
      for iRow1 := 2 to  StringGrid1.RowCount-1 do
      begin
        StringGrid1.RowHeights[iRow1] := IntTextHeight * IntGridRatioIndex;
      end;
      Memo1.Font.Name        := FontDialog1.Font.Name;
      Memo1.Font.Color       := FontDialog1.Font.Color;
      Memo1.Font.Size        := FontDialog1.Font.Size;
      Memo1.Color            := ColorDialog1.Color;
      Memo1.CrLfColor        := ColorDialog4.Color;
      Memo1.Height                          := (IntTextHeight * IntMemoHeightIndex) + (IntTextHeight * 2);
      Panel3.Height                         := (IntTextHeight * IntMemoHeightIndex) + (IntTextHeight * 2);
  end;
end;

procedure TFormViewGrid.Memo1Click(Sender: TObject);
begin
  //XV
  MemoStatus(Sender);
end;

procedure TFormViewGrid.Memo1KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  //XV
  MemoStatus(Sender);
end;

procedure TFormViewGrid.Memo1KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  //XV
  MemoStatus(Sender);

  //s\邽߂Ɏs
  SendMessage(Memo1.Handle, WM_PAINT, 0, 0);
end;

procedure TFormViewGrid.Memo1MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  //XV
  MemoStatus(Sender);
end;

procedure TFormViewGrid.MemoStatus(Sender: TObject);
var
  iCount1: Integer;
  iLine1: Integer;
  iSel1: Integer;
  iSel2: Integer;
begin
  if Memo1.Text = '' then
  begin
    if BoolJapanese = TRUE then
    begin
      //\
      StatusBar3.Panels.Items[0].Text := Format('%d', [0]) + 's,' +
                                         Format('%d', [0]) + '';
      StatusBar3.Panels.Items[1].Text := Format('%d', [0]) + 's';
    end
    else
    begin
      //\
      StatusBar3.Panels.Items[0].Text := 'Row:'   + Format('%d', [0]) +
                                         '  Col:' + Format('%d', [0]);
      StatusBar3.Panels.Items[1].Text := 'Row:'   + Format('%d', [0]);
    end;
    Exit;
  end;

  // -1nƌ݂̃Lbgʒu̍sԍ(Ox[X)
  //@@  iLine1  := RichEdit1.Perform(EM_LINEFROMCHAR, -1, 0);
  iLine1  := Memo1.Perform(EM_LINEFROMCHAR, UINT(-1), 0);
  // -1nČ݂̃Lbgʒu̍s̐擪ʒu(Ox[X)
  //@@  iSel1   := RichEdit1.Perform(EM_LINEINDEX,    -1, 0);
  iSel1   := Memo1.Perform(EM_LINEINDEX,    UINT(-1), 0);
  //Lbgʒu(Ox[X)
  iSel2   := Memo1.SelStart;
  //s
  //  iTotal1 := RichEdit1.Lines.Count;
  iCount1 := Memo1.Lines.Count;

  if Memo1.SelLength > 0 then
  begin
    if BoolJapanese = TRUE then
    begin
      //\
      StatusBar3.Panels.Items[0].Text := Format('%d', [iLine1+1]) + 's,' +
                                         Format('%d', [0])        + '';
      StatusBar3.Panels.Items[1].Text := Format('%d', [iCount1])  + 's';
    end
    else
    begin
      //\
      StatusBar3.Panels.Items[0].Text := 'Row:'   + Format('%d', [iLine1+1]) +
                                         '  Col:' + Format('%d', [0]);
      StatusBar3.Panels.Items[1].Text := 'Row:'   + Format('%d', [iCount1]);
    end;
  end
  else
  begin
    if BoolJapanese = TRUE then
    begin
      //\
      StatusBar3.Panels.Items[0].Text := Format('%d', [iLine1+1])      + 's,' +
                                         Format('%d', [iSel2-iSel1+1]) + '';
      StatusBar3.Panels.Items[1].Text := Format('%d', [iCount1])       + 's';
    end
    else
    begin
      //\
      StatusBar3.Panels.Items[0].Text := 'Row:'   + Format('%d', [iLine1+1]) +
                                         '  Col:' + Format('%d', [iSel2-iSel1+1]);
      StatusBar3.Panels.Items[1].Text := 'Row:'   + Format('%d', [iCount1]);
    end;
  end;
end;

function  TFormViewGrid.ConvertCellToCsvEx2(SrcString: String; Sender: TObject): String;
var
  sStr1: String;
  sStr2: String;
  iPos1: Integer;
  iPos2: Integer;
  iPos3: Integer;
begin
  sStr1 := StringReplace(SrcString, #13#10, #10, [rfReplaceAll]);
  iPos1 := Pos(',', sStr1);
  iPos2 := Pos('"', sStr1);
  iPos3 := Pos(#10, sStr1);
  if (iPos1 > 0) or (iPos2 > 0) or (iPos3 > 0) then
  begin
    sStr2 := StringReplace(sStr1, '"', '""', [rfReplaceAll]);
    Result := '"' + sStr2 + '"';
  end
  else
  begin
    Result := sStr1;
  end;
end;

end.
